%{
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include "az80.tab.h"

extern char lstbuff[256];
extern char *plst;
extern int line_no;
extern int pass_no;
extern FILE *inpfile;
extern unsigned short PC;

static const char *valid_b = "01";
static const char *valid_o = "01234567";
static const char *valid_d = "0123456789";
static const char *valid_x = "0123456789ABCDEF";
int aton(char *src, int base, const char *valid, char end);

#define	TOKEN do { \
	if (!plst) \
		break; \
	plst += snprintf(plst, sizeof(lstbuff) - (size_t)(plst - lstbuff), "%s", yytext); \
} while (0)

%}

delim	[\t ,;()]
bin	[0-1]
oct	[0-7]
dec	[0-9]
alpha	[A-Za-z_]
hex	[0-9A-Fa-f]
alnum	[0-9A-Za-z_]

%s	cc

%%

<cc>[Nn][Zz]		{ TOKEN; return _cNZ; }
<cc>[Zz]		{ TOKEN; return _cZ; }
<cc>[Nn][Cc]		{ TOKEN; return _cNC; }
<cc>[Cc]		{ TOKEN; return _cC; }
<cc>[Pp][Oo]		{ TOKEN; return _cPO; }
<cc>[Pp][Ee]		{ TOKEN; return _cPE; }
<cc>[Pp]		{ TOKEN; return _cP; }
<cc>[Mm]		{ TOKEN; return _cM; }

[Bb][Cc]		{ TOKEN; return _BC; }
[Dd][Ee]		{ TOKEN; return _DE; }
[Hh][Ll]		{ TOKEN; return _HL; }
[Ss][Pp]		{ TOKEN; return _SP; }
[Aa][Ff]		{ TOKEN; return _AF; }

[Ii][Xx]		{ TOKEN; return _IX; }
[Ii][Yy]		{ TOKEN; return _IY; }

[Bb]			{ TOKEN; return _B; }
[Cc]			{ TOKEN; return _C; }
[Dd]			{ TOKEN; return _D; }
[Ee]			{ TOKEN; return _E; }
[Hh]			{ TOKEN; return _H; }
[Ll]			{ TOKEN; return _L; }
"("[Hh][Ll]")"		{ TOKEN; return _M; }
[Aa]			{ TOKEN; return _A; }
[Ff]			{ TOKEN; return _F; }
[Ii]			{ TOKEN; return _I; }
[Rr]			{ TOKEN; return _R; }
[Hh][Xx]		{ TOKEN; return _HX; }
[Ll][Xx]		{ TOKEN; return _LX; }
[Hh][Yy]		{ TOKEN; return _HY; }
[Ll][Yy]		{ TOKEN; return _LY; }

"("[Cc]")"		{ TOKEN; return _PORTC; }

[Aa][Dd][Cc]		{ TOKEN; return _ADC; }
[Aa][Dd][Dd]		{ TOKEN; return _ADD; }
[Aa][Nn][Dd]		{ TOKEN; return _AND; }
[Bb][Ii][Tt]		{ TOKEN; return _BIT; }
[Cc][Aa][Ll][Ll]	{ TOKEN; BEGIN(cc); return _CALL; }
[Cc][Cc][Ff]		{ TOKEN; return _CCF; }
[Cc][Pp]		{ TOKEN; return _CP; }
[Cc][Pp][Dd]		{ TOKEN; return _CPD; }
[Cc][Pp][Dd][Rr]	{ TOKEN; return _CPDR; }
[Cc][Pp][Ii]		{ TOKEN; return _CPI; }
[Cc][Pp][Ii][Rr]	{ TOKEN; return _CPIR; }
[Cc][Pp][Ll]		{ TOKEN; return _CPL; }
[Dd][Aa][Aa]		{ TOKEN; return _DAA; }
[Dd][Ee][Cc]		{ TOKEN; return _DEC; }
[Dd][Ii]		{ TOKEN; return _DI; }
[Dd][Jj][Nn][Zz]	{ TOKEN; return _DJNZ; }
[Ee][Ii]		{ TOKEN; return _EI; }
[Ee][Xx]		{ TOKEN; return _EX; }
[Ee][Xx][Xx]		{ TOKEN; return _EXX; }
[Hh][Aa][Ll][Tt]	{ TOKEN; return _HALT; }
[Ii][Mm]		{ TOKEN; return _IM; }
[Ii][Nn]		{ TOKEN; return _IN; }
[Ii][Nn][Cc]		{ TOKEN; return _INC; }
[Ii][Nn][Dd]		{ TOKEN; return _IND; }
[Ii][Nn][Dd][Rr]	{ TOKEN; return _INDR; }
[Ii][Nn][Ii]		{ TOKEN; return _INI; }
[Ii][Nn][Ii][Rr]	{ TOKEN; return _INIR; }
[Jj][Pp]		{ TOKEN; BEGIN(cc); return _JP; }
[Jj][Rr]		{ TOKEN; BEGIN(cc); return _JR; }
[Ll][Dd]		{ TOKEN; return _LD; }
[Ll][Dd][Dd]		{ TOKEN; return _LDD; }
[Ll][Dd][Dd][Rr]	{ TOKEN; return _LDDR; }
[Ll][Dd][Ii]		{ TOKEN; return _LDI; }
[Ll][Dd][Ii][Rr]	{ TOKEN; return _LDIR; }
[Nn][Ee][Gg]		{ TOKEN; return _NEG; }
[Nn][Oo][Pp]		{ TOKEN; return _NOP; }
[Oo][Rr]		{ TOKEN; return _OR; }
[Oo][Tt][Dd][Rr]	{ TOKEN; return _OTDR; }
[Oo][Tt][Ii][Rr]	{ TOKEN; return _OTIR; }
[Oo][Uu][Tt]		{ TOKEN; return _OUT; }
[Oo][Uu][Tt][Dd]	{ TOKEN; return _OUTD; }
[Oo][Uu][Tt][Ii]	{ TOKEN; return _OUTI; }
[Pp][Oo][Pp]		{ TOKEN; return _POP; }
[Pp][Uu][Ss][Hh]	{ TOKEN; return _PUSH; }
[Rr][Ee][Ss]		{ TOKEN; return _RES; }
[Rr][Ee][Tt]		{ TOKEN; BEGIN(cc); return _RET; }
[Rr][Ee][Tt][Ii]	{ TOKEN; return _RETI; }
[Rr][Ee][Tt][Nn]	{ TOKEN; return _RETN; }
[Rr][Ll]		{ TOKEN; return _RL; }
[Rr][Ll][Aa]		{ TOKEN; return _RLA; }
[Rr][Ll][Cc]		{ TOKEN; return _RLC; }
[Rr][Ll][Cc][Aa]	{ TOKEN; return _RLCA; }
[Rr][Ll][Dd]		{ TOKEN; return _RLD; }
[Rr][Rr]		{ TOKEN; return _RR; }
[Rr][Rr][Aa]		{ TOKEN; return _RRA; }
[Rr][Rr][Cc]		{ TOKEN; return _RRC; }
[Rr][Rr][Cc][Aa]	{ TOKEN; return _RRCA; }
[Rr][Rr][Dd]		{ TOKEN; return _RRD; }
[Rr][Ss][Tt]		{ TOKEN; return _RST; }
[Ss][Bb][Cc]		{ TOKEN; return _SBC; }
[Ss][Cc][Ff]		{ TOKEN; return _SCF; }
[Ss][Ee][Tt]		{ TOKEN; return _SET; }
[Ss][Ll][Aa]		{ TOKEN; return _SLA; }
[Ss][Rr][Aa]		{ TOKEN; return _SRA; }
[Ss][Ll][Ll]		{ TOKEN; return _SLL; }
[Ss][Rr][Ll]		{ TOKEN; return _SRL; }
[Ss][Uu][Bb]		{ TOKEN; return _SUB; }
[Xx][Oo][Rr]		{ TOKEN; return _XOR; }

[Dd][Bb]		{ TOKEN; return _DB; }
[Dd][Ww]		{ TOKEN; return _DW; }
[Dd][Dd]		{ TOKEN; return _DD; }
[Dd][Ss]		{ TOKEN; return _DS; }
[Dd][Ee][Ff][Bb]	{ TOKEN; return _DEFB; }
[Dd][Ee][Ff][Ww]	{ TOKEN; return _DEFW; }
[Dd][Ee][Ff][Dd]	{ TOKEN; return _DEFD; }
[Dd][Ee][Ff][Ll]	{ TOKEN; return _DEFL; }
[Dd][Ee][Ff][Mm]	{ TOKEN; return _DEFM; }
[Dd][Ee][Ff][Ss]	{ TOKEN; return _DEFS; }
[Ee][Nn][Dd]		{ TOKEN; return _END; }
[Ee][Qq][Uu]		{ TOKEN; return _EQU; }
[Aa][Ll][Ii][Gg][Nn]	{ TOKEN; return _ALIGN; }
[Ee][Vv][Ee][Nn]	{ TOKEN; return _EVEN; }
[Oo][Rr][Gg]		{ TOKEN; return _ORG; }

{bin}+[Bb]		{
				TOKEN;
				yylval.i = aton(yytext,2,valid_b,'B');
				return _IMM;
			}

{oct}+[Oo]		{
				TOKEN;
				yylval.i = aton(yytext,8,valid_o,'O');
				return _IMM;
			}

{dec}+[Dd]?		{
				TOKEN;
				yylval.i = aton(yytext,10,valid_d,'D');
				return _IMM;
			}

{hex}+[Hh]		{
				TOKEN;
				yylval.i = aton(yytext,16,valid_x,'H');
				return _IMM;
			}

"0"[Xx]{hex}+		{
				TOKEN;
				yylval.i = aton(yytext+2,16,valid_x,0);
				return _IMM;
			}

"$"			{ TOKEN; yylval.i = PC; return _IMM; }
","			{ TOKEN; return COMMA; }
":"                     { TOKEN; return COLON; }
"("			{ TOKEN; return LPAREN; }
")"			{ TOKEN; return RPAREN; }
"'"			{ TOKEN; return QUOTE; }
"+"			{ TOKEN; return PLUS; }
"-"			{ TOKEN; return MINUS; }
"*"			{ TOKEN; return MUL; }
"/"			{ TOKEN; return DIV; }
"<<"			{ TOKEN; return SHL; }
">>"			{ TOKEN; return SHR; }
"&"                     { TOKEN; return AND; }
"|"                     { TOKEN; return OR; }
"^"                     { TOKEN; return XOR; }
"~"                     { TOKEN; return NOT; }

\"[^\"\n]+\"            {	/* anything enclosed in double quotes, w/o CR */
				TOKEN;
				yylval.s = strdup(yytext + 1);
				yylval.s[strlen(yylval.s)-1] = '\0';
				return _STR;
			}

\'[^\'\n]+\'		{	/* anything enclosed in single quotes, w/o CR */
				TOKEN;
				yylval.i = yytext[1];
				if (yyleng > 3)
					yylval.i += 256 * yytext[2];
				return _IMM;
			}

{alpha}{alnum}*		{	/* alpha followed by alpha numerical */
				TOKEN;
				yylval.s = strdup(yytext);
				return _SYM;
			}

[\t ]			{	/* white space */
				TOKEN;
			}

";"[^\n]*\n		|	/* comment until CR */
[\n]			{	/* CR */
				TOKEN;
				if (plst)
					plst = lstbuff;
				BEGIN(INITIAL);
				line_no++;
				return EOL;
			}

%%

int aton(char *src, int base, const char *valid, char end)
{
	int result = 0;
	int sign = 1;
	char * pos;

	while (*src && !strchr(valid, toupper(*src)) && (toupper(*src) != end)) {
		if (*src == '+')
			sign = +1;
		if (*src == '-')
			sign = -sign;
		src++;
	}
	while (*src && ((pos = strchr(valid, toupper(*src))) != NULL)) {
		result *= base;
		if (pos)
			result += (int)(pos - valid);
		src++;
	}
	result *= sign;
	return result;
}

void yyerror(char * msg, ...)
{
	va_list arg;
	fprintf(stderr, "line %d: ", line_no);
	va_start(arg, msg);
	fprintf(stderr, msg, arg);
	va_end(arg);
	fprintf(stderr, " at \"%s\"\n", yytext);
}

int yywrap(void)
{
	if (pass_no < 2) {
		rewind(inpfile);
		PC = 0;
		line_no = 1;
		pass_no++;
		printf("pass %d\n", pass_no);
		return 0;
	}
	return 1;
}

