'******************************************************************************
'*  pm80.spin - a recompiling Z80 emulator
'*
'*  (c) Juergen Buchmueller <pullmoll@t-online.de>
'*
'* $Id: pm80.spin,v 1.88 2010-03-16 17:11:54 pm Exp $
'******************************************************************************
CON
	_xinfreq		= 5_000_000		' Quartz is 5MHz
	_clkmode		= xtal1 + pll16x	' System clock is 80MHz

' Define if you want to run this on the TriBladeProp #2
'#define TriBladeProp

' Uncomment to have debug code inserted
'#define DEBUG
'#define SSTEP

'#define XYFLAGS

	CF			=	%00000001	' carry
	NF			=	%00000010	' negative
	PF			=	%00000100	' parity (also overflow)
	XF			=	%00001000	' undocumented x
	HF			=	%00010000	' auxiliaray (half) carry
	YF			=	%00100000	' undocumented y
	ZF			=	%01000000	' zero
	SF			=	%10000000	' sign

	bit0			=	%00000001
	bit1			=	%00000010
	bit2			=	%00000100
	bit3			=	%00001000
	bit4			=	%00010000
	bit5			=	%00100000
	bit6			=	%01000000
	bit7			=	%10000000

	io_cmd_out		=	$01
	io_cmd_in		=	$02
	io_cmd_break		=	$03

	constat_port		=	$10
	condata_port		=	$11

	PATCH_SHIFT		=	16

'******************************************************************************
'
' Structure of a compiled Z80 opcode.
'
'  long #     meaning
' ---------------------------------------------------------------------
'       0     contains the Z80 PC for the opcode (of the first byte)
'       1     contains "add R, #1" PASM code
'       2     code pasted from the code snippet
'      ...
'      n+2    last code long
'      n+3    contains the Z80 PC of the next opcode
'
'******************************************************************************

VAR
	long  cog

PUB start(ram, io) : okay

'' Start CPU simulator - starts a cog
'' returns false if no cog available
''
'' ram is a pointer to Z80 RAM space
'' io is a pointer to Z80 I/O space

	stop

#ifdef TriBladeProp
#else
	ram_base := ram
#endif
	io_command := io
	io_port := io + 4
	io_data := io + 8

	okay := cog := COGNEW(@entry, 0) + 1

PUB stop

'' Stop CPU simulator - frees a cog

	if cog
		cogstop(cog~ - 1)

DAT
'---------------------------------------------------------------------------------------------------------

		org	0
entry
'---------------------------------------------------------------------------------------------------------
#ifdef TriBladeProp
'The TriBladeProp Blade #2 driver presets the latch to enable SRAM -CE(U23)=0 and hold LE=0
datax		mov     outa, ram_latched	' -WE=1 -OE=1, all other bits=0
outx		mov     dira, ram_dir_read	' setup direction to read
#else
hubptr		jmp	#startup		' contains hub RAM address for ea (overwrites jmp)
#endif

'******************************************************************************
'
' cog variables
'
#ifdef TriBladeProp
#else
ram_base	long	0-0		' base address of the Z80 RAM in hub RAM
#endif
io_command	long    0-0		' io command address
io_port		long    0-0		' io port number
io_data		long    0-0		' io port data

table_00	long	@@@snippets_00
table_cb	long	@@@snippets_cb
table_xy	long	@@@snippets_xy
table_ed	long	@@@snippets_ed
table_xy_cb	long	@@@snippets_xy_cb
alt_regs	long	@@@alt_regs_array
IX		long	@@@IX_reg
IY		long	@@@IY_reg

low_16bits	long	$0000ffff

dst_pc		long	0		' compiler output program counter
max_pc		long	0		' max. address for compiler output
snippet		long	0		' start address of current code snippet in cog RAM
count		long	0		' count of longs to copy
daa_adj					' re-use this long in DAA
patch		long	0		' opcode patch coming from the table_xx

t1		long	0

ea		long	0		' effective address
opcode		long	0		' opcode

alu		long	0		' arithmetic logic unit (sort of :)
tmp		long	0		' generic temporary
aux		long	0		' auxiliary flag temporary

B		long	0		' B register \_ combined into BC on occasion
C		long	0		' C register /
D		long	0		' D register \_ combined into DE on occasion
E		long	0		' E register /
H		long	0		' H register \_ combined into HL on occasion
L		long	0		' L register /
F		long	0		' flags       \_ combined into AF on occasion
A		long	0		' accumulator /

XY		long	0		' pointer to the IX or IY register
SP		long	0		' stack pointer
PC		long	0		' program counter
PC0		long	0		' program counter at the start of this compiler sweep
PC1		long	0		' program counter at the start of this instruction
R		long	0		' R register (Z80 DRAM refresh)
R2_I		long	0		' bit #7 of last write to R register, I register in 15..8
HALT		long	0		' <>0 if halted

#ifdef	DEBUG
char		long	0
#endif

startup
		mov	XY, #0
'---------------------------------------------------------------------------------------------------------
' Z80 compiler
'---------------------------------------------------------------------------------------------------------
compile
#ifdef	DEBUG
		tjnz	HALT, #stackdump
#else
		tjnz	HALT, #compile
#endif
		mov	dst_pc, #ORIGIN			' start over to compile to ORIGIN++
		mov	max_pc, #$1f0
		mov	PC0, PC				' remember the program counter of this compiled block
compile_next
#ifdef	SSTEP
		cmp	dst_pc, #ORIGIN	WZ
	if_nz	jmp	#execute
#endif
		mov	PC1, PC				' save PC at start of this opcode
		mov	tmp, table_00
compile_op
		call	#rd_opcode			' get opcode
compile_op_2
		shl	opcode, #1			' * 2
		add	tmp, opcode			' offset into one of the code snippets tables (words)
compile_op_3
		rdword	ea, tmp				' get the address of the code snippet from the table
		rdlong	count, ea			' get number of longs to copy + the patch jump address
		mov	patch, count			' save a copy in patch for later
		and	count, low_16bits		' mask count only

#ifdef	DEBUG
			mov	hexval, PC1
			call	#hex_w
			mov	char, #$3a
			call	#conout_chr
			mov	hexval, opcode
			shr	hexval, #1
			call	#hex_b
			call	#conout_space
			mov	tmp, #B
			mov	aux, #8				' dump B, C, D, E, H, L, A and F
:regs8			movs	:getreg8, tmp
			add	tmp, #1
:getreg8		mov	hexval, 0-0
			call	#hex_b
			test	aux, #1	WZ
		if_nz	call	#conout_space
			djnz	aux, #:regs8
			mov	tmp, #8
			add	tmp, alt_regs
			rdword	hexval, tmp
			call	#hex_w
			call	#conout_space
			add	tmp, #2
			rdword	hexval, tmp
			call	#hex_w
			call	#conout_space
			mov	tmp, #SP
			mov	aux, #7				' dump SP, PC, PC0, PC1, R, R2_I, HALT
:regs16			movs	:getreg16, tmp
			add	tmp, #1
:getreg16		mov	hexval, 0-0
			call	#hex_w
			call	#conout_space
			djnz	aux, #:regs16
			mov	hexval, dst_pc
			sub	hexval, #ORIGIN
			call	#hex_b
			call	#conout_cr
#endif
		' verify that this code snippet still fits in cog RAM
		mov	tmp, dst_pc			' destination
		add	tmp, count			' + count longs
		add	tmp, #6				' + 6 longs safety margin
		cmp	tmp, max_pc	WZ,WC		' < last possible cog RAM address?
	if_ae	mov	PC, PC1				' no, have to compile this opcode again after execution
	if_ae	jmp	#execute			' and now execute what we compiled

		movd	:padr0, dst_pc			' patch destination PC address
		add	dst_pc, #1
:padr0		mov	0-0, PC1			' store PC of the Z80 instruction in the snippet's first long

		movd	:padr1, dst_pc			' patch destination PC address 
		add	dst_pc, #1
:padr1		mov	0-0, opcode_add_r_1		' store an "add R, #1" opcode in the cog RAM slot

		tjz	XY, #:notxy			' skip if not IX or IY instruction
		movd	:padr2, dst_pc
		movd	:padr3, dst_pc
:padr2		mov	0-0, opcode_mov_xy_src		' store a "mov XY, ..." opcode
:padr3		movs	0-0, XY				' store the address of either IX or IY in the source field
		add	dst_pc, #1
		mov	XY, #0				' clear XY for next opcode

:notxy		mov	snippet, dst_pc			' remember the start of pasted code in the snippet
		tjz	count, #:patcher		' if count is zero, copy nothing at all
:copy1		movd	:padr4, dst_pc			' store destination address in rdlong opcode
		add	ea, #4				' next source address (+ 1 long)
		rdlong	tmp, ea				' get code snippet PASM opcode
		shr	tmp, #(32-9)	WZ, NR		' see if this is a NOP (paste code snippet marker)
	if_z	jmp	#:snippet			' yes, paste the corresponding snippet of code
:padr4		mov	0-0, tmp			' store the PASM opcode
		add	dst_pc, #1			' next cog RAM address
:more		djnz	count, #:copy1			' copy the code snippet
:patcher	shr	patch, #PATCH_SHIFT	WZ	' get patcher function to bits 15..0
	if_nz	jmp	patch				' go to the patcher, if any
		jmp	#compile_next			' otherwise compile next instruction

'*******************************************************
' Paste another snippet of code into the current one
' snippet source address (in cog RAM) is in tmp
'
:snippet
		rdlong	aux, tmp			' fetch the number of longs to copy
:copy2		movd	:padr5, dst_pc			' store destination address in rdlong opcode
		add	tmp, #4				' next source address (+ 1 long)
:padr5		rdlong	0-0, tmp			' read the PASM opcode to cog RAM
		add	dst_pc, #1			' next cog RAM address
		sub	count, #1			' decrement count also
		djnz	aux, #:copy2			' copy the code snippet
		add	count, #1			' add 1 to count for the long that was skipped
		jmp	#:more

#ifdef	DEBUG
'*******************************************************
' Send a core dump to the terminal
'
stackdump
		mov	tmp, SP
		shl	aux, #64
:dump		mov	ea, tmp
		call	#rd_word
		mov	hexval, alu
		add	tmp, #2
		call	#hex_w
		call	#conout_space
		test	tmp, #15 wz
	if_z	call	#conout_cr
		djnz	aux, #:dump
:loop		jmp	#:loop
#endif

'*******************************************************
' Patch an immediate 16 bit (word) load in snippet 
'
do_patch_imm16
		sub	max_pc, #1			' one less long available
		movd	:padr1, max_pc			' set patch address for long at end of cog RAM
		movd	:padr2, snippet			' set patch address for address in code
		call	#rd_opword
:padr1		mov	0-0, alu			' store word in a long at the end of cog RAM
:padr2		movs	0-0, max_pc			' patch the code with the address of the word
do_patch_imm16_ret
		ret

'*******************************************************
' Patch an immediate 16 bit (word) load in snippet 
'
patch_imm16
		call	#do_patch_imm16
		jmp	#compile_next

'*******************************************************
' Try and see if the program counter stored in alu
' is found in the compiled code and if so, patch the
' 2nd instruction "mov ea, #0-0" to the address after
' the NOP containing the PC
'
patch_compiled_pc
		cmp	alu, PC0	WZ,WC
	if_b	jmp	#patch_compiled_pc_ret		' new PC is below this compiled block
		cmp	alu, PC	WZ,WC
	if_a	jmp	#patch_compiled_pc_ret		' new PC is beyond the current address
		mov	ea, dst_pc			' get the end of the compiler output
		sub	ea, #1
:loop		movs	:padr0, ea			' and start to search backwards
		sub	ea, #1
:padr0		mov	aux, 0-0			' get the PASM code
		cmp	aux, alu	WZ		' found the expected PC?
	if_nz	cmp	ea, #ORIGIN	WC		' hit bottom of compiled code?
  if_nz_and_c	jmp	#patch_compiled_pc_ret		' not found
	if_nz	jmp	#:loop
		mov	tmp, #1				' + 1 into the snippet's code
		add	tmp, snippet
		movd	:padr1, tmp			' set patch address
		add	ea, #2				' ea = address after the tinted NOP
:padr1		movs	0-0, ea				' patch this snippet's code with the found snippet's address
patch_compiled_pc_ret
		ret

'*******************************************************
' Patch an absolute jump address into the snippet
'
patch_jp_imm16
		call	#do_patch_imm16
		call	#patch_compiled_pc
		jmp	#compile_next

'*******************************************************
' Patch two imm8 (bytes) into the snippet
'
patch_imm8_1
		movd	:padr0, snippet			' set patch address for disp8 in code
		call	#rd_opcode
:padr0		movs	0-0, opcode			' patch the code with the byte
		add	snippet, #1
		' fall through

'*******************************************************
' Patch an imm8 (byte) into the snippet
'
patch_imm8
		movd	:padr0, snippet			' set patch address for imm8 in code
		call	#rd_opcode
:padr0		movs	0-0, opcode			' patch the code with the byte
		jmp	#compile_next

'*******************************************************
' Patch a relative jump address into the snippet
'
patch_disp8
		movd	:padr1, snippet			' set patch address for new PC in code
		sub	max_pc, #1			' make room for another imm16
		movd	:padr2, max_pc			' where to store the imm16
		call	#rd_opcode
		shl	alu, #24			' sign extend the byte
		sar	alu, #24			' into a long
		add	alu, PC				' alu = new target address
		and	alu, low_16bits			' mask bits 15..0
:padr1		movs	0-0, max_pc			' store address of new PC imm16 in the snippet
:padr2		mov	0-0, alu			' store alu in the top cog RAM slot
		call	#patch_compiled_pc
		jmp	#compile_next

'*******************************************************
' Patch an call immediate 16 in snippet
'
patch_call_imm16
		call	#do_patch_imm16
rst_execute
		sub	max_pc, #1			' one less long available
		add	snippet, #1
		movd	:padr1, max_pc			' set patch address for long at end of cog RAM
		movd	:padr2, snippet			' set patch address for address in code
:padr1		mov	0-0, PC				' store next PC in a long at the end of cog RAM
:padr2		movs	0-0, max_pc			' patch the code with the address of the word
		jmp	#execute

'*******************************************************
' Patch the address of the snippet into the snippet
'
patch_self_1
		movd	:padr, snippet
		sub	snippet, #1
:padr		movs	0-0, snippet
		jmp	#compile_next

'*******************************************************
' Get opcode snippet for CB xx opcode
'
prefix_cb
		sub	dst_pc, #2
		mov	tmp, table_cb
		jmp	#compile_op

'*******************************************************
' Get opcode snippet for ED xx opcode
'
prefix_ed
		sub	dst_pc, #2
		mov	tmp, table_ed
		jmp	#compile_op

'*******************************************************
' Get opcode snippet for DD xx opcode
'
prefix_dd
		mov	tmp, #IX			' IX contains the address of IX_reg
prefix_ddfd
		sub	dst_pc, #2			' discard previous "long PC" and "add R, #1" opcodes
		mov	XY, tmp				' store in XY so the code loading it will be prepended
		call	#rd_opcode
		cmp	opcode, #$cb		WZ
	if_z	jmp	#prefix_xd_cb
		mov	tmp, table_xy
		jmp	#compile_op_2

'*******************************************************
' Get opcode snippet for FD xx opcode
'
prefix_fd
		mov	tmp, #IY			' IY contains the address of IY_reg
		jmp	#prefix_ddfd

'*******************************************************
' Get opcode snippet for DD CB disp8 xx opcode
'
prefix_xd_cb
		call	#rd_opcode			' read disp8 byte
		movs	prefix_xy_cb_padr1, opcode	' store in patcher
		mov	tmp, table_xy_cb
		call	#rd_opcode			' fetch 4th byte opcode
		shr	alu, #3				' ignore bits 2..0
		shl	alu, #1				' * 2
		add	tmp, alu			' table_xy_cb offset
		jmp	#compile_op_3

'*******************************************************
' Get opcode snippet for DD/FD CB disp8 xx opcode
'
prefix_xy_cb
		movd	prefix_xy_cb_padr1, snippet	' destination for the disp8 is the first opcode long
		mov	alu, opcode			' get opcode
prefix_xy_cb_padr1
		movs	0-0, #0				' move imm8 into snippet
		mov	tmp, alu
		and	alu, #%00000111			' mask lower 3 bits
		cmp	alu, #6	WZ			' regular (IXY+disp8) opcode?
	if_z	jmp	#compile_next
		and	tmp, #%11000000			' mask upper 2 bits
		cmp	tmp, #%01000000	WZ		' is it the "BIT n, (IXY+disp8)" group?
	if_z	jmp	#compile_next
		movd	:padr3, dst_pc			' write to dst_pc
		movd	:padr4, dst_pc			' here too
:padr3		mov	0-0, opcode_mov_dst_alu		' store a "MOV dst, alu" opcode after the snippet
		add	alu, #B
		movs	:padr4, alu			' address of B, C, D, E, H, L or A in alu
		add	dst_pc, #1			' one more long for the snippet
:padr4		movd	0-0, #0-0			' store the 8bit register address to also receive the alu result
		jmp	#compile_next

'*******************************************************
opcode_add_r_1
		add	R, #1				' code to prepend to every instruction
opcode_mov_xy_src
		mov	XY, 0-0				' code to prepend to IX and IY opcodes
'*******************************************************

'*******************************************************
' A snippet has modified compiled code. We need to find
' the Z80 PC of the following snippet and then
' recompile from there on
'snippet_pc
'		and	tmp, #$1ff			' mask the return address of the modifying snippet
':loop		movs	:padr0, tmp
'		add	tmp, #1
':padr0		mov	aux, 0-0			' fetch the PASM opcode
'		shr	aux, #(32-9)	WZ, NR		' test for a NOP opcode
'	if_nz	jmp	#:loop				' let's hope this cannot fail to find the last PC long
'		jmp	#compile_from_aux

'*******************************************************
' Handle a return from subroutine
'
handle_ret
		call	#pop_alu
opcode_mov_dst_alu
		mov	PC, alu				' and set the new Z80 program counter
opcode_jmp_compile
		jmp	#compile

'*******************************************************
' Handle an absolute call to address aux
'
handle_call
		call	#push_alu
handle_jp
		mov	PC, aux				' set the new Z80 program counter
		jmp	#compile

#ifdef TriBladeProp

'******************************************************************************
'
'	Memory access functions (TriBladeProp)
'
'******************************************************************************
'                                          +-------- -CS (U26 FLASH)   -+                                  
'                                          |+------- -CS (J22 microSD)  +- Latched pins...                 
'                                          ||+------ -RST3              |    passes when LE=1              
'             (P30) SO ---+                |||+----- -RST1              |    latches on  LE low going edge 
'             (P31) SI --+|                ||||+---- -CE (U24 SRAM-hi)  |    latched on  LE=0              
'                        ||+---- -WE(SDA)  |||||+--- -CE (U23 SRAM-lo)  |                                  
'                        |||+--- -OE(SCL)  ||||||+-- A20                |                                  
'                        ||||+--  LE       |||||||+- A19               -+                                  
'                        |||||             ||||||||                                                        
'          P31...P0 --> %00xx0_xxxxxxxxxxx_xxxxxxxx_xxxxxxxx  
ram_latched	long	%00110_00000000000_00000000_00000000	' -WE=1 -OE=1 (LE=0)                               
ram_read	long	%00100_00000000000_00000000_00000000	' -WE=1 -OE=0 (LE=0)    
ram_write	long	%00010_00000000000_00000000_00000000	' -WE=0 -OE=1 (LE=0)
ram_dir_read	long	%00110_11111111111_11111111_00000000	' outputs WE, OE, A0-18, inputs D0-7
ram_dir_write	long	%00110_11111111111_11111111_11111111	' outputs WE, OE, A0-18, D0-7
ram_dir_input	long	%00000_00000000000_00000000_00000000	' all inputs (for Ram Disk access)
'                              +---------+ +------+ +------+
'                              A18......A8 A7....A0 D7....D0
'                                          Q7....Q0
'

'*******************************************************
' Read a byte from RAM[HL] into the alu
'
rd_byte_hl
		mov	ea, H
		shl	ea, #8
		or	ea, L
		' fall through

'Read a BYTE from location given by "address" of Z80 RAM into byte data_8
rd_byte
		and	ea, low_16bits
		mov	outx, ea
		shl	outx, #8			' shift address 8 bits
		or	outx, ram_read			' add -WE=1 -OE=0
		mov	outa, outx			' pc (shifted 8 bits + -WE=1 -OE=0)
		nop					' delay for sram access time
		mov     alu, ina			' read SRAM
		and     alu, #$ff			' extract 8 bits
rd_byte_ret
rd_byte_hl_ret
		ret

'*******************************************************
' Write a byte from alu to RAM[HL]
'
wr_byte
		and	ea, low_16bits
		mov	outx, ea
		shl	outx, #8			' shift address 8 bits
		or	outx, ram_write			' add -WE=0 -OE=1
		mov	datax, alu
		and	datax, #$ff			' ensure upper bytes=0
		or	outx, datax			' add data bits to be written
		mov	outa, outx			' output -WE=0 -OE=1, address bits, data bits
		mov	dira, ram_dir_write		' set data bits to outputs
'		nop					' 25nS min delay (nothing reqd)
		or      outa, ram_latched		' set -WE=0->1 -OE=1->1  (end write)
		mov     dira, ram_dir_read		' set data bits to inputs
		mov     outa, ram_read			' set -WE=1 -OE=0
wr_byte_ret
		ret
#else

'******************************************************************************
'
'	Memory access functions (hub RAM)
'
'******************************************************************************

'*******************************************************
' Read a byte from RAM[HL] into the alu
'
rd_byte_hl
		mov	ea, H
		shl	ea, #8
		or	ea, L
		' fall through

'*******************************************************
' Read a byte from RAM[ea] into the alu
'
rd_byte
		and	ea, low_16bits
		mov	hubptr, ea
		add	hubptr, ram_base
		rdbyte	alu, hubptr
rd_byte_ret
rd_byte_hl_ret
		ret

'*******************************************************
' Write a byte from the alu to RAM[ea]
'
wr_byte
		and	ea, low_16bits
		mov	hubptr, ea
		add	hubptr, ram_base
		wrbyte	alu, hubptr
wr_byte_ret
		ret
#endif

'*******************************************************
' Read a byte from RAM[PC++] into the opcode
'
rd_opcode
		mov	ea, PC
		call	#rd_byte
		add	PC, #1
		and	PC, low_16bits
		mov	opcode, alu
rd_opcode_ret
		ret

'*******************************************************
' Write a word from the alu to RAM[ea]
'
wr_word
		call	#wr_byte
		add	ea, #1
		ror	alu, #8
		call	#wr_byte
		rol	alu, #8				' restore alu
		sub	ea, #1				' restore ea, as it may be re-used
:ok_to_write
wr_word_ret
		ret

rd_opword
		mov	ea, PC
		add	PC, #2
		and	PC, low_16bits

'*******************************************************
' Read a word from RAM[ea] into the alu
'
' clobbers aux
'
rd_word
		call	#rd_byte
		add	ea, #1
		mov	aux, alu
		call	#rd_byte
		sub	ea, #1				' restore ea, as it may be re-used
		shl	alu, #8
		or	alu, aux
rd_word_ret
rd_opword_ret
		ret

'*******************************************************
' Write a byte from the alu to an I/O port
'
wr_port_alu
		wrword	ea, io_port
		wrbyte	alu, io_data
#ifdef TriBladeProp
		mov     dira, ram_dir_input		' pass I/O to another cog
#endif
		mov	t1, #io_cmd_out
		wrbyte	t1, io_command
:wait
		rdbyte	t1, io_command	WZ		' ...and wait for it to be completed
	if_nz	jmp	#:wait
#ifdef TriBladeProp
		mov     dira, ram_dir_read		' take I/O from another cog
#endif
wr_port_alu_ret
		ret

'*******************************************************
' Read a byte from an I/O port into the alu
'
rd_port_alu
		wrword	ea, io_port
#ifdef TriBladeProp
		mov     dira, ram_dir_input		' pass I/O to another cog
#endif
		mov	t1, #io_cmd_in
		wrbyte	t1, io_command
:wait
		rdbyte	t1, io_command WZ		'...and wait for it to be completed.
	if_nz	jmp	#:wait
#ifdef TriBladeProp
		mov     dira, ram_dir_read		' take I/O from another cog
#endif
		rdbyte	alu, io_data
rd_port_alu_ret
		ret


'*******************************************************
' Pop a word from the Z80 stack to alu
'
pop_alu
		mov	ea, SP
		add	SP, #2
		and	SP, low_16bits
		call	#rd_word			' into alu
pop_alu_ret
		ret

'*******************************************************
' Push a word from alu to the Z80 stack
'
push_alu
		sub	SP, #2
		and	SP, low_16bits
		mov	ea, SP
		call	#wr_word			' from alu
push_alu_ret
		ret

'******************************************************************************
'******************************************************************************
'
'	Debug code
'
'******************************************************************************
'******************************************************************************
#ifdef	DEBUG
conout_space
		mov	char, #32
		jmp	#conout_chr
conout_cr
		mov	char, #13
conout_chr
		mov	t1, #condata_port
		wrbyte	t1, io_port
		wrbyte	char, io_data
		mov	t1, #io_cmd_out
		wrbyte	t1, io_command
:wait
		rdbyte	t1, io_command
		tjnz	t1, #:wait
conout_chr_ret
conout_cr_ret
conout_space_ret
		ret

hexcnt		long	0
hexval		long	0
hex_l
		mov	hexcnt, #8
hex_loop
		rol	hexval, #4
		mov	char, hexval
		and	char, #15
		cmp	char, #10	WZ,WC
	if_b	add	char, #48
	if_ae	add	char, #55
		call	#conout_chr
		djnz	hexcnt, #hex_loop
hex_loop_ret
hex_l_ret
		ret

hex_b
		mov	hexcnt, #2
		shl	hexval, #24
		call	#hex_loop
hex_b_ret
		ret
hex_w
		mov	hexcnt, #4
		shl	hexval, #16
		call	#hex_loop
hex_w_ret
		ret
#endif


'*******************************************************
' Patch an imm16 into the code and then prepare
' the execution of the compiled code
'
execute_imm16
		call	#do_patch_imm16

'*******************************************************
' Prepare execution of the compiled code
'
execute
		' store the current Z80 PC in the code
		movd	:padr0, dst_pc			' store destination address in the mov opcode
		add	dst_pc, #1
:padr0		mov	0-0, PC				' next PC
		' store a "jmp #compile" at the end of the cog RAM at dst_pc
		movd	:padr1, dst_pc			' store destination address in the mov opcode
		add	dst_pc, #1
:padr1		mov	0-0, opcode_jmp_compile
#ifdef	DEBUG
		mov	ea, #ORIGIN
:dump		movs	:padr2, ea
		add	ea, #1
:padr2		mov	hexval, 0-0
		call	#hex_l
		call	#conout_cr
		cmp	ea, dst_pc	WZ,WC
	if_b	jmp	#:dump
#endif
		' now execute what we compiled

'******************************************************************************
'
' compiled code goes here
'
ORIGIN					' here is the origin of the compiler output
		fit	$1f0

alt_regs_array	byte	0,0,0,0,0,0,0,0	' alternate B, C, D, E, H, L, F and A registers
IX_reg		word	0
IY_reg		word	0

'******************************************************************************
'
' Compute an 8 bit increment and set the flags
'
alu_inc		long	(alu_inc_end - $ - 1)
		add	alu, #1				' increment alu
		andn	F, #NF				' reset negative flag
		and	alu, #$FF	WZ
		muxz	F, #ZF
		test	alu, #$80	WC		' bit #7 set?
		muxc	F, #SF				' S flag used to indicate sign
		cmp	alu, #$80	WZ		' increment 7f -> 80 ?
		muxz	F, #PF				' P flag used to indicate overflow
		mov	aux, alu
		and	aux, #15
		cmp	aux, #0	WZ			' lower nibble wrapped to 0?
		muxz	F, #HF				' set half carry
alu_inc_end


'******************************************************************************
'
' Compute an 8 bit decrement and set the flags
'
alu_dec		long	(alu_dec_end - $ - 1)
		sub	alu, #1				' decrement alu
		or	F, #NF				' set negative flag
		and	alu, #$ff	WZ
		muxz	F, #ZF
		test	alu, #$80	WC		' bit #7 set?
		muxc	F, #SF				' S flag used to indicate sign
		cmp	alu, #$7f	WZ		' decrement 80 -> 7f ?
		muxz	F, #PF				' P flag used to indicate overflow
		mov	aux, alu
		and	aux, #15
		cmp	aux, #15	WZ		' lower nibble wrapped to 15?
		muxz	F, #HF				' set half carry
alu_dec_end

'******************************************************************************
'
' Compute an arithmetic or logic alu operation and set the flags
'
alu_add		long	(alu_add_end - $ - 1)
		mov	F, #0				' clear all flags
		mov	aux, A
		add	A, tmp				' add the tmp value
		xor	aux, tmp
		xor	aux, A
		test	aux, #HF	WC		' H flag change?
		muxc	F, #HF
		test	A, #$100	WC
		muxc	F, #(PF|CF)
		test	A, #$80	WC
	if_c	xor	F, #PF
		and	A, #$ff	WZ
		muxz	F, #ZF
		test	A, #$80	WZ
		muxnz	F, #SF
alu_add_end

alu_adc		long	(alu_adc_end - $ - 1)
		test	F, #CF	WC			' get the C flag into carry
		mov	F, #0				' clear all flags
		mov	aux, A
		addx	A, tmp				' add the tmp value with carry
		xor	aux, tmp
		xor	aux, A
		and	aux, #HF			' H flag change?
		or	F, aux
		test	A, #$100	WC
		muxc	F, #(PF|CF)
		test	A, #$80	WC
	if_c	xor	F, #PF
		and	A, #$ff	WZ
		muxz	F, #ZF
		test	A, #$80	WZ
		muxnz	F, #SF
alu_adc_end

alu_sub		long	(alu_sub_end - $ - 1)
		mov	F, #NF				' clear all flags but the N flag
		mov	aux, A
		sub	A, tmp				' subtract the tmp value
		xor	aux, tmp
		xor	aux, A
		and	aux, #HF			' H flag change?
		or	F, aux
		test	A, #$100	WC
		muxc	F, #(PF|CF)
		test	A, #$80	WC
	if_c	xor	F, #PF
		and	A, #$ff	WZ
		muxz	F, #ZF
		test	A, #$80	WZ
		muxnz	F, #SF
alu_sub_end

alu_sbc		long	(alu_sbc_end - $ - 1)
		test	F, #CF	WC			' get the C flag into carry
		mov	F, #NF				' clear all flags but the N flag
		mov	aux, A
		subx	A, tmp				' subtract the tmp value
		xor	aux, tmp
		xor	aux, A
		and	aux, #HF			' H flag change?
		or	F, aux
		test	A, #$100	WC
		muxc	F, #(PF|CF)
		test	A, #$80	WC
	if_c	xor	F, #PF
		and	A, #$ff	WZ
		muxz	F, #ZF
		test	A, #$80	WZ
		muxnz	F, #SF
alu_sbc_end

alu_and		long	(alu_and_end - $ - 1)
		mov	F, #0				' clear flags
		and	A, tmp		WC,WZ		' and with the tmp value
		muxz	F, #ZF				' if zero set Z flag
		muxnc	F, #PF				' if even parity set the P flag
		test	A, #$80	WC			' get sign in prop's carry flag
		muxc	F, #SF				' set S flag
		or	F, #HF				' set H flag
alu_and_end

alu_xor		long	(alu_xor_end - $ - 1)
		mov	F, #0				' clear flags
		xor	A, tmp		WC,WZ		' xor with the tmp value
		muxz	F, #ZF				' if zero set Z flag
		muxnc	F, #PF				' if even parity set the P flag
		test	A, #$80	WC			' get sign in prop's carry flag
		muxc	F, #SF				' set S flag
alu_xor_end

alu_or		long	(alu_or_end - $ - 1)
		mov	F, #0				' clear flags
		or	A, tmp		WC,WZ		' or with the tmp value
		muxz	F, #ZF				' if zero set Z flag
		muxnc	F, #PF				' if even parity set the P flag
		test	A, #$80	WC			' get sign in prop's carry flag
		muxc	F, #SF				' set S flag
alu_or_end

alu_cp		long	(alu_cp_end - $ - 1)
		mov	alu, A
		mov	F, #NF				' clear all flags but the N flag
		mov	aux, alu
		sub	alu, tmp			' subtract the tmp value
		xor	aux, tmp
		xor	aux, alu
		and	aux, #HF			' H flag change?
		or	F, aux
		test	alu, #$100	WC
		muxc	F, #(PF|CF)
		test	alu, #$80	WC
	if_c	xor	F, #PF
		and	alu, #$ff	WZ
		muxz	F, #ZF
		test	alu, #$80	WZ
		muxnz	F, #SF
alu_cp_end

'*******************************************************
' Add a 16 bit register to HL
'
add16_hl	long	(add16_hl_end - $ - 1)
		mov	alu, H
		shl	alu, #8
		or	alu, L
		and	F, #(SF | ZF | PF)
		mov	aux, alu
		add	alu, tmp
		shl	alu, #16	WC		' get bit #16 into prop's carry
		muxc	F, #CF				' clear or set carry flag
		shr	alu, #16			' to bits 15..0
		xor	aux, alu
		xor	aux, tmp
		shr	aux, #8				' get MSB to LSB
		and	aux, #HF	WC
		or	F, #HF
		mov	L, alu
		and	L, #$ff
		mov	H, alu
		shr	H, #8
add16_hl_end

'*******************************************************
' Add a 16 bit register to IX or IY
'
add16_xy	long	(add16_xy_end - $ - 1)
		rdword	alu, XY				' fetch IX or IY
		and	F, #(SF | ZF | PF)
		mov	aux, alu
		add	alu, tmp
		shl	alu, #16	WC		' get bit #16 into prop's carry
		muxc	F, #CF				' clear or set carry flag
		shr	alu, #16			' to bits 15..0
		xor	aux, alu
		xor	aux, tmp
		shr	aux, #8				' get MSB to LSB
		and	aux, #HF	WC
		or	F, #HF
		wrword	alu, XY				' store IX or IY
add16_xy_end

'******************************************************************************
'
' Compute rotate left circular and the flags
'
'    alu: [7][6][5][4][3][2][1][0]  carry: [0]
' before   h  g  f  e  d  c  b  a           ?
' after    g  f  e  d  c  b  a  h           h
'
alu_rlc		long	(alu_rlc_end - $ - 1)
		mov	F, #0				' clear all flags
		shl	alu, #1
		test	alu, #$100	WZ		' get bit #7 into zero flag
		muxnz	alu, #$01			' get carry in bit #0
		muxnz	F, #CF				' set C flag from carry
		and	alu, #$ff	WZ,WC		' get zero and parity flags
		muxz	F, #ZF				' set Z on zero
		muxnc	F, #PF				' clear P on parity
		test	alu, #$80	WZ		' get sign flag
		muxnz	F, #SF				' set S on sign
alu_rlc_end

'******************************************************************************
'
' Compute rotate right circular and the flags
'
'    alu: [7][6][5][4][3][2][1][0]  carry: [0]
' before   h  g  f  e  d  c  b  a           ?
' after    a  h  g  f  e  d  c  b           a
'
alu_rrc		long	(alu_rrc_end - $ - 1)
		mov	F, #0				' clear all flags
		shr	alu, #1	WC
		muxc	alu, #$80			' set bit #7 from carry
		muxc	F, #CF				' set C flag
		and	alu, #$ff	WZ,WC		' get zero and parity flags
		muxz	F, #ZF				' set Z on zero
		muxnc	F, #PF				' clear P on parity
		test	alu, #$80	WZ		' get sign flag
		muxnz	F, #SF				' set S on sign
alu_rrc_end

'******************************************************************************
'
' Compute rotate left through carry and the flags
'
'    alu: [7][6][5][4][3][2][1][0]  carry: [0]
' before   i  h  g  f  e  d  c  b           a
' after    h  g  f  e  d  c  b  a           i
'
alu_rl		long	(alu_rl_end - $ - 1)
		shr	F, #1	WC			' get current C flag in carry
		mov	F, #0				' clear all flags
		rcl	alu, #1				' alu in bits 8..1, carry in bit 0
		test	alu, #$100	WZ		' test new carry
		muxnz	F, #CF				' set C flag
		and	alu, #$ff	WZ,WC		' get zero and parity flags
		muxz	F, #ZF				' set Z on zero
		muxnc	F, #PF				' clear P on parity
		test	alu, #$80	WZ		' get sign flag
		muxnz	F, #SF				' set S on sign
alu_rl_end

'******************************************************************************
'
' Compute rotate right through carry and the flags
'
'    alu: [7][6][5][4][3][2][1][0]  carry: [0]
' before   i  h  g  f  e  d  c  b           a
' after    a  i  h  g  f  e  d  c           b
'
alu_rr		long	(alu_rr_end - $ - 1)
		shr	F, #1	WC			' get current C flag in carry
		mov	F, #0				' clear all flags
		test	alu, #$01	WZ		' get bit 0 in zero flag
		shr	alu, #1				' alu in bits 31 and 6..0
		muxc	alu, #$80			' set bit #7 from current carry
		muxnz	F, #CF				' set C from bit #0
		and	alu, #$ff	WZ,WC		' get zero and parity flags
		muxz	F, #ZF				' set Z on zero
		muxnc	F, #PF				' clear P on parity
		test	alu, #$80	WZ		' get sign flag
		muxnz	F, #SF				' set S on sign
alu_rr_end

'******************************************************************************
'
' Compute shift left arithmetic and the flags
'
'    alu: [7][6][5][4][3][2][1][0]  carry: [0]
' before   h  g  f  e  d  c  b  a           ?
' after    g  f  e  d  c  b  a  0           h
'
alu_sla		long	(alu_sla_end - $ - 1)
		mov	F, #0				' clear all flags
		shl	alu, #25	WC		' alu in bits 31..25, bit #7 in carry
		muxc	F, #CF				' set C flag from bit #7
		shr	alu, #24			' result in 7..0
		and	alu, #$ff	WZ,WC		' get zero and parity flags
		muxz	F, #ZF				' set Z on zero
		muxnc	F, #PF				' clear P on parity
		test	alu, #$80	WZ		' get sign flag
		muxnz	F, #SF				' set S on sign
alu_sla_end

'******************************************************************************
'
' Compute shift right arithmetic and the flags
'
'    alu: [7][6][5][4][3][2][1][0]  carry: [0]
' before   h  g  f  e  d  c  b  a           ?
' after    h  h  g  f  e  d  c  b           a
'
alu_sra		long	(alu_sra_end - $ - 1)
		mov	F, #0				' clear all flags
		shl	alu, #24			' alu in bits 31..24
		sar	alu, #25	WC		' alu in bits 6..0, bit #0 in carry
		muxc	F, #CF				' set C flag
		and	alu, #$ff	WZ,WC		' get zero and parity flags
		muxz	F, #ZF				' set Z on zero
		muxnc	F, #PF				' clear P on parity
		test	alu, #$80	WZ		' get sign flag
		muxnz	F, #SF				' set S on sign
alu_sra_end

'******************************************************************************
'
' Compute shift left inverted and the flags
'
'    alu: [7][6][5][4][3][2][1][0]  carry: [0]
' before   h  g  f  e  d  c  b  a           ?
' after    g  f  e  d  c  b  a  1           h
'
alu_sli		long	(alu_sli_end - $ - 1)
		mov	F, #0				' clear all flags
		shl	alu, #25	WC		' alu in bits 31..25
		muxc	F, #CF				' set C flag
		shr	alu, #24			' result in 7..0
		or	alu, #$01			' set bit #0
		and	alu, #$ff	WZ,WC		' get zero and parity flags
		muxz	F, #ZF				' set Z on zero
		muxnc	F, #PF				' clear P on parity
		test	alu, #$80	WZ		' get sign flag
		muxnz	F, #SF				' set S on sign
alu_sli_end

'******************************************************************************
'
' Compute shift right logical and the flags
'
'    alu: [7][6][5][4][3][2][1][0]  carry: [0]
' before   h  g  f  e  d  c  b  a           ?
' after    0  h  g  f  e  d  c  b           a
'
alu_srl		long	(alu_srl_end - $ - 1)
		mov	F, #0				' clear all flags
		shr	alu, #1	WC			' alu in bits 6..0
		muxc	F, #CF				' set C flag
		and	alu, #$ff	WZ,WC		' get zero and parity flags
		muxz	F, #ZF				' set Z on zero
		muxnc	F, #PF				' clear P on parity
		test	alu, #$80	WZ		' get sign flag
		muxnz	F, #SF				' set S on sign
alu_srl_end

'******************************************************************************
'
' Compute bit test flags
'
alu_bit		long	(alu_bit_end - $ - 1)
		and	alu, tmp	WZ		' mask bit
		muxz	F, #ZF				' set Z flag from ~bit
		test	alu, #$80	WC		' test sign bit
		muxc	F, #SF				' set S flag depending on carry
		andn	F, #NF				' clear N flag
		or	F, #HF				' set H flag
alu_bit_end

'*******************************************************
' Build the address IX or IY + disp8 in ea
'
mem_xy		long	(mem_xy_end - $ - 1)
		rdword	tmp, XY				' fetch IX or IY
		shl	ea, #24
		sar	ea, #24
		add	ea, tmp
mem_xy_end

'*******************************************************
' 00 No operation
'
z_nop		long	(:end - $ - 1)
:end

'*******************************************************
' 08 Exchange AF and AF'
'
z_ex_af_af2	long	(:end - $ - 1)
		mov	ea, alt_regs
		add	ea, #6
		rdbyte	tmp, ea
		wrbyte	F, ea
		add	ea, #1
		rdbyte	tmp, ea
		mov	F, tmp
		wrbyte	A, ea
		mov	A, tmp
:end

'*******************************************************
' 10 Decrement B and jump if non zero
'
z_djnz		long	(:end - $ - 1) + (patch_disp8 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with address containing Z80 PC of the target
		mov	ea, #0-0		' source may be patched with the compiled address
		sub	B, #1	WZ		' B--
	if_nz	tjnz	ea, ea			' NZ then ea contains the cog RAM address for the rel jump
	if_nz	jmp	#handle_jp
:end

'*******************************************************
' 18 Jump relative
'
z_jr		long	(:end - $ - 1) + (patch_disp8 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with address containing Z80 PC of the target
		mov	ea, #0-0		' source may be patched with the compiled address
		tjnz	ea, ea			' NZ then ea contains the cog RAM address for the rel jump
		jmp	#handle_jp
:end

'*******************************************************
' 20 Jump relative if Z flag is clear
'
z_jr_nz		long	(:end - $ - 1) + (patch_disp8 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with address containing Z80 PC of the target
		mov	ea, #0-0		' source may be patched with the compiled address
		test	F, #ZF	WC		' test if Z flag is set
	if_nc	tjnz	ea, ea			' NZ then ea contains the cog RAM address for the rel jump
	if_nc	jmp	#handle_jp
:end

'*******************************************************
' 28 Jump relative if Z flag is set
'
z_jr_z		long	(:end - $ - 1) + (patch_disp8 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with address containing Z80 PC of the target
		mov	ea, #0-0		' source may be patched with the compiled address
		test	F, #ZF	WC		' test if Z flag is clear
	if_c	tjnz	ea, ea			' NZ then ea contains the cog RAM address for the rel jump
	if_c	jmp	#handle_jp
:end

'*******************************************************
' 30 Jump relative if C flag is clear
'
z_jr_nc		long	(:end - $ - 1) + (patch_disp8 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with address containing Z80 PC of the target
		mov	ea, #0-0		' source may be patched with the compiled address
		test	F, #CF	WC		' test if C flag is clear
	if_nc	tjnz	ea, ea			' NZ then ea contains the cog RAM address for the rel jump
	if_nc	jmp	#handle_jp
:end

'*******************************************************
' 38 Jump relative if C flag is set
'
z_jr_c		long	(:end - $ - 1) + (patch_disp8 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with address containing Z80 PC of the target
		mov	ea, #$0	WZ		' source may be patched with the compiled address
		test	F, #CF	WC		' test if C flag is set
	if_c	tjnz	ea, ea			' NZ then ea contains the cog RAM address for the rel jump
	if_c	jmp	#handle_jp
:end

'*******************************************************
' 01 Load BC with imm16
'
z_ld_bc_imm16	long	(:end - $ - 1)	| (patch_imm16 << PATCH_SHIFT)
		mov	C, 0-0			' source is patched with the address of imm16
		mov	B, C
		and	C, #$ff
		shr	B, #8
:end

'*******************************************************
' 11 Load DE with imm16
'
z_ld_de_imm16	long	(:end - $ - 1)	| (patch_imm16 << PATCH_SHIFT)
		mov	E, 0-0			' source is patched with the address of imm16
		mov	D, E
		and	E, #$ff
		shr	D, #8
:end

'*******************************************************
' 21 Load HL with imm16
'
z_ld_hl_imm16	long	(:end - $ - 1)	| (patch_imm16 << PATCH_SHIFT)
		mov	L, 0-0			' source is patched with the address of imm16
		mov	H, L
		and	L, #$ff
		shr	H, #8
:end

'*******************************************************
' 31 Load SP with imm16
'
z_ld_sp_imm16	long	(:end - $ - 1)	| (patch_imm16 << PATCH_SHIFT)
		mov	SP, 0-0			' source is patched with the address of imm16
:end

'*******************************************************
' 09 Add BC to HL
'
z_add_hl_bc	long	(:end - $ - 1) + (add16_hl_end - add16_hl - 1) - 1
		mov	tmp, B
		shl	tmp, #8
		or	tmp, C
		long	@@@add16_hl	
:end

'*******************************************************
' 19 Add DE to HL
'
z_add_hl_de	long	(:end - $ - 1) + (add16_hl_end - add16_hl - 1) - 1
		mov	tmp, D
		shl	tmp, #8
		or	tmp, E
		long	@@@add16_hl	
:end

'*******************************************************
' 29 Add HL to HL
'
z_add_hl_hl	long	(:end - $ - 1) + (add16_hl_end - add16_hl - 1) - 1
		mov	tmp, H
		shl	tmp, #8
		or	tmp, L
		long	@@@add16_hl	
:end

'*******************************************************
' 39 Add SP to HL
'
z_add_hl_sp	long	(:end - $ - 1) + (add16_hl_end - add16_hl - 1) - 1
		mov	tmp, SP
		long	@@@add16_hl	
:end

'*******************************************************
' 02 Load memory (BC) with accu
'
z_ld_mem_bc_a	long	(:end - $ - 1)
		mov	ea, B
		shl	ea, #8
		or	ea, C
		mov	alu, A
		call	#wr_byte
:end

'*******************************************************
' 0a Load accu from memory (BC)
'
z_ld_a_mem_bc	long	(:end - $ - 1)
		mov	ea, B
		shl	ea, #8
		or	ea, C
		call	#rd_byte
		mov	A, alu
:end

'*******************************************************
' 12 Load memory (DE) with accu
'
z_ld_mem_de_a	long	(:end - $ - 1)
		mov	alu, A
		mov	ea, D
		shl	ea, #8
		or	ea, E
		call	#wr_byte
:end

'*******************************************************
' 1a Load accu from memory (DE)
'
z_ld_a_mem_de	long	(:end - $ - 1)
		mov	ea, D
		shl	ea, #8
		or	ea, E
		call	#rd_byte
		mov	A, alu
:end

'*******************************************************
' 22 Load memory (imm16) with HL
'
z_ld_abs16_hl	long	(:end - $ - 1) + (patch_imm16 << PATCH_SHIFT)
		mov	ea, 0-0			' source is patched with the address of imm16
		mov	alu, H
		shl	alu, #8
		or	alu, L
		call	#wr_word
:end

'*******************************************************
' 2a Load HL from memory (imm16)
'
z_ld_hl_abs16	long	(:end - $ - 1) + (patch_imm16 << PATCH_SHIFT)
		mov	ea, 0-0			' source is patched with the address of imm16
		call	#rd_word
		mov	L, alu
		and	L, #$ff
		mov	H, alu
		shr	H, #8
:end

'*******************************************************
' 32 Load memory (imm16) with A
'
z_ld_mem_a	long	(:end - $ - 1) + (patch_imm16 << PATCH_SHIFT)
		mov	ea, 0-0			' source is patched with the address of imm16
		mov	alu, A
		call	#wr_byte
:end

'*******************************************************
' 3a Load A from memory (imm16)
'
z_ld_a_mem	long	(:end - $ - 1) + (patch_imm16 << PATCH_SHIFT)
		mov	ea, 0-0			' source is patched with the address of imm16
		call	#rd_byte
		mov	A, alu
:end

'*******************************************************
' 03 Increment BC
'
z_inc_bc	long	(:end - $ - 1)
		add	C, #1
		test	C, #$100	WC
		and	C, #$ff
		addx	B, #0
		and	B, #$ff
:end

'*******************************************************
' 0b Decrement BC
'
z_dec_bc	long	(:end - $ - 1)
		sub	C, #1	WC
		and	C, #$ff
		subx	B, #0
		and	B, #$ff
:end

'*******************************************************
' 13 Increment DE
'
z_inc_de	long	(:end - $ - 1)
		add	E, #1
		test	E, #$100	WC
		and	E, #$ff
		addx	D, #0
		and	D, #$ff
:end

'*******************************************************
' 1b Decrement DE
'
z_dec_de	long	(:end - $ - 1)
		sub	E, #1	WC
		and	E, #$ff
		subx	D, #0
		and	D, #$ff
:end

'*******************************************************
' 23 Increment HL
'
z_inc_hl	long	(:end - $ - 1)
		add	L, #1
		test	L, #$100	WC
		and	L, #$ff
		addx	H, #0
		and	H, #$ff
:end

'*******************************************************
' 2b Decrement HL
'
z_dec_hl	long	(:end - $ - 1)
		sub	L, #1	WC
		and	L, #$ff
		subx	H, #0
		and	H, #$ff
:end

'*******************************************************
' 33 Increment SP
'
z_inc_sp	long	(:end - $ - 1)
		add	SP, #1
		and	SP, low_16bits
:end

'*******************************************************
' 3b Decrement SP
'
z_dec_sp	long	(:end - $ - 1)
		sub	SP, #1
		and	SP, low_16bits
:end

'*******************************************************
' 04 Increment B
'
z_inc_b		long	(:end - $ - 1) + (alu_inc_end - alu_inc - 1) - 1
		mov	alu, B
		long	@@@alu_inc
		mov	B, alu
:end

'*******************************************************
' 0c Increment C
'
z_inc_c		long	(:end - $ - 1) + (alu_inc_end - alu_inc - 1) - 1
		mov	alu, C
		long	@@@alu_inc
		mov	C, alu
:end

'*******************************************************
' 14 Increment D
'
z_inc_d		long	(:end - $ - 1) + (alu_inc_end - alu_inc - 1) - 1
		mov	alu, D
		long	@@@alu_inc
		mov	D, alu
:end


'*******************************************************
' 1c Increment E
'
z_inc_e		long	(:end - $ - 1) + (alu_inc_end - alu_inc - 1) - 1
		mov	alu, E
		long	@@@alu_inc
		mov	E, alu
:end

'*******************************************************
' 24 Increment H
'
z_inc_h		long	(:end - $ - 1) + (alu_inc_end - alu_inc - 1) - 1
		mov	alu, H
		long	@@@alu_inc
		mov	H, alu
:end


'*******************************************************
' 2c Increment L
'
z_inc_l		long	(:end - $ - 1) + (alu_inc_end - alu_inc - 1) - 1
		mov	alu, L
		long	@@@alu_inc
		mov	L, alu
:end

'*******************************************************
' 34 Increment memory (HL)
'
z_inc_m		long	(:end - $ - 1) + (alu_inc_end - alu_inc - 1) - 1
		call	#rd_byte_hl
		long	@@@alu_inc
		call	#wr_byte
:end

'*******************************************************
' 3c Increment A
'
z_inc_a		long	(:end - $ - 1) + (alu_inc_end - alu_inc - 1) - 1
		mov	alu, A
		long	@@@alu_inc
		mov	A, alu
:end

'*******************************************************
' 05 Decrement B
'
z_dec_b		long	(:end - $ - 1) + (alu_dec_end - alu_dec - 1) - 1
		mov	alu, B
		long	@@@alu_dec
		mov	B, alu
:end

'*******************************************************
' 0d Decrement C
'
z_dec_c		long	(:end - $ - 1) + (alu_dec_end - alu_dec - 1) - 1
		mov	alu, C
		long	@@@alu_dec
		mov	C, alu
:end

'*******************************************************
' 15 Decrement D
'
z_dec_d		long	(:end - $ - 1) + (alu_dec_end - alu_dec - 1) - 1
		mov	alu, D
		long	@@@alu_dec
		mov	D, alu
:end

'*******************************************************
' 1d Decrement E
'
z_dec_e		long	(:end - $ - 1) + (alu_dec_end - alu_dec - 1) - 1
		mov	alu, E
		long	@@@alu_dec
		mov	E, alu
:end

'*******************************************************
' 25 Decrement H
'
z_dec_h		long	(:end - $ - 1) + (alu_dec_end - alu_dec - 1) - 1
		mov	alu, H
		long	@@@alu_dec
		mov	H, alu
:end

'*******************************************************
' 2d Decrement L
'
z_dec_l		long	(:end - $ - 1) + (alu_dec_end - alu_dec - 1) - 1
		mov	alu, L
		long	@@@alu_dec
		mov	L, alu
:end

'*******************************************************
' 35 Decrement memory (HL)
'
z_dec_m		long	(:end - $ - 1) + (alu_dec_end - alu_dec - 1) - 1
		call	#rd_byte_hl
		long	@@@alu_dec
		call	#wr_byte
:end

'*******************************************************
' 3d Decrement A
'
z_dec_a		long	(:end - $ - 1) + (alu_dec_end - alu_dec - 1) - 1
		mov	alu, A
		long	@@@alu_dec
		mov	A, alu
:end

'*******************************************************
' 06 Load B with imm8
'
z_ld_b_imm8	long	(:end - $ - 1) + (patch_imm8 << PATCH_SHIFT)
		mov	B, #0			' source is patched with the imm8
:end

'*******************************************************
' 0e Load c with imm8
'
z_ld_c_imm8	long	(:end - $ - 1) + (patch_imm8 << PATCH_SHIFT)
		mov	C, #0			' source is patched with the imm8
:end

'*******************************************************
' 16 Load D with imm8
'
z_ld_d_imm8	long	(:end - $ - 1) + (patch_imm8 << PATCH_SHIFT)
		mov	D, #0			' source is patched with the imm8
:end

'*******************************************************
' 1e Load E with imm8
'
z_ld_e_imm8	long	(:end - $ - 1) + (patch_imm8 << PATCH_SHIFT)
		mov	E, #0			' source is patched with the imm8
:end

'*******************************************************
' 26 Load H with imm8
'
z_ld_h_imm8	long	(:end - $ - 1) + (patch_imm8 << PATCH_SHIFT)
		mov	H, #0			' source is patched with the imm8
:end

'*******************************************************
' 2e Load L with imm8
'
z_ld_l_imm8	long	(:end - $ - 1) + (patch_imm8 << PATCH_SHIFT)
		mov	L, #0			' source is patched with the imm8
:end

'*******************************************************
' 36 Load memory (HL) with imm8
'
z_ld_m_imm8	long	(:end - $ - 1) + (patch_imm8 << PATCH_SHIFT)
		mov	alu, #0			' source is patched with the imm8
		mov	ea, H
		shl	ea, #8
		or	ea, L
		call	#wr_byte
:end

'*******************************************************
' 3e Load A with imm8
'
z_ld_a_imm8	long	(:end - $ - 1) + (patch_imm8 << PATCH_SHIFT)
		mov	A, #0			' source is patched with the imm8
:end

'*******************************************************
' 07 Rotate left circular accu
'
z_rlca		long	(:end - $ - 1)
		shl	A, #1
		test	A, #$100	WC	' check carry
		and	A, #$ff			' keep bits 7..0
		muxc	A, #$01			' set low bit from carry
		and	F, #(SF | ZF | PF)	' keep S, Z and P flags
#ifdef	XYFLAGS
		mov	aux, A
		and	aux, #(YF | XF | CF)
		or	F, aux			' set Y, X and C flags
#endif
:end


'*******************************************************
' 0f Rotate right circular accu
'
z_rrca		long	(:end - $ - 1)
		shr	A, #1	WC		' rotate right and check carry
		muxc	A, #$80			' set bit #7 from carry flag
		and	F, #(SF | ZF | PF)	' keep S, Z and P flags
		muxc	F, #CF			' set C from carry flag
#ifdef	XYFLAGS
		mov	aux, A
		and	aux, #(YF | XF)
		or	F, aux			' set Y and X flags
#endif
:end

'*******************************************************
' 17 Rotate left accu
'
z_rla		long	(:end - $ - 1)
		shl	A, #1
		test	A, #$100	WZ	' test bit #7
		and	A, #$ff			' keep bits 7..0
		test	F, #CF	WC		' get C flag to carry
		muxnz	F, #CF			' set C from zero flag
		muxc	A, #$01			' set bit #0 from previous C flag
		and	F, #(SF | ZF | PF | CF)	' keep S, Z, P and C flags
#ifdef	XYFLAGS
		mov	aux, A
		and	aux, #(YF | XF)
		or	F, aux			' set Y and X flags
#endif
:end

'*******************************************************
' 1f Rotate right accu
'
z_rra		long	(:end - $ - 1)
		shr	A, #1	WC		' rotate right and check carry
		test	F, #CF	WZ		' get C flag
		muxc	F, #CF			' set C flag from carry flag
		muxnz	A, #$80			' set bit #7 from zero flag
		and	F, #(SF | ZF | PF | CF)	' keep S, Z, P and C flags
#ifdef	XYFLAGS
		mov	aux, A
		and	aux, #(YF | XF)
		or	F, aux			' set Y and X flags
#endif
:end

'*******************************************************
' 27 Decimal adjust accumulator
'
z_daa		long	(:end - $ - 1)
		mov	tmp, A
		mov	aux, A			' save for auxiliary flag test
		and	tmp, #15		' isolate low nibble of accumulator
		mov	daa_adj, #$06		' +$06
		test	F, #NF	WZ		' test negative flag
		negnz	daa_adj, daa_adj	' -$06
		test	F, #HF	WZ
		cmp	tmp, #$0a	WC	' aux isn't set; is nibble > $9 ?
  if_nc_or_nz	add	A, daa_adj		' aux was set or nibble is > $9 => add $06 or -$06
		shl	daa_adj, #4		' +$60 or -$60
		test	F, #CF	WZ
		cmp	A, #$9a	WC		' carry isn't set; is data > $99 ?
  if_nc_or_nz	add	A, daa_adj		' carry was set or data is > $99 => add $60 or -$60
		and	F, #(NF|CF)		' preserve old CF and NF
		cmp	A, #$9a	WC
	if_nc	or	F, #CF			' set carry if result > $99
		xor	aux, A
		and	aux, #HF		' if bit #4 of result changed
		or	F, aux			' set the auxiliary carry flag
		and	A, #$ff	WZ,WC		' get zero and parity flags
		muxz	F, #ZF			' set Z on zero
		muxnc	F, #PF			' clear P on parity
		test	A, #$80	WZ		' get sign flag
		muxnz	F, #SF			' set S on sign
:end

'*******************************************************
' 2f Complement A
'
z_cpl		long	(:end - $ - 1)
		xor	A, #$ff
		andn	F, #(XF|YF)		' clear the X and Y flags
		or	F, #(HF|NF)		' set the H and N flags
#ifdef	XYFLAGS
		mov	aux, A
		and	aux, #(XF|YF)
		or	F, aux			' set the X and Y flags
#endif
:end

'*******************************************************
' 37 Set carry flag
'
z_scf		long	(:end - $ - 1)
		or	F, #CF			' set the C flag
#ifdef	XYFLAGS
		mov	aux, A
		and	aux, #(XF|YF)
		or	F, aux			' or in the X and Y flags (sic!)
#endif
:end

'*******************************************************
' 3f Complement carry flag
'
z_ccf		long	(:end - $ - 1)
		mov	aux, A
		test	F, #1	WC		' get the C flag into the carry
	if_c	muxc	F, #HF			' set the H flag if C was set
		xor	F, #CF			' toggle the C flag
#ifdef	XYFLAGS
		and	aux, #(XF|YF)
		or	F, aux			' or in the X and Y flags (sic!)
#endif
:end


'*******************************************************
' 40 Load B with B (NOP)
'
'z_ld_b_b	long	(:end - $ - 1)
':end

'*******************************************************
' 41 Load B with C
'
z_ld_b_c	long	(:end - $ - 1)
		mov	B, C
:end

'*******************************************************
' 42 Load B with D
'
z_ld_b_d	long	(:end - $ - 1)
		mov	B, D
:end

'*******************************************************
' 43 Load B with E
'
z_ld_b_e	long	(:end - $ - 1)
		mov	B, E
:end

'*******************************************************
' 44 Load B with H
'
z_ld_b_h	long	(:end - $ - 1)
		mov	B, H
:end

'*******************************************************
' 45 Load B with L
'
z_ld_b_l	long	(:end - $ - 1)
		mov	B, L
:end

'*******************************************************
' 46 Load B with memory (HL)
'
z_ld_b_m	long	(:end - $ - 1)
		call	#rd_byte_hl
		mov	B, alu
:end

'*******************************************************
' 47 Load B with A
'
z_ld_b_a	long	(:end - $ - 1)
		mov	B, A
:end

'*******************************************************
' 48 Load C with B
'
z_ld_c_b	long	(:end - $ - 1)
		mov	C, B
:end

'*******************************************************
' 49 Load C with C (NOP)
'
'z_ld_c_c	long	(:end - $ - 1)
':end

'*******************************************************
' 4a Load C with D
'
z_ld_c_d	long	(:end - $ - 1)
		mov	C, D
:end

'*******************************************************
' 4b Load C with E
'
z_ld_c_e	long	(:end - $ - 1)
		mov	C, E
:end

'*******************************************************
' 4c Load B with H
'
z_ld_c_h	long	(:end - $ - 1)
		mov	C, H
:end

'*******************************************************
' 4d Load C with L
'
z_ld_c_l	long	(:end - $ - 1)
		mov	C, L
:end

'*******************************************************
' 4e Load C with memory (HL)
'
z_ld_c_m	long	(:end - $ - 1)
		call	#rd_byte_hl
		mov	C, alu
:end

'*******************************************************
' 4f Load C with A
'
z_ld_c_a	long	(:end - $ - 1)
		mov	C, A
:end

'*******************************************************
' 50 Load D with B
'
z_ld_d_b	long	(:end - $ - 1)
		mov	D, B
:end

'*******************************************************
' 51 Load D with C
'
z_ld_d_c	long	(:end - $ - 1)
		mov	D, C
:end

'*******************************************************
' 52 Load D with D (NOP)
'
'z_ld_d_d	long	(:end - $ - 1)
':end

'*******************************************************
' 53 Load D with E
'
z_ld_d_e	long	(:end - $ - 1)
		mov	D, E
:end

'*******************************************************
' 54 Load D with H
'
z_ld_d_h	long	(:end - $ - 1)
		mov	D, H
:end

'*******************************************************
' 55 Load D with L
'
z_ld_d_l	long	(:end - $ - 1)
		mov	D, L
:end

'*******************************************************
' 56 Load D with memory (HL)
'
z_ld_d_m	long	(:end - $ - 1)
		call	#rd_byte_hl
		mov	D, alu
:end

'*******************************************************
' 57 Load D with A
'
z_ld_d_a	long	(:end - $ - 1)
		mov	D, A
:end

'*******************************************************
' 58 Load E with B
'
z_ld_e_b	long	(:end - $ - 1)
		mov	E, B
:end

'*******************************************************
' 59 Load E with C
'
z_ld_e_c	long	(:end - $ - 1)
		mov	E, C
:end

'*******************************************************
' 5a Load E with D
'
z_ld_e_d	long	(:end - $ - 1)
		mov	E, D
:end

'*******************************************************
' 5b Load E with E (NOP)
'
'z_ld_e_e	long	(:end - $ - 1)
':end

'*******************************************************
' 5c Load E with H
'
z_ld_e_h	long	(:end - $ - 1)
		mov	E, H
:end

'*******************************************************
' 5d Load E with L
'
z_ld_e_l	long	(:end - $ - 1)
		mov	E, L
:end

'*******************************************************
' 5e Load E with memory (HL)
'
z_ld_e_m	long	(:end - $ - 1)
		call	#rd_byte_hl
		mov	E, alu
:end

'*******************************************************
' 5f Load E with A
'
z_ld_e_a	long	(:end - $ - 1)
		mov	E, A
:end

'*******************************************************
' 60 Load H with B
'
z_ld_h_b	long	(:end - $ - 1)
		mov	H, B
:end

'*******************************************************
' 61 Load H with C
'
z_ld_h_c	long	(:end - $ - 1)
		mov	H, C
:end

'*******************************************************
' 62 Load H with D
'
z_ld_h_d	long	(:end - $ - 1)
		mov	H, D
:end

'*******************************************************
' 63 Load H with E
'
z_ld_h_e	long	(:end - $ - 1)
		mov	H, E
:end

'*******************************************************
' 64 Load H with H (NOP)
'
'z_ld_h_h	long	(:end - $ - 1)
':end

'*******************************************************
' 65 Load H with L
'
z_ld_h_l	long	(:end - $ - 1)
		mov	H, L
:end

'*******************************************************
' 66 Load H with memory (HL)
'
z_ld_h_m	long	(:end - $ - 1)
		call	#rd_byte_hl
		mov	H, alu
:end

'*******************************************************
' 67 Load H with A
'
z_ld_h_a	long	(:end - $ - 1)
		mov	H, A
:end

'*******************************************************
' 68 Load L with B
'
z_ld_l_b	long	(:end - $ - 1)
		mov	L, B
:end

'*******************************************************
' 69 Load L with C
'
z_ld_l_c	long	(:end - $ - 1)
		mov	L, C
:end

'*******************************************************
' 6a Load L with D
'
z_ld_l_d	long	(:end - $ - 1)
		mov	L, D
:end

'*******************************************************
' 6b Load L with E
'
z_ld_l_e	long	(:end - $ - 1)
		mov	L, E
:end

'*******************************************************
' 6c Load L with H
'
z_ld_l_h	long	(:end - $ - 1)
		mov	L, H
:end

'*******************************************************
' 6d Load L with L (NOP)
'
'z_ld_l_l	long	(:end - $ - 1)
':end

'*******************************************************
' 6e Load L with memory (HL)
'
z_ld_l_m	long	(:end - $ - 1)
		call	#rd_byte_hl
		mov	L, alu
:end

'*******************************************************
' 6f Load L with A
'
z_ld_l_a	long	(:end - $ - 1)
		mov	L, A
:end

'*******************************************************
' 70 Load memory (HL) with B
'
z_ld_m_b	long	(:end - $ - 1)
		mov	alu, B
		mov	ea, H
		shl	ea, #8
		or	ea, L
		call	#wr_byte
:end

'*******************************************************
' 71 Load memory (HL) with C
'
z_ld_m_c	long	(:end - $ - 1)
		mov	alu, C
		mov	ea, H
		shl	ea, #8
		or	ea, L
		call	#wr_byte
:end

'*******************************************************
' 72 Load memory (HL) with D
'
z_ld_m_d	long	(:end - $ - 1)
		mov	alu, D
		mov	ea, H
		shl	ea, #8
		or	ea, L
		call	#wr_byte
:end

'*******************************************************
' 73 Load memory (HL) with E
'
z_ld_m_e	long	(:end - $ - 1)
		mov	alu, E
		mov	ea, H
		shl	ea, #8
		or	ea, L
		call	#wr_byte
:end

'*******************************************************
' 74 Load memory (HL) with H
'
z_ld_m_h	long	(:end - $ - 1)
		mov	alu, H
		mov	ea, H
		shl	ea, #8
		or	ea, L
		call	#wr_byte
:end

'*******************************************************
' 75 Load memory (HL) with L
'
z_ld_m_l	long	(:end - $ - 1)
		mov	alu, L
		mov	ea, H
		shl	ea, #8
		or	ea, L
		call	#wr_byte
:end

'*******************************************************
' 76 HALT
'
z_halt		long	(:end - $ - 1) + (execute << PATCH_SHIFT)
		mov	HALT, #1
:end

'*******************************************************
' 77 Load memory (HL) with A
'
z_ld_m_a	long	(:end - $ - 1)
		mov	alu, A
		mov	ea, H
		shl	ea, #8
		or	ea, L
		call	#wr_byte
:end

'*******************************************************
' 78 Load A with B
'
z_ld_a_b	long	(:end - $ - 1)
		mov	A, B
:end

'*******************************************************
' 79 Load A with C
'
z_ld_a_c	long	(:end - $ - 1)
		mov	A, C
:end

'*******************************************************
' 7a Load A with D
'
z_ld_a_d	long	(:end - $ - 1)
		mov	A, D
:end

'*******************************************************
' 7b Load A with E
'
z_ld_a_e	long	(:end - $ - 1)
		mov	A, E
:end

'*******************************************************
' 7c Load A with H
'
z_ld_a_h	long	(:end - $ - 1)
		mov	A, H
:end

'*******************************************************
' 7d Load A with L
'
z_ld_a_l	long	(:end - $ - 1)
		mov	A, L
:end

'*******************************************************
' 7e Load A with memory (HL)
'
z_ld_a_m	long	(:end - $ - 1)
		call	#rd_byte_hl
		mov	A, alu
:end

'*******************************************************
' 7f Load A with A (NOP)
'
'z_ld_a_a	long	(:end - $ - 1)
':end

'*******************************************************
' 80 Add B to A
'
z_add_a_b	long	(:end - $ - 1) + (alu_add_end - alu_add - 1) - 1
		mov	tmp, B
		long	@@@alu_add		' paste the alu_add code snippet here
:end

'*******************************************************
' 81 Add C to A
'
z_add_a_c	long	(:end - $ - 1) + (alu_add_end - alu_add - 1) - 1
		mov	tmp, C
		long	@@@alu_add		' paste the alu_add code snippet here
:end

'*******************************************************
' 82 Add D to A
'
z_add_a_d	long	(:end - $ - 1) + (alu_add_end - alu_add - 1) - 1
		mov	tmp, D
		long	@@@alu_add		' paste the alu_add code snippet here
:end

'*******************************************************
' 83 Add E to A
'
z_add_a_e	long	(:end - $ - 1) + (alu_add_end - alu_add - 1) - 1
		mov	tmp, E
		long	@@@alu_add		' paste the alu_add code snippet here
:end

'*******************************************************
' 84 Add H to A
'
z_add_a_h	long	(:end - $ - 1) + (alu_add_end - alu_add - 1) - 1
		mov	tmp, H
		long	@@@alu_add		' paste the alu_add code snippet here
:end

'*******************************************************
' 85 Add L to A
'
z_add_a_l	long	(:end - $ - 1) + (alu_add_end - alu_add - 1) - 1
		mov	tmp, L
		long	@@@alu_add		' paste the alu_add code snippet here
:end

'*******************************************************
' 86 Add memory (HL) to A
'
z_add_a_m	long	(:end - $ - 1) + (alu_add_end - alu_add - 1) - 1
		call	#rd_byte_hl
		mov	tmp, alu
		long	@@@alu_add		' paste the alu_add code snippet here
:end

'*******************************************************
' 87 Add A to A
'
z_add_a_a	long	(:end - $ - 1) + (alu_add_end - alu_add - 1) - 1
		mov	tmp, A
		long	@@@alu_add		' paste the alu_add code snippet here
:end

'****************************************************
' 88 Add with carry B to A
'
z_adc_a_b	long	(:end - $ - 1) + (alu_adc_end - alu_adc - 1) - 1
		mov	tmp, B
		long	@@@alu_adc		' paste the alu_adc code snippet here
:end

'*******************************************************
' 89 Add with carry C to A
'
z_adc_a_c	long	(:end - $ - 1) + (alu_adc_end - alu_adc - 1) - 1
		mov	tmp, C
		long	@@@alu_adc		' paste the alu_adc code snippet here
:end

'*******************************************************
' 8a Add with carry D to A
'
z_adc_a_d	long	(:end - $ - 1) + (alu_adc_end - alu_adc - 1) - 1
		mov	tmp, D
		long	@@@alu_adc		' paste the alu_adc code snippet here
:end

'*******************************************************
' 8b Add with carry E to A
'
z_adc_a_e	long	(:end - $ - 1) + (alu_adc_end - alu_adc - 1) - 1
		mov	tmp, E
		long	@@@alu_adc		' paste the alu_adc code snippet here
:end

'*******************************************************
' 8c Add with carry H to A
'
z_adc_a_h	long	(:end - $ - 1) + (alu_adc_end - alu_adc - 1) - 1
		mov	tmp, H
		long	@@@alu_adc		' paste the alu_adc code snippet here
:end

'*******************************************************
' 8d Add with carry L to A
'
z_adc_a_l	long	(:end - $ - 1) + (alu_adc_end - alu_adc - 1) - 1
		mov	tmp, L
		long	@@@alu_adc		' paste the alu_adc code snippet here
:end

'*******************************************************
' 8e Add with carry memory (HL) to A
'
z_adc_a_m	long	(:end - $ - 1) + (alu_adc_end - alu_adc - 1) - 1
		call	#rd_byte_hl
		mov	tmp, alu
		long	@@@alu_adc		' paste the alu_adc code snippet here
:end

'*******************************************************
' 8f Add with carry A to A
'
z_adc_a_a	long	(:end - $ - 1) + (alu_adc_end - alu_adc - 1) - 1
		mov	tmp, A
		long	@@@alu_adc		' paste the alu_adc code snippet here
:end

'*******************************************************
' 90 Subtract B from A 
'
z_sub_a_b	long	(:end - $ - 1) + (alu_sub_end - alu_sub - 1) - 1
		mov	tmp, B
		long	@@@alu_sub		' paste the alu_sub code snippet here
:end

'*******************************************************
' 91 Subtract C from A 
'
z_sub_a_c	long	(:end - $ - 1) + (alu_sub_end - alu_sub - 1) - 1
		mov	tmp, C
		long	@@@alu_sub		' paste the alu_sub code snippet here
:end

'*******************************************************
' 92 Subtract D from A 
'
z_sub_a_d	long	(:end - $ - 1) + (alu_sub_end - alu_sub - 1) - 1
		mov	tmp, D
		long	@@@alu_sub		' paste the alu_sub code snippet here
:end

'*******************************************************
' 93 Subtract E from A 
'
z_sub_a_e	long	(:end - $ - 1) + (alu_sub_end - alu_sub - 1) - 1
		mov	tmp, E
		long	@@@alu_sub		' paste the alu_sub code snippet here
:end

'*******************************************************
' 94 Subtract H from A 
'
z_sub_a_h	long	(:end - $ - 1) + (alu_sub_end - alu_sub - 1) - 1
		mov	tmp, H
		long	@@@alu_sub		' paste the alu_sub code snippet here
:end

'*******************************************************
' 95 Subtract L from A 
'
z_sub_a_l	long	(:end - $ - 1) + (alu_sub_end - alu_sub - 1) - 1
		mov	tmp, L
		long	@@@alu_sub		' paste the alu_sub code snippet here
:end

'*******************************************************
' 96 Subtract memory (HL) from A 
'
z_sub_a_m	long	(:end - $ - 1) + (alu_sub_end - alu_sub - 1) - 1
		call	#rd_byte_hl
		mov	tmp, alu
		long	@@@alu_sub		' paste the alu_sub code snippet here
:end

'*******************************************************
' 97 Subtract A from A 
'
z_sub_a_a	long	(:end - $ - 1) + (alu_sub_end - alu_sub - 1) - 1
		mov	tmp, A
		long	@@@alu_sub		' paste the alu_sub code snippet here
:end

'*******************************************************
' 98 Subtract with carry B from A 
'
z_sbc_a_b	long	(:end - $ - 1) + (alu_sbc_end - alu_sbc - 1) - 1
		mov	tmp, B
		long	@@@alu_sbc		' paste the alu_sbc code snippet here
:end

'*******************************************************
' 99 Subtract with carry C from A 
'
z_sbc_a_c	long	(:end - $ - 1) + (alu_sbc_end - alu_sbc - 1) - 1
		mov	tmp, C
		long	@@@alu_sbc		' paste the alu_sbc code snippet here
:end

'*******************************************************
' 9a Subtract with carry D from A 
'
z_sbc_a_d	long	(:end - $ - 1) + (alu_sbc_end - alu_sbc - 1) - 1
		mov	tmp, D
		long	@@@alu_sbc		' paste the alu_sbc code snippet here
:end

'*******************************************************
' 9b Subtract with carry E from A 
'
z_sbc_a_e	long	(:end - $ - 1) + (alu_sbc_end - alu_sbc - 1) - 1
		mov	tmp, E
		long	@@@alu_sbc		' paste the alu_sbc code snippet here
:end

'*******************************************************
' 9c Subtract with carry H from A 
'
z_sbc_a_h	long	(:end - $ - 1) + (alu_sbc_end - alu_sbc - 1) - 1
		mov	tmp, H
		long	@@@alu_sbc		' paste the alu_sbc code snippet here
:end

'*******************************************************
' 9d Subtract with carry L from A 
'
z_sbc_a_l	long	(:end - $ - 1) + (alu_sbc_end - alu_sbc - 1) - 1
		mov	tmp, L
		long	@@@alu_sbc		' paste the alu_sbc code snippet here
:end

'*******************************************************
' 9e Subtract with carry memory (HL) from A 
'
z_sbc_a_m	long	(:end - $ - 1) + (alu_sbc_end - alu_sbc - 1) - 1
		call	#rd_byte_hl
		mov	tmp, alu
		long	@@@alu_sbc		' paste the alu_sbc code snippet here
:end

'*******************************************************
' 9f Subtract with carry A from A 
'
z_sbc_a_a	long	(:end - $ - 1) + (alu_sbc_end - alu_sbc - 1) - 1
		mov	tmp, A
		long	@@@alu_sbc		' paste the alu_sbc code snippet here
:end

'*******************************************************
' a0 And A with B
'
z_and_a_b	long	(:end - $ - 1) + (alu_and_end - alu_and - 1) - 1
		mov	tmp, B
		long	@@@alu_and		' paste the alu_and code snippet here
:end

'*******************************************************
' a1 And A with C
'
z_and_a_c	long	(:end - $ - 1) + (alu_and_end - alu_and - 1) - 1
		mov	tmp, C
		long	@@@alu_and		' paste the alu_and code snippet here
:end

'*******************************************************
' a2 And A with D
'
z_and_a_d	long	(:end - $ - 1) + (alu_and_end - alu_and - 1) - 1
		mov	tmp, D
		long	@@@alu_and		' paste the alu_and code snippet here
:end

'*******************************************************
' a3 And A with E
'
z_and_a_e	long	(:end - $ - 1) + (alu_and_end - alu_and - 1) - 1
		mov	tmp, E
		long	@@@alu_and		' paste the alu_and code snippet here
:end

'*******************************************************
' a4 And A with H
'
z_and_a_h	long	(:end - $ - 1) + (alu_and_end - alu_and - 1) - 1
		mov	tmp, H
		long	@@@alu_and		' paste the alu_and code snippet here
:end

'*******************************************************
' a5 And A with L
'
z_and_a_l	long	(:end - $ - 1) + (alu_and_end - alu_and - 1) - 1
		mov	tmp, L
		long	@@@alu_and		' paste the alu_and code snippet here
:end

'*******************************************************
' a6 And A with memory (HL)
'
z_and_a_m	long	(:end - $ - 1) + (alu_and_end - alu_and - 1) - 1
		call	#rd_byte_hl
		mov	tmp, alu
		long	@@@alu_and		' paste the alu_and code snippet here
:end

'*******************************************************
' a7 And A with A
'
z_and_a_a	long	(:end - $ - 1) + (alu_and_end - alu_and - 1) - 1
		mov	tmp, A
		long	@@@alu_and		' paste the alu_and code snippet here
:end

'*******************************************************
' a8 Exclusive or A with B
'
z_xor_a_b	long	(:end - $ - 1) + (alu_xor_end - alu_xor - 1) - 1
		mov	tmp, B
		long	@@@alu_xor		' paste the alu_xor code snippet here
:end

'*******************************************************
' a9 Exclusive or A with C
'
z_xor_a_c	long	(:end - $ - 1) + (alu_xor_end - alu_xor - 1) - 1
		mov	tmp, C
		long	@@@alu_xor		' paste the alu_xor code snippet here
:end

'*******************************************************
' aa Exclusive or A with D
'
z_xor_a_d	long	(:end - $ - 1) + (alu_xor_end - alu_xor - 1) - 1
		mov	tmp, D
		long	@@@alu_xor		' paste the alu_xor code snippet here
:end

'*******************************************************
' ab Exclusive or A with E
'
z_xor_a_e	long	(:end - $ - 1) + (alu_xor_end - alu_xor - 1) - 1
		mov	tmp, E
		long	@@@alu_xor		' paste the alu_xor code snippet here
:end

'*******************************************************
' ac Exclusive or A with H
'
z_xor_a_h	long	(:end - $ - 1) + (alu_xor_end - alu_xor - 1) - 1
		mov	tmp, H
		long	@@@alu_xor		' paste the alu_xor code snippet here
:end

'*******************************************************
' ad Exclusive or A with L
'
z_xor_a_l	long	(:end - $ - 1) + (alu_xor_end - alu_xor - 1) - 1
		mov	tmp, L
		long	@@@alu_xor		' paste the alu_xor code snippet here
:end

'*******************************************************
' ae Exclusive or A with memory (HL)
'
z_xor_a_m	long	(:end - $ - 1) + (alu_xor_end - alu_xor - 1) - 1
		call	#rd_byte_hl
		mov	tmp, alu
		long	@@@alu_xor		' paste the alu_xor code snippet here
:end

'*******************************************************
' af Exclusive or A with A
'
z_xor_a_a	long	(:end - $ - 1) + (alu_xor_end - alu_xor - 1) - 1
		mov	tmp, A
		long	@@@alu_xor		' paste the alu_xor code snippet here
:end

'*******************************************************
' b0 Or A with B
'
z_or_a_b	long	(:end - $ - 1) + (alu_or_end - alu_or - 1) - 1
		mov	tmp, B
		long	@@@alu_or		' paste the alu_or code snippet here
:end

'*******************************************************
' b1 Or A with C
'
z_or_a_c	long	(:end - $ - 1) + (alu_or_end - alu_or - 1) - 1
		mov	tmp, C
		long	@@@alu_or		' paste the alu_or code snippet here
:end

'*******************************************************
' b2 Or A with D
'
z_or_a_d	long	(:end - $ - 1) + (alu_or_end - alu_or - 1) - 1
		mov	tmp, D
		long	@@@alu_or		' paste the alu_or code snippet here
:end

'*******************************************************
' b3 Or A with E
'
z_or_a_e	long	(:end - $ - 1) + (alu_or_end - alu_or - 1) - 1
		mov	tmp, E
		long	@@@alu_or		' paste the alu_or code snippet here
:end

'*******************************************************
' b4 Or A with H
'
z_or_a_h	long	(:end - $ - 1) + (alu_or_end - alu_or - 1) - 1
		mov	tmp, H
		long	@@@alu_or		' paste the alu_or code snippet here
:end

'*******************************************************
' b5 Or A with L
'
z_or_a_l	long	(:end - $ - 1) + (alu_or_end - alu_or - 1) - 1
		mov	tmp, L
		long	@@@alu_or		' paste the alu_or code snippet here
:end

'*******************************************************
' b6 Or A with memory (HL)
'
z_or_a_m	long	(:end - $ - 1) + (alu_or_end - alu_or - 1) - 1
		call	#rd_byte_hl
		mov	tmp, alu
		long	@@@alu_or		' paste the alu_or code snippet here
:end

'*******************************************************
' b7 Or A with A
'
z_or_a_a	long	(:end - $ - 1) + (alu_or_end - alu_or - 1) - 1
		mov	tmp, A
		long	@@@alu_or		' paste the alu_or code snippet here
:end

'*******************************************************
' b8 Compare A with B
'
z_cp_a_b	long	(:end - $ - 1) + (alu_cp_end - alu_cp - 1) - 1
		mov	tmp, B
		long	@@@alu_cp		' paste the alu_cp code snippet here
:end

'*******************************************************
' b9 Compare A with C
'
z_cp_a_c	long	(:end - $ - 1) + (alu_cp_end - alu_cp - 1) - 1
		mov	tmp, C
		long	@@@alu_cp		' paste the alu_cp code snippet here
:end

'*******************************************************
' ba Compare A with D
'
z_cp_a_d	long	(:end - $ - 1) + (alu_cp_end - alu_cp - 1) - 1
		mov	tmp, D
		long	@@@alu_cp		' paste the alu_cp code snippet here
:end

'*******************************************************
' bb Compare A with E
'
z_cp_a_e	long	(:end - $ - 1) + (alu_cp_end - alu_cp - 1) - 1
		mov	tmp, E
		long	@@@alu_cp		' paste the alu_cp code snippet here
:end

'*******************************************************
' bc Compare A with H
'
z_cp_a_h	long	(:end - $ - 1) + (alu_cp_end - alu_cp - 1) - 1
		mov	tmp, H
		long	@@@alu_cp		' paste the alu_cp code snippet here
:end

'*******************************************************
' bd Compare A with L
'
z_cp_a_l	long	(:end - $ - 1) + (alu_cp_end - alu_cp - 1) - 1
		mov	tmp, L
		long	@@@alu_cp		' paste the alu_cp code snippet here
:end

'*******************************************************
' be Compare A with memory (HL)
'
z_cp_a_m	long	(:end - $ - 1) + (alu_cp_end - alu_cp - 1) - 1
		call	#rd_byte_hl
		mov	tmp, alu
		long	@@@alu_cp		' paste the alu_cp code snippet here
:end

'*******************************************************
' bf Compare A with A
'
z_cp_a_a	long	(:end - $ - 1) + (alu_cp_end - alu_cp - 1) - 1
		mov	tmp, A
		long	@@@alu_cp		' paste the alu_cp code snippet here
:end

'*******************************************************
' c0 Return on Z flag clear (non zero)
'
z_ret_nz	long	(:end - $ - 1)
		test	F, #ZF		WC	' get Z flag into carry
	if_nc	jmp	#handle_ret		' do the return if Z is clear
:end

'*******************************************************
' c8 Return on Z flag set (zero)
'
z_ret_z		long	(:end - $ - 1)
		test	F, #ZF		WC	' get Z flag into carry
	if_c	jmp	#handle_ret		' do the return if Z is set
:end

'*******************************************************
' d0 Return on C flag clear (non carry)
'
z_ret_nc	long	(:end - $ - 1)
		test	F, #CF		WC	' get C flag into carry
	if_nc	jmp	#handle_ret		' do the return if C is clear
:end

'*******************************************************
' d8 Return on C flag set (carry)
'
z_ret_c		long	(:end - $ - 1)
		test	F, #CF		WC	' get C flag into carry
	if_c	jmp	#handle_ret		' do the return if C is set
:end

'*******************************************************
' e0 Return on P flag clear (parity odd)
'
z_ret_po	long	(:end - $ - 1)
		test	F, #PF		WC	' get P flag into carry
	if_nc	jmp	#handle_ret		' do the return if P is clear
:end

'*******************************************************
' e8 Return on P flag set (parity even)
'
z_ret_pe	long	(:end - $ - 1)
		test	F, #PF		WC	' get P flag into carry
	if_c	jmp	#handle_ret		' do the return if P is set
:end

'*******************************************************
' f0 Return on S flag clear (plus)
'
z_ret_p		long	(:end - $ - 1)
		test	F, #SF		WC	' get S flag into carry
	if_nc	jmp	#handle_ret		' do the return if S is clear
:end

'*******************************************************
' e8 Return on S flag set (minus)
'
z_ret_m		long	(:end - $ - 1)
		test	F, #SF		WC	' get S flag into carry
	if_c	jmp	#handle_ret		' do the return if S is set
:end

'*******************************************************
' c1 Pop BC off the stack
'
z_pop_bc	long	(:end - $ - 1)
		call	#pop_alu
		mov	C, alu
		and	C, #$ff
		mov	B, alu
		shr	B, #8
:end

'*******************************************************
' d1 Pop DE off the stack
'
z_pop_de	long	(:end - $ - 1)
		call	#pop_alu
		mov	E, alu
		and	E, #$ff
		mov	D, alu
		shr	D, #8
:end

'*******************************************************
' e1 Pop HL off the stack
'
z_pop_hl	long	(:end - $ - 1)
		call	#pop_alu
		mov	L, alu
		and	L, #$ff
		mov	H, alu
		shr	H, #8
:end

'*******************************************************
' f1 Pop AF off the stack
'
z_pop_af	long	(:end - $ - 1)
		call	#pop_alu
		mov	F, alu
		and	F, #$ff
		mov	A, alu
		shr	A, #8
:end

'*******************************************************
' c9 Return unconditionally
'
z_ret		long	(:end - $ - 1) + (execute << PATCH_SHIFT)
		jmp	#handle_ret		' do the return, i.e. recompile from new PC
:end

'*******************************************************
' d9 Exchange register sets
'
z_exx		long	(:end - $ - 1)
		mov	ea, alt_regs
		rdbyte	tmp, ea
		wrbyte	B, ea
		add	ea, #1
		mov	B, tmp
		rdbyte	tmp, ea
		wrbyte	C, ea
		add	ea, #1
		mov	C, tmp
		rdbyte	tmp, ea
		wrbyte	D, ea
		add	ea, #1
		mov	D, tmp
		rdbyte	tmp, ea
		wrbyte	E, ea
		add	ea, #1
		mov	E, tmp
		rdbyte	tmp, ea
		wrbyte	H, ea
		add	ea, #1
		mov	H, tmp
		rdbyte	tmp, ea
		wrbyte	L, ea
		add	ea, #1
		mov	L, tmp
:end

'*******************************************************
' e9 Jump to address in HL
'
z_jp_hl		long	(:end - $ - 1) + (execute << PATCH_SHIFT)
		mov	PC, H
		shl	PC, #8
		or	PC, L
		jmp	#compile
:end

'*******************************************************
' f9 Load SP with HL
'
z_ld_sp_hl	long	(:end - $ - 1)
		mov	SP, H
		shl	SP, #8
		or	SP, L
:end

'*******************************************************
' c2 Jump on Z flag clear (non zero)
'
z_jp_nz		long	(:end - $ - 1) + (patch_jp_imm16 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with the address of the imm16
		mov	ea, #0-0		' source may be patched with the destination address if inside PC0..PC
		test	F, #ZF		WC	' get Z flag into carry
	if_nc	tjnz	ea, ea			' jump to already compiled code
	if_nc	jmp	#handle_jp		' or otherwise recompile from aux (new PC)
:end

'*******************************************************
' ca Jump on Z flag set (zero)
'
z_jp_z		long	(:end - $ - 1) + (patch_jp_imm16 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with the address of the imm16
		mov	ea, #0-0		' source may be patched with the destination address if inside PC0..PC
		test	F, #ZF		WC	' get Z flag into carry
	if_c	tjnz	ea, ea			' jump to already compiled code
	if_c	jmp	#handle_jp		' or otherwise recompile from aux (new PC)
:end

'*******************************************************
' d2 Jump on C flag clear (non carry)
'
z_jp_nc		long	(:end - $ - 1) + (patch_jp_imm16 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with the address of the imm16
		mov	ea, #0-0		' source may be patched with the destination address if inside PC0..PC
		test	F, #CF		WC	' get C flag into carry
	if_nc	tjnz	ea, ea			' jump to already compiled code
	if_nc	jmp	#handle_jp		' or otherwise recompile from aux (new PC)
:end

'*******************************************************
' da Jump on C flag set (carry)
'
z_jp_c		long	(:end - $ - 1) + (patch_jp_imm16 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with the address of the imm16
		mov	ea, #0-0		' source may be patched with the destination address if inside PC0..PC
		test	F, #CF		WC	' get C flag into carry
	if_c	tjnz	ea, ea			' jump to already compiled code
	if_c	jmp	#handle_jp		' or otherwise recompile from aux (new PC)
:end

'*******************************************************
' e2 Jump on P flag clear (parity odd)
'
z_jp_po		long	(:end - $ - 1) + (patch_jp_imm16 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with the address of the imm16
		mov	ea, #0-0		' source may be patched with the destination address if inside PC0..PC
		test	F, #PF		WC	' get P flag into carry
	if_nc	tjnz	ea, ea			' jump to already compiled code
	if_nc	jmp	#handle_jp		' or otherwise recompile from aux (new PC)
:end

'*******************************************************
' ea Jump on P flag set (parity even)
'
z_jp_pe		long	(:end - $ - 1) + (patch_jp_imm16 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with the address of the imm16
		mov	ea, #0-0		' source may be patched with the destination address if inside PC0..PC
		test	F, #PF		WC	' get P flag into carry
	if_c	tjnz	ea, ea			' jump to already compiled code
	if_c	jmp	#handle_jp		' or otherwise recompile from aux (new PC)
:end

'*******************************************************
' f2 Jump on S flag clear (plus)
'
z_jp_p		long	(:end - $ - 1) + (patch_jp_imm16 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with the address of the imm16
		mov	ea, #0-0		' source may be patched with the destination address if inside PC0..PC
		test	F, #SF		WC	' get S flag into carry
	if_nc	tjnz	ea, ea			' jump to already compiled code
	if_nc	jmp	#handle_jp		' or otherwise recompile from aux (new PC)
:end

'*******************************************************
' ca Jump on S flag set (minus)
'
z_jp_m		long	(:end - $ - 1) + (patch_jp_imm16 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with the address of the imm16
		mov	ea, #0-0		' source may be patched with the destination address if inside PC0..PC
		test	F, #SF		WC	' get S flag into carry
	if_c	tjnz	ea, ea			' jump to already compiled code
	if_c	jmp	#handle_jp		' or otherwise recompile from aux (new PC)
:end

'*******************************************************
' c3 Jump unconditionally
'
z_jp		long	(:end - $ - 1) + (execute_imm16 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with the address of the imm16
		jmp	#handle_jp		' or otherwise recompile from aux (new PC)
:end

'*******************************************************
' cb Prefixed opcodes (rotate, shift, bit-ops)
'
z_cb_xx		long	(:end - $ - 1) + (prefix_cb << PATCH_SHIFT)
:end

'*******************************************************
' d3 Output A to port imm8 FIXME: flags
'
z_out_port_a	long	(:end - $ - 1) + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with the imm8
		mov	alu, A
		call	#wr_port_alu
:end

'*******************************************************
' db Input port imm8 to A FIXME: flags
'
z_in_a_port	long	(:end - $ - 1) + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with the imm8
		call	#rd_port_alu
		mov	A, alu
:end

'*******************************************************
' e3 Exchange stack contents with HL
'
z_ex_sp_hl	long	(:end - $ - 1)
		mov	ea, SP
		call	#rd_byte
		xor	L, alu
		xor	alu, L
		xor	L, alu
		call	#wr_byte
		add	ea, #1
		call	#rd_byte
		xor	H, alu
		xor	alu, H
		xor	H, alu
		call	#wr_byte
:end

'*******************************************************
' eb Exchange DE with HL
'
z_ex_de_hl	long	(:end - $ - 1)
		mov	tmp, L
		mov	L, E
		mov	E, tmp
		mov	tmp, H
		mov	H, D
		mov	D, tmp
:end

'*******************************************************
' f3 Disable interrupts
'
z_di		long	(:end - $ - 1)
		' nothing to do - yet
:end

'*******************************************************
' fb Enable interrupts
'
z_ei		long	(:end - $ - 1)
		' nothing to do - yet
:end

'*******************************************************
' c4 Call on Z flag clear (non zero)
'
z_call_nz	long	(:end - $ - 1) + (patch_call_imm16 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with the address of imm16
		mov	alu, 0-0		' source is patched with the address of the next PC
		test	F, #ZF		WC	' get Z flag into carry
	if_nc	jmp	#handle_call		' do the call if Z is clear
:end

'*******************************************************
' cc Call on Z flag set (zero)
'
z_call_z	long	(:end - $ - 1) + (patch_call_imm16 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with the address of imm16
		mov	alu, 0-0		' source is patched with the address of the next PC
		test	F, #ZF		WC	' get Z flag into carry
	if_c	jmp	#handle_call		' do the call if Z is set
:end

'*******************************************************
' d4 Call on C flag clear (non carry)
'
z_call_nc	long	(:end - $ - 1) + (patch_call_imm16 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with the address of imm16
		mov	alu, 0-0		' source is patched with the address of the next PC
		test	F, #CF		WC	' get C flag into carry
	if_nc	jmp	#handle_call		' do the call if C is clear
:end

'*******************************************************
' dc Call on C flag set (carry)
'
z_call_c	long	(:end - $ - 1) + (patch_call_imm16 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with the address of imm16
		mov	alu, 0-0		' source is patched with the address of the next PC
		test	F, #CF		WC	' get C flag into carry
	if_c	jmp	#handle_call		' do the call if C is set
:end

'*******************************************************
' e4 Call on P flag clear (odd parity)
'
z_call_po	long	(:end - $ - 1) + (patch_call_imm16 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with the address of imm16
		mov	alu, 0-0		' source is patched with the address of the next PC
		test	F, #PF		WC	' get P flag into carry
	if_nc	jmp	#handle_call		' do the call if P is clear
:end

'*******************************************************
' cc Call on P flag set (even parity)
'
z_call_pe	long	(:end - $ - 1) + (patch_call_imm16 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with the address of imm16
		mov	alu, 0-0		' source is patched with the address of the next PC
		test	F, #PF		WC	' get P flag into carry
	if_c	jmp	#handle_call		' do the call if P is set
:end

'*******************************************************
' f4 Call on S flag clear (plus)
'
z_call_p	long	(:end - $ - 1) + (patch_call_imm16 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with the address of imm16
		mov	alu, 0-0		' source is patched with the address of the next PC
		test	F, #SF		WC	' get S flag into carry
	if_nc	jmp	#handle_call		' do the call if S is clear
:end

'*******************************************************
' fc Call on S flag set (minus)
'
z_call_m	long	(:end - $ - 1) + (patch_call_imm16 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with the address of imm16
		mov	alu, 0-0		' source is patched with the address of the next PC
		test	F, #SF		WC	' get S flag into carry
	if_c	jmp	#handle_call		' do the call if S is set
:end

'*******************************************************
' c5 Push BC to the stack
'
z_push_bc	long	(:end - $ - 1)
		mov	alu, B
		shl	alu, #8
		or	alu, C
		call	#push_alu
:end

'*******************************************************
' d5 Push DE to the stack
'
z_push_de	long	(:end - $ - 1)
		mov	alu, D
		shl	alu, #8
		or	alu, E
		call	#push_alu
:end

'*******************************************************
' e5 Push HL to the stack
'
z_push_hl	long	(:end - $ - 1)
		mov	alu, H
		shl	alu, #8
		or	alu, L
		call	#push_alu
:end

'*******************************************************
' f5 Push AF to the stack
'
z_push_af	long	(:end - $ - 1)
		mov	alu, A
		shl	alu, #8
		or	alu, F
		call	#push_alu
:end

'*******************************************************
' cd Call unconditional
'
z_call		long	(:end - $ - 1) + (patch_call_imm16 << PATCH_SHIFT)
		mov	aux, 0-0		' source is patched with the address of imm16
		mov	alu, 0-0		' source is patched with the address of the next PC
		jmp	#handle_call		' do the call
:end

'*******************************************************
' dd Prefixed opcodes (IX register)
'
z_dd_xx		long	(:end - $ - 1) + (prefix_dd << PATCH_SHIFT)
:end

'*******************************************************
' ed Prefixed opcodes (extended opcodes)
'
z_ed_xx		long	(:end - $ - 1) + (prefix_ed << PATCH_SHIFT)
:end

'*******************************************************
' fd Prefixed opcodes (IY register)
'
z_fd_xx		long	(:end - $ - 1) + (prefix_fd << PATCH_SHIFT)
:end

'*******************************************************
' c6 Add immediate 8bit to A
'
z_add_a_imm8	long	(:end - $ - 1) + (alu_add_end - alu_add - 1) - 1 + (patch_imm8 << PATCH_SHIFT)
		mov	tmp, #0			' source is patched with the imm8
		long	@@@alu_add
:end

'*******************************************************
' ce Add with carry immediate 8bit to A
'
z_adc_a_imm8	long	(:end - $ - 1) + (alu_adc_end - alu_adc - 1) - 1 + (patch_imm8 << PATCH_SHIFT)
		mov	tmp, #0			' source is patched with the imm8
		long	@@@alu_adc
:end

'*******************************************************
' d6 Subtract immediate 8bit from A
'
z_sub_imm8	long	(:end - $ - 1) + (alu_sub_end - alu_sub - 1) - 1 + (patch_imm8 << PATCH_SHIFT)
		mov	tmp, #0			' source is patched with the imm8
		long	@@@alu_sub
:end

'*******************************************************
' de Subtract with carry immediate 8bit from A
'
z_sbc_a_imm8	long	(:end - $ - 1) + (alu_sbc_end - alu_sbc - 1) - 1 + (patch_imm8 << PATCH_SHIFT)
		mov	tmp, #0			' source is patched with the imm8
		long	@@@alu_sbc
:end

'*******************************************************
' e6 And A with immediate 8bit
'
z_and_imm8	long	(:end - $ - 1) + (alu_and_end - alu_and - 1) - 1 + (patch_imm8 << PATCH_SHIFT)
		mov	tmp, #0			' source is patched with the imm8
		long	@@@alu_and
:end

'*******************************************************
' ee Exclusive or A with immediate 8bit
'
z_xor_imm8	long	(:end - $ - 1) + (alu_xor_end - alu_xor - 1) - 1 + (patch_imm8 << PATCH_SHIFT)
		mov	tmp, #0			' source is patched with the imm8
		long	@@@alu_xor
:end

'*******************************************************
' f6 Or A with immediate 8bit
'
z_or_imm8	long	(:end - $ - 1) + (alu_or_end - alu_or - 1) - 1 + (patch_imm8 << PATCH_SHIFT)
		mov	tmp, #0			' source is patched with the imm8
		long	@@@alu_or
:end

'*******************************************************
' fe Compare A with immediate 8bit
'
z_cp_imm8	long	(:end - $ - 1) + (alu_cp_end - alu_cp - 1) - 1 + (patch_imm8 << PATCH_SHIFT)
		mov	tmp, #0			' source is patched with the imm8
		long	@@@alu_cp
:end

'*******************************************************
' c7 Restart at vector $00
'
z_rst_00	long	(:end - $ - 1) + (rst_execute << PATCH_SHIFT)
		mov	aux, #$00
		mov	alu, 0-0		' source is patched with the address of the next PC
		jmp	#handle_call
:end

'*******************************************************
' cf Restart at vector $08
'
z_rst_08	long	(:end - $ - 1) + (rst_execute << PATCH_SHIFT)
		mov	aux, #$08
		mov	alu, 0-0		' source is patched with the address of the next PC
		jmp	#handle_call
:end

'*******************************************************
' d7 Restart at vector $10
'
z_rst_10	long	(:end - $ - 1) + (rst_execute << PATCH_SHIFT)
		mov	aux, #$10
		mov	alu, 0-0		' source is patched with the address of the next PC
		jmp	#handle_call
:end

'*******************************************************
' df Restart at vector $18
'
z_rst_18	long	(:end - $ - 1) + (rst_execute << PATCH_SHIFT)
		mov	aux, #$18
		mov	alu, 0-0		' source is patched with the address of the next PC
		jmp	#handle_call
:end

'*******************************************************
' e7 Restart at vector $20
'
z_rst_20	long	(:end - $ - 1) + (rst_execute << PATCH_SHIFT)
		mov	aux, #$20
		mov	alu, 0-0		' source is patched with the address of the next PC
		jmp	#handle_call
:end

'*******************************************************
' ef Restart at vector $28
'
z_rst_28	long	(:end - $ - 1) + (rst_execute << PATCH_SHIFT)
		mov	aux, #$28
		mov	alu, 0-0		' source is patched with the address of the next PC
		jmp	#handle_call
:end

'*******************************************************
' f7 Restart at vector $30
'
z_rst_30	long	(:end - $ - 1) + (rst_execute << PATCH_SHIFT)
		mov	aux, #$30
		mov	alu, 0-0		' source is patched with the address of the next PC
		jmp	#handle_call
:end

'*******************************************************
' ff Restart at vector $38
'
z_rst_38	long	(:end - $ - 1) + (rst_execute << PATCH_SHIFT)
		mov	aux, #$38
		mov	alu, 0-0		' source is patched with the address of the next PC
		jmp	#handle_call
:end

'******************************************************************************
'******************************************************************************
'
'	CB xx opcodes
'
'******************************************************************************
'******************************************************************************

'*******************************************************
' cb 00 Rotate left circular B
'
z_rlc_b		long	(:end - $ - 1) + (alu_rlc_end - alu_rlc - 1) - 1
		mov	alu, B
		long	@@@alu_rlc
		mov	B, alu
:end

'*******************************************************
' cb 01 Rotate left circular C
'
z_rlc_c		long	(:end - $ - 1) + (alu_rlc_end - alu_rlc - 1) - 1
		mov	alu, C
		long	@@@alu_rlc
		mov	C, alu
:end

'*******************************************************
' cb 02 Rotate left circular D
'
z_rlc_d		long	(:end - $ - 1) + (alu_rlc_end - alu_rlc - 1) - 1
		mov	alu, D
		long	@@@alu_rlc
		mov	D, alu
:end

'*******************************************************
' cb 03 Rotate left circular E
'
z_rlc_e		long	(:end - $ - 1) + (alu_rlc_end - alu_rlc - 1) - 1
		mov	alu, E
		long	@@@alu_rlc
		mov	E, alu
:end

'*******************************************************
' cb 04 Rotate left circular H
'
z_rlc_h		long	(:end - $ - 1) + (alu_rlc_end - alu_rlc - 1) - 1
		mov	alu, H
		long	@@@alu_rlc
		mov	H, alu
:end

'*******************************************************
' cb 05 Rotate left circular L
'
z_rlc_l		long	(:end - $ - 1) + (alu_rlc_end - alu_rlc - 1) - 1
		mov	alu, L
		long	@@@alu_rlc
		mov	L, alu
:end

'*******************************************************
' cb 06 Rotate left circular memory (HL)
'
z_rlc_m		long	(:end - $ - 1) + (alu_rlc_end - alu_rlc - 1) - 1
		call	#rd_byte_hl
		long	@@@alu_rlc
		call	#wr_byte
:end

'*******************************************************
' cb 07 Rotate left circular A
'
z_rlc_a		long	(:end - $ - 1) + (alu_rlc_end - alu_rlc - 1) - 1
		mov	alu, A
		long	@@@alu_rlc
		mov	A, alu
:end

'*******************************************************
' cb 08 Rotate right circular B
'
z_rrc_b		long	(:end - $ - 1) + (alu_rrc_end - alu_rrc - 1) - 1
		mov	alu, B
		long	@@@alu_rrc
		mov	B, alu
:end

'*******************************************************
' cb 09 Rotate right circular C
'
z_rrc_c		long	(:end - $ - 1) + (alu_rrc_end - alu_rrc - 1) - 1
		mov	alu, C
		long	@@@alu_rrc
		mov	C, alu
:end

'*******************************************************
' cb 0a Rotate right circular D
'
z_rrc_d		long	(:end - $ - 1) + (alu_rrc_end - alu_rrc - 1) - 1
		mov	alu, D
		long	@@@alu_rrc
		mov	D, alu
:end

'*******************************************************
' cb 0b Rotate right circular E
'
z_rrc_e		long	(:end - $ - 1) + (alu_rrc_end - alu_rrc - 1) - 1
		mov	alu, E
		long	@@@alu_rrc
		mov	E, alu
:end

'*******************************************************
' cb 0c Rotate right circular H
'
z_rrc_h		long	(:end - $ - 1) + (alu_rrc_end - alu_rrc - 1) - 1
		mov	alu, H
		long	@@@alu_rrc
		mov	H, alu
:end

'*******************************************************
' cb 0d Rotate right circular L
'
z_rrc_l		long	(:end - $ - 1) + (alu_rrc_end - alu_rrc - 1) - 1
		mov	alu, L
		long	@@@alu_rrc
		mov	L, alu
:end

'*******************************************************
' cb 0e Rotate right circular memory (HL)
'
z_rrc_m		long	(:end - $ - 1) + (alu_rrc_end - alu_rrc - 1) - 1
		call	#rd_byte_hl
		long	@@@alu_rrc
		call	#wr_byte
:end

'*******************************************************
' cb 0f Rotate right circular A
'
z_rrc_a		long	(:end - $ - 1) + (alu_rrc_end - alu_rrc - 1) - 1
		mov	alu, A
		long	@@@alu_rrc
		mov	A, alu
:end

'*******************************************************
' cb 10 Rotate left through carry B
'
z_rl_b		long	(:end - $ - 1) + (alu_rl_end - alu_rl - 1) - 1
		mov	alu, B
		long	@@@alu_rl
		mov	B, alu
:end

'*******************************************************
' cb 11 Rotate left through carry C
'
z_rl_c		long	(:end - $ - 1) + (alu_rl_end - alu_rl - 1) - 1
		mov	alu, C
		long	@@@alu_rl
		mov	C, alu
:end

'*******************************************************
' cb 12 Rotate left through carry D
'
z_rl_d		long	(:end - $ - 1) + (alu_rl_end - alu_rl - 1) - 1
		mov	alu, D
		long	@@@alu_rl
		mov	D, alu
:end

'*******************************************************
' cb 13 Rotate left through carry E
'
z_rl_e		long	(:end - $ - 1) + (alu_rl_end - alu_rl - 1) - 1
		mov	alu, E
		long	@@@alu_rl
		mov	E, alu
:end

'*******************************************************
' cb 14 Rotate left through carry H
'
z_rl_h		long	(:end - $ - 1) + (alu_rl_end - alu_rl - 1) - 1
		mov	alu, H
		long	@@@alu_rl
		mov	H, alu
:end

'*******************************************************
' cb 15 Rotate left through carry L
'
z_rl_l		long	(:end - $ - 1) + (alu_rl_end - alu_rl - 1) - 1
		mov	alu, L
		long	@@@alu_rl
		mov	L, alu
:end

'*******************************************************
' cb 16 Rotate left through carry memory (HL)
'
z_rl_m		long	(:end - $ - 1) + (alu_rl_end - alu_rl - 1) - 1
		call	#rd_byte_hl
		long	@@@alu_rl
		call	#wr_byte
:end

'*******************************************************
' cb 17 Rotate left through carry A
'
z_rl_a		long	(:end - $ - 1) + (alu_rl_end - alu_rl - 1) - 1
		mov	alu, A
		long	@@@alu_rl
		mov	A, alu
:end

'*******************************************************
' cb 18 Rotate right through carry B
'
z_rr_b		long	(:end - $ - 1) + (alu_rr_end - alu_rr - 1) - 1
		mov	alu, B
		long	@@@alu_rr
		mov	B, alu
:end

'*******************************************************
' cb 19 Rotate right through carry C
'
z_rr_c		long	(:end - $ - 1) + (alu_rr_end - alu_rr - 1) - 1
		mov	alu, C
		long	@@@alu_rr
		mov	C, alu
:end

'*******************************************************
' cb 1a Rotate right through carry D
'
z_rr_d		long	(:end - $ - 1) + (alu_rr_end - alu_rr - 1) - 1
		mov	alu, D
		long	@@@alu_rr
		mov	D, alu
:end

'*******************************************************
' cb 1b Rotate right through carry E
'
z_rr_e		long	(:end - $ - 1) + (alu_rr_end - alu_rr - 1) - 1
		mov	alu, E
		long	@@@alu_rr
		mov	E, alu
:end

'*******************************************************
' cb 1c Rotate right through carry H
'
z_rr_h		long	(:end - $ - 1) + (alu_rr_end - alu_rr - 1) - 1
		mov	alu, H
		long	@@@alu_rr
		mov	H, alu
:end

'*******************************************************
' cb 1d Rotate right through carry L
'
z_rr_l		long	(:end - $ - 1) + (alu_rr_end - alu_rr - 1) - 1
		mov	alu, L
		long	@@@alu_rr
		mov	L, alu
:end

'*******************************************************
' cb 1e Rotate right through carry memory (HL)
'
z_rr_m		long	(:end - $ - 1) + (alu_rr_end - alu_rr - 1) - 1
		call	#rd_byte_hl
		long	@@@alu_rr
		call	#wr_byte
:end

'*******************************************************
' cb 1f Rotate right through carry A
'
z_rr_a		long	(:end - $ - 1) + (alu_rr_end - alu_rr - 1) - 1
		mov	alu, A
		long	@@@alu_rr
		mov	A, alu
:end

'*******************************************************
' cb 20 Shift arithmetic left B
'
z_sla_b		long	(:end - $ - 1) + (alu_sla_end - alu_sla - 1) - 1
		mov	alu, B
		long	@@@alu_sla
		mov	B, alu
:end

'*******************************************************
' cb 21 Shift arithmetic left C
'
z_sla_c		long	(:end - $ - 1) + (alu_sla_end - alu_sla - 1) - 1
		mov	alu, C
		long	@@@alu_sla
		mov	C, alu
:end

'*******************************************************
' cb 22 Shift arithmetic left D
'
z_sla_d		long	(:end - $ - 1) + (alu_sla_end - alu_sla - 1) - 1
		mov	alu, D
		long	@@@alu_sla
		mov	D, alu
:end

'*******************************************************
' cb 23 Shift arithmetic left E
'
z_sla_e		long	(:end - $ - 1) + (alu_sla_end - alu_sla - 1) - 1
		mov	alu, E
		long	@@@alu_sla
		mov	E, alu
:end

'*******************************************************
' cb 24 Shift arithmetic left H
'
z_sla_h		long	(:end - $ - 1) + (alu_sla_end - alu_sla - 1) - 1
		mov	alu, H
		long	@@@alu_sla
		mov	H, alu
:end

'*******************************************************
' cb 25 Shift arithmetic left L
'
z_sla_l		long	(:end - $ - 1) + (alu_sla_end - alu_sla - 1) - 1
		mov	alu, L
		long	@@@alu_sla
		mov	L, alu
:end

'*******************************************************
' cb 26 Shift arithmetic left memory (HL)
'
z_sla_m		long	(:end - $ - 1) + (alu_sla_end - alu_sla - 1) - 1
		call	#rd_byte_hl
		long	@@@alu_sla
		call	#wr_byte
:end

'*******************************************************
' cb 27 Shift arithmetic left A
'
z_sla_a		long	(:end - $ - 1) + (alu_sla_end - alu_sla - 1) - 1
		mov	alu, A
		long	@@@alu_sla
		mov	A, alu
:end

'*******************************************************
' cb 28 Shift arithmetic right B
'
z_sra_b		long	(:end - $ - 1) + (alu_sra_end - alu_sra - 1) - 1
		mov	alu, B
		long	@@@alu_sra
		mov	B, alu
:end

'*******************************************************
' cb 29 Shift arithmetic right C
'
z_sra_c		long	(:end - $ - 1) + (alu_sra_end - alu_sra - 1) - 1
		mov	alu, C
		long	@@@alu_sra
		mov	C, alu
:end

'*******************************************************
' cb 2a Shift arithmetic right D
'
z_sra_d		long	(:end - $ - 1) + (alu_sra_end - alu_sra - 1) - 1
		mov	alu, D
		long	@@@alu_sra
		mov	D, alu
:end

'*******************************************************
' cb 2b Shift arithmetic right E
'
z_sra_e		long	(:end - $ - 1) + (alu_sra_end - alu_sra - 1) - 1
		mov	alu, E
		long	@@@alu_sra
		mov	E, alu
:end

'*******************************************************
' cb 2c Shift arithmetic right H
'
z_sra_h		long	(:end - $ - 1) + (alu_sra_end - alu_sra - 1) - 1
		mov	alu, H
		long	@@@alu_sra
		mov	H, alu
:end

'*******************************************************
' cb 2d Shift arithmetic right L
'
z_sra_l		long	(:end - $ - 1) + (alu_sra_end - alu_sra - 1) - 1
		mov	alu, L
		long	@@@alu_sra
		mov	L, alu
:end

'*******************************************************
' cb 2e Shift arithmetic right memory (HL)
'
z_sra_m		long	(:end - $ - 1) + (alu_sra_end - alu_sra - 1) - 1
		call	#rd_byte_hl
		long	@@@alu_sra
		call	#wr_byte
:end

'*******************************************************
' cb 2f Shift arithmetic right A
'
z_sra_a		long	(:end - $ - 1) + (alu_sra_end - alu_sra - 1) - 1
		mov	alu, A
		long	@@@alu_sra
		mov	A, alu
:end

'*******************************************************
' cb 30 Shift left inverted B
'
z_sli_b		long	(:end - $ - 1) + (alu_sli_end - alu_sli - 1) - 1
		mov	alu, B
		long	@@@alu_sli
		mov	B, alu
:end

'*******************************************************
' cb 31 Shift left inverted C
'
z_sli_c		long	(:end - $ - 1) + (alu_sli_end - alu_sli - 1) - 1
		mov	alu, C
		long	@@@alu_sli
		mov	C, alu
:end

'*******************************************************
' cb 32 Shift left inverted D
'
z_sli_d		long	(:end - $ - 1) + (alu_sli_end - alu_sli - 1) - 1
		mov	alu, D
		long	@@@alu_sli
		mov	D, alu
:end

'*******************************************************
' cb 33 Shift left inverted E
'
z_sli_e		long	(:end - $ - 1) + (alu_sli_end - alu_sli - 1) - 1
		mov	alu, E
		long	@@@alu_sli
		mov	E, alu
:end

'*******************************************************
' cb 34 Shift left inverted H
'
z_sli_h		long	(:end - $ - 1) + (alu_sli_end - alu_sli - 1) - 1
		mov	alu, H
		long	@@@alu_sli
		mov	H, alu
:end

'*******************************************************
' cb 35 Shift left inverted L
'
z_sli_l		long	(:end - $ - 1) + (alu_sli_end - alu_sli - 1) - 1
		mov	alu, L
		long	@@@alu_sli
		mov	L, alu
:end

'*******************************************************
' cb 36 Shift left inverted memory (HL)
'
z_sli_m		long	(:end - $ - 1) + (alu_sli_end - alu_sli - 1) - 1
		call	#rd_byte_hl
		long	@@@alu_sli
		call	#wr_byte
:end

'*******************************************************
' cb 37 Shift left inverted A
'
z_sli_a		long	(:end - $ - 1) + (alu_sli_end - alu_sli - 1) - 1
		mov	alu, A
		long	@@@alu_sli
		mov	A, alu
:end

'*******************************************************
' cb 38 Shift right logical B
'
z_srl_b		long	(:end - $ - 1) + (alu_srl_end - alu_srl - 1) - 1
		mov	alu, B
		long	@@@alu_srl
		mov	B, alu
:end

'*******************************************************
' cb 39 Shift right logical C
'
z_srl_c		long	(:end - $ - 1) + (alu_srl_end - alu_srl - 1) - 1
		mov	alu, C
		long	@@@alu_srl
		mov	C, alu
:end

'*******************************************************
' cb 3a Shift right logical D
'
z_srl_d		long	(:end - $ - 1) + (alu_srl_end - alu_srl - 1) - 1
		mov	alu, D
		long	@@@alu_srl
		mov	D, alu
:end

'*******************************************************
' cb 3b Shift right logical E
'
z_srl_e		long	(:end - $ - 1) + (alu_srl_end - alu_srl - 1) - 1
		mov	alu, E
		long	@@@alu_srl
		mov	E, alu
:end

'*******************************************************
' cb 3c Shift right logical H
'
z_srl_h		long	(:end - $ - 1) + (alu_srl_end - alu_srl - 1) - 1
		mov	alu, H
		long	@@@alu_srl
		mov	H, alu
:end

'*******************************************************
' cb 3d Shift right logical L
'
z_srl_l		long	(:end - $ - 1) + (alu_srl_end - alu_srl - 1) - 1
		mov	alu, L
		long	@@@alu_srl
		mov	L, alu
:end

'*******************************************************
' cb 3e Shift right logical memory (HL)
'
z_srl_m		long	(:end - $ - 1) + (alu_srl_end - alu_srl - 1) - 1
		call	#rd_byte_hl
		long	@@@alu_srl
		call	#wr_byte
:end

'*******************************************************
' cb 3f Shift right logical A
'
z_srl_a		long	(:end - $ - 1) + (alu_srl_end - alu_srl - 1) - 1
		mov	alu, A
		long	@@@alu_srl
		mov	A, alu
:end

'*******************************************************
' cb 40 Test bit 0 in B
'
z_bit_0_b	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, B
		mov	tmp, #bit0
		long	@@@alu_bit
:end

'*******************************************************
' cb 41 Test bit 0 in C
'
z_bit_0_c	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, C
		mov	tmp, #bit0
		long	@@@alu_bit
:end

'*******************************************************
' cb 42 Test bit 0 in D
'
z_bit_0_d	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, D
		mov	tmp, #bit0
		long	@@@alu_bit
:end

'*******************************************************
' cb 43 Test bit 0 in E
'
z_bit_0_e	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, E
		mov	tmp, #bit0
		long	@@@alu_bit
:end

'*******************************************************
' cb 44 Test bit 0 in H
'
z_bit_0_h	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, H
		mov	tmp, #bit0
		long	@@@alu_bit
:end

'*******************************************************
' cb 45 Test bit 0 in L
'
z_bit_0_l	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, L
		mov	tmp, #bit0
		long	@@@alu_bit
:end

'*******************************************************
' cb 46 Test bit 0 in memory (HL)
'
z_bit_0_m	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		call	#rd_byte_hl
		mov	tmp, #bit0
		long	@@@alu_bit
:end

'*******************************************************
' cb 47 Test bit 0 in A
'
z_bit_0_a	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, A
		mov	tmp, #bit0
		long	@@@alu_bit
:end

'*******************************************************
' cb 48 Test bit 1 in B
'
z_bit_1_b	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, B
		mov	tmp, #bit1
		long	@@@alu_bit
:end

'*******************************************************
' cb 49 Test bit 1 in C
'
z_bit_1_c	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, C
		mov	tmp, #bit1
		long	@@@alu_bit
:end

'*******************************************************
' cb 4a Test bit 1 in D
'
z_bit_1_d	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, D
		mov	tmp, #bit1
		long	@@@alu_bit
:end

'*******************************************************
' cb 4b Test bit 1 in E
'
z_bit_1_e	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, E
		mov	tmp, #bit1
		long	@@@alu_bit
:end

'*******************************************************
' cb 4c Test bit 1 in H
'
z_bit_1_h	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, H
		mov	tmp, #bit1
		long	@@@alu_bit
:end

'*******************************************************
' cb 4d Test bit 1 in L
'
z_bit_1_l	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, L
		mov	tmp, #bit1
		long	@@@alu_bit
:end

'*******************************************************
' cb 4e Test bit 1 in memory (HL)
'
z_bit_1_m	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		call	#rd_byte_hl
		mov	tmp, #bit1
		long	@@@alu_bit
:end

'*******************************************************
' cb 4f Test bit 1 in A
'
z_bit_1_a	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, A
		mov	tmp, #bit1
		long	@@@alu_bit
:end

'*******************************************************
' cb 50 Test bit 2 in B
'
z_bit_2_b	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, B
		mov	tmp, #bit2
		long	@@@alu_bit
:end

'*******************************************************
' cb 51 Test bit 2 in C
'
z_bit_2_c	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, C
		mov	tmp, #bit2
		long	@@@alu_bit
:end

'*******************************************************
' cb 52 Test bit 2 in D
'
z_bit_2_d	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, D
		mov	tmp, #bit2
		long	@@@alu_bit
:end

'*******************************************************
' cb 53 Test bit 2 in E
'
z_bit_2_e	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, E
		mov	tmp, #bit2
		long	@@@alu_bit
:end

'*******************************************************
' cb 54 Test bit 2 in H
'
z_bit_2_h	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, H
		mov	tmp, #bit2
		long	@@@alu_bit
:end

'*******************************************************
' cb 55 Test bit 2 in L
'
z_bit_2_l	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, L
		mov	tmp, #bit2
		long	@@@alu_bit
:end

'*******************************************************
' cb 56 Test bit 2 in memory (HL)
'
z_bit_2_m	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		call	#rd_byte_hl
		mov	tmp, #bit2
		long	@@@alu_bit
:end

'*******************************************************
' cb 57 Test bit 2 in A
'
z_bit_2_a	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, A
		mov	tmp, #bit2
		long	@@@alu_bit
:end

'*******************************************************
' cb 58 Test bit 3 in B
'
z_bit_3_b	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, B
		mov	tmp, #bit3
		long	@@@alu_bit
:end

'*******************************************************
' cb 59 Test bit 3 in C
'
z_bit_3_c	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, C
		mov	tmp, #bit3
		long	@@@alu_bit
:end

'*******************************************************
' cb 5a Test bit 3 in D
'
z_bit_3_d	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, D
		mov	tmp, #bit3
		long	@@@alu_bit
:end

'*******************************************************
' cb 5b Test bit 3 in E
'
z_bit_3_e	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, E
		mov	tmp, #bit3
		long	@@@alu_bit
:end

'*******************************************************
' cb 5c Test bit 3 in H
'
z_bit_3_h	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, H
		mov	tmp, #bit3
		long	@@@alu_bit
:end

'*******************************************************
' cb 5d Test bit 3 in L
'
z_bit_3_l	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, L
		mov	tmp, #bit3
		long	@@@alu_bit
:end

'*******************************************************
' cb 5e Test bit 3 in memory (HL)
'
z_bit_3_m	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		call	#rd_byte_hl
		mov	tmp, #bit3
		long	@@@alu_bit
:end

'*******************************************************
' cb 5f Test bit 3 in A
'
z_bit_3_a	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, A
		mov	tmp, #bit3
		long	@@@alu_bit
:end

'*******************************************************
' cb 60 Test bit 4 in B
'
z_bit_4_b	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, B
		mov	tmp, #bit4
		long	@@@alu_bit
:end

'*******************************************************
' cb 61 Test bit 4 in C
'
z_bit_4_c	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, C
		mov	tmp, #bit4
		long	@@@alu_bit
:end

'*******************************************************
' cb 62 Test bit 4 in D
'
z_bit_4_d	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, D
		mov	tmp, #bit4
		long	@@@alu_bit
:end

'*******************************************************
' cb 63 Test bit 4 in E
'
z_bit_4_e	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, E
		mov	tmp, #bit4
		long	@@@alu_bit
:end

'*******************************************************
' cb 64 Test bit 4 in H
'
z_bit_4_h	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, H
		mov	tmp, #bit4
		long	@@@alu_bit
:end

'*******************************************************
' cb 65 Test bit 4 in L
'
z_bit_4_l	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, L
		mov	tmp, #bit4
		long	@@@alu_bit
:end

'*******************************************************
' cb 66 Test bit 4 in memory (HL)
'
z_bit_4_m	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		call	#rd_byte_hl
		mov	tmp, #bit4
		long	@@@alu_bit
:end

'*******************************************************
' cb 67 Test bit 4 in A
'
z_bit_4_a	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, A
		mov	tmp, #bit4
		long	@@@alu_bit
:end

'*******************************************************
' cb 68 Test bit 5 in B
'
z_bit_5_b	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, B
		mov	tmp, #bit5
		long	@@@alu_bit
:end

'*******************************************************
' cb 69 Test bit 5 in C
'
z_bit_5_c	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, C
		mov	tmp, #bit5
		long	@@@alu_bit
:end

'*******************************************************
' cb 6a Test bit 5 in D
'
z_bit_5_d	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, D
		mov	tmp, #bit5
		long	@@@alu_bit
:end

'*******************************************************
' cb 6b Test bit 5 in E
'
z_bit_5_e	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, E
		mov	tmp, #bit5
		long	@@@alu_bit
:end

'*******************************************************
' cb 6c Test bit 5 in H
'
z_bit_5_h	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, H
		mov	tmp, #bit5
		long	@@@alu_bit
:end

'*******************************************************
' cb 6d Test bit 5 in L
'
z_bit_5_l	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, L
		mov	tmp, #bit5
		long	@@@alu_bit
:end

'*******************************************************
' cb 6e Test bit 5 in memory (HL)
'
z_bit_5_m	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		call	#rd_byte_hl
		mov	tmp, #bit5
		long	@@@alu_bit
:end

'*******************************************************
' cb 6f Test bit 5 in A
'
z_bit_5_a	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, A
		mov	tmp, #bit5
		long	@@@alu_bit
:end

'*******************************************************
' cb 70 Test bit 6 in B
'
z_bit_6_b	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, B
		mov	tmp, #bit6
		long	@@@alu_bit
:end

'*******************************************************
' cb 71 Test bit 6 in C
'
z_bit_6_c	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, C
		mov	tmp, #bit6
		long	@@@alu_bit
:end

'*******************************************************
' cb 72 Test bit 6 in D
'
z_bit_6_d	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, D
		mov	tmp, #bit6
		long	@@@alu_bit
:end

'*******************************************************
' cb 73 Test bit 6 in E
'
z_bit_6_e	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, E
		mov	tmp, #bit6
		long	@@@alu_bit
:end

'*******************************************************
' cb 74 Test bit 6 in H
'
z_bit_6_h	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, H
		mov	tmp, #bit6
		long	@@@alu_bit
:end

'*******************************************************
' cb 75 Test bit 6 in L
'
z_bit_6_l	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, L
		mov	tmp, #bit6
		long	@@@alu_bit
:end

'*******************************************************
' cb 76 Test bit 6 in memory (HL)
'
z_bit_6_m	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		call	#rd_byte_hl
		mov	tmp, #bit6
		long	@@@alu_bit
:end

'*******************************************************
' cb 77 Test bit 6 in A
'
z_bit_6_a	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, A
		mov	tmp, #bit6
		long	@@@alu_bit
:end

'*******************************************************
' cb 78 Test bit 7 in B
'
z_bit_7_b	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, B
		mov	tmp, #bit7
		long	@@@alu_bit
:end

'*******************************************************
' cb 79 Test bit 7 in C
'
z_bit_7_c	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, C
		mov	tmp, #bit7
		long	@@@alu_bit
:end

'*******************************************************
' cb 7a Test bit 7 in D
'
z_bit_7_d	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, D
		mov	tmp, #bit7
		long	@@@alu_bit
:end

'*******************************************************
' cb 7b Test bit 7 in E
'
z_bit_7_e	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, E
		mov	tmp, #bit7
		long	@@@alu_bit
:end

'*******************************************************
' cb 7c Test bit 7 in H
'
z_bit_7_h	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, H
		mov	tmp, #bit7
		long	@@@alu_bit
:end

'*******************************************************
' cb 7d Test bit 7 in L
'
z_bit_7_l	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, L
		mov	tmp, #bit7
		long	@@@alu_bit
:end

'*******************************************************
' cb 7e Test bit 7 in memory (HL)
'
z_bit_7_m	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		call	#rd_byte_hl
		mov	tmp, #bit7
		long	@@@alu_bit
:end

'*******************************************************
' cb 7f Test bit 7 in A
'
z_bit_7_a	long	(:end - $ - 1) + (alu_bit_end - alu_bit - 1) - 1
		mov	alu, A
		mov	tmp, #bit7
		long	@@@alu_bit
:end

'*******************************************************
' cb 80 Reset bit 0 in B
'
z_res_0_b	long	(:end - $ - 1)
		andn	B, #bit0
:end

'*******************************************************
' cb 81 Reset bit 0 in C
'
z_res_0_c	long	(:end - $ - 1)
		andn	C, #bit0
:end

'*******************************************************
' cb 82 Reset bit 0 in D
'
z_res_0_d	long	(:end - $ - 1)
		andn	D, #bit0
:end

'*******************************************************
' cb 83 Reset bit 0 in E
'
z_res_0_e	long	(:end - $ - 1)
		andn	E, #bit0
:end

'*******************************************************
' cb 84 Reset bit 0 in H
'
z_res_0_h	long	(:end - $ - 1)
		andn	H, #bit0
:end

'*******************************************************
' cb 85 Reset bit 0 in L
'
z_res_0_l	long	(:end - $ - 1)
		andn	L, #bit0
:end

'*******************************************************
' cb 86 Reset bit 0 in memory (HL)
'
z_res_0_m	long	(:end - $ - 1)
		call	#rd_byte_hl
		andn	alu, #bit0
		call	#wr_byte
:end

'*******************************************************
' cb 87 Reset bit 0 in A
'
z_res_0_a	long	(:end - $ - 1)
		andn	A, #bit0
:end

'*******************************************************
' cb 88 Reset bit 1 in B
'
z_res_1_b	long	(:end - $ - 1)
		andn	B, #bit1
:end

'*******************************************************
' cb 89 Reset bit 1 in C
'
z_res_1_c	long	(:end - $ - 1)
		andn	C, #bit1
:end

'*******************************************************
' cb 8a Reset bit 1 in D
'
z_res_1_d	long	(:end - $ - 1)
		andn	D, #bit1
:end

'*******************************************************
' cb 8b Reset bit 1 in E
'
z_res_1_e	long	(:end - $ - 1)
		andn	E, #bit1
:end

'*******************************************************
' cb 8c Reset bit 1 in H
'
z_res_1_h	long	(:end - $ - 1)
		andn	H, #bit1
:end

'*******************************************************
' cb 8d Reset bit 1 in L
'
z_res_1_l	long	(:end - $ - 1)
		andn	L, #bit1
:end

'*******************************************************
' cb 8e Reset bit 1 in memory (HL)
'
z_res_1_m	long	(:end - $ - 1)
		call	#rd_byte_hl
		andn	alu, #bit1
		call	#wr_byte
:end

'*******************************************************
' cb 8f Reset bit 1 in A
'
z_res_1_a	long	(:end - $ - 1)
		andn	A, #bit1
:end

'*******************************************************
' cb 90 Reset bit 2 in B
'
z_res_2_b	long	(:end - $ - 1)
		andn	B, #bit2
:end

'*******************************************************
' cb 91 Reset bit 2 in C
'
z_res_2_c	long	(:end - $ - 1)
		andn	C, #bit2
:end

'*******************************************************
' cb 92 Reset bit 2 in D
'
z_res_2_d	long	(:end - $ - 1)
		andn	D, #bit2
:end

'*******************************************************
' cb 93 Reset bit 2 in E
'
z_res_2_e	long	(:end - $ - 1)
		andn	E, #bit2
:end

'*******************************************************
' cb 94 Reset bit 2 in H
'
z_res_2_h	long	(:end - $ - 1)
		andn	H, #bit2
:end

'*******************************************************
' cb 95 Reset bit 2 in L
'
z_res_2_l	long	(:end - $ - 1)
		andn	L, #bit2
:end

'*******************************************************
' cb 96 Reset bit 2 in memory (HL)
'
z_res_2_m	long	(:end - $ - 1)
		call	#rd_byte_hl
		andn	alu, #bit2
		call	#wr_byte
:end

'*******************************************************
' cb 97 Reset bit 2 in A
'
z_res_2_a	long	(:end - $ - 1)
		andn	A, #bit2
:end

'*******************************************************
' cb 98 Reset bit 3 in B
'
z_res_3_b	long	(:end - $ - 1)
		andn	B, #bit3
:end

'*******************************************************
' cb 99 Reset bit 3 in C
'
z_res_3_c	long	(:end - $ - 1)
		andn	C, #bit3
:end

'*******************************************************
' cb 9a Reset bit 3 in D
'
z_res_3_d	long	(:end - $ - 1)
		andn	D, #bit3
:end

'*******************************************************
' cb 9b Reset bit 3 in E
'
z_res_3_e	long	(:end - $ - 1)
		andn	E, #bit3
:end

'*******************************************************
' cb 9c Reset bit 3 in H
'
z_res_3_h	long	(:end - $ - 1)
		andn	H, #bit3
:end

'*******************************************************
' cb 9d Reset bit 3 in L
'
z_res_3_l	long	(:end - $ - 1)
		andn	L, #bit3
:end

'*******************************************************
' cb 9e Reset bit 3 in memory (HL)
'
z_res_3_m	long	(:end - $ - 1)
		call	#rd_byte_hl
		andn	alu, #bit3
		call	#wr_byte
:end

'*******************************************************
' cb 9f Reset bit 3 in A
'
z_res_3_a	long	(:end - $ - 1)
		andn	A, #bit3
:end

'*******************************************************
' cb a0 Reset bit 4 in B
'
z_res_4_b	long	(:end - $ - 1)
		andn	B, #bit4
:end

'*******************************************************
' cb a1 Reset bit 4 in C
'
z_res_4_c	long	(:end - $ - 1)
		andn	C, #bit4
:end

'*******************************************************
' cb a2 Reset bit 4 in D
'
z_res_4_d	long	(:end - $ - 1)
		andn	D, #bit4
:end

'*******************************************************
' cb a3 Reset bit 4 in E
'
z_res_4_e	long	(:end - $ - 1)
		andn	E, #bit4
:end

'*******************************************************
' cb a4 Reset bit 4 in H
'
z_res_4_h	long	(:end - $ - 1)
		andn	H, #bit4
:end

'*******************************************************
' cb a5 Reset bit 4 in L
'
z_res_4_l	long	(:end - $ - 1)
		andn	L, #bit4
:end

'*******************************************************
' cb a6 Reset bit 4 in memory (HL)
'
z_res_4_m	long	(:end - $ - 1)
		call	#rd_byte_hl
		andn	alu, #bit4
		call	#wr_byte
:end

'*******************************************************
' cb a7 Reset bit 4 in A
'
z_res_4_a	long	(:end - $ - 1)
		andn	A, #bit4
:end

'*******************************************************
' cb a8 Reset bit 5 in B
'
z_res_5_b	long	(:end - $ - 1)
		andn	B, #bit5
:end

'*******************************************************
' cb a9 Reset bit 5 in C
'
z_res_5_c	long	(:end - $ - 1)
		andn	C, #bit5
:end

'*******************************************************
' cb aa Reset bit 5 in D
'
z_res_5_d	long	(:end - $ - 1)
		andn	D, #bit5
:end

'*******************************************************
' cb ab Reset bit 5 in E
'
z_res_5_e	long	(:end - $ - 1)
		andn	E, #bit5
:end

'*******************************************************
' cb ac Reset bit 5 in H
'
z_res_5_h	long	(:end - $ - 1)
		andn	H, #bit5
:end

'*******************************************************
' cb ad Reset bit 5 in L
'
z_res_5_l	long	(:end - $ - 1)
		andn	L, #bit5
:end

'*******************************************************
' cb ae Reset bit 5 in memory (HL)
'
z_res_5_m	long	(:end - $ - 1)
		call	#rd_byte_hl
		andn	alu, #bit5
		call	#wr_byte
:end

'*******************************************************
' cb af Reset bit 5 in A
'
z_res_5_a	long	(:end - $ - 1)
		andn	A, #bit5
:end

'*******************************************************
' cb b0 Reset bit 6 in B
'
z_res_6_b	long	(:end - $ - 1)
		andn	B, #bit6
:end

'*******************************************************
' cb b1 Reset bit 6 in C
'
z_res_6_c	long	(:end - $ - 1)
		andn	C, #bit6
:end

'*******************************************************
' cb b2 Reset bit 6 in D
'
z_res_6_d	long	(:end - $ - 1)
		andn	D, #bit6
:end

'*******************************************************
' cb b3 Reset bit 6 in E
'
z_res_6_e	long	(:end - $ - 1)
		andn	E, #bit6
:end

'*******************************************************
' cb b4 Reset bit 6 in H
'
z_res_6_h	long	(:end - $ - 1)
		andn	H, #bit6
:end

'*******************************************************
' cb b5 Reset bit 6 in L
'
z_res_6_l	long	(:end - $ - 1)
		andn	L, #bit6
:end

'*******************************************************
' cb b6 Reset bit 6 in memory (HL)
'
z_res_6_m	long	(:end - $ - 1)
		call	#rd_byte_hl
		andn	alu, #bit6
		call	#wr_byte
:end

'*******************************************************
' cb b7 Reset bit 6 in A
'
z_res_6_a	long	(:end - $ - 1)
		andn	A, #bit6
:end

'*******************************************************
' cb b8 Reset bit 7 in B
'
z_res_7_b	long	(:end - $ - 1)
		andn	B, #bit7
:end

'*******************************************************
' cb b9 Reset bit 7 in C
'
z_res_7_c	long	(:end - $ - 1)
		andn	C, #bit7
:end

'*******************************************************
' cb ba Reset bit 7 in D
'
z_res_7_d	long	(:end - $ - 1)
		andn	D, #bit7
:end

'*******************************************************
' cb bb Reset bit 7 in E
'
z_res_7_e	long	(:end - $ - 1)
		andn	E, #bit7
:end

'*******************************************************
' cb bc Reset bit 7 in H
'
z_res_7_h	long	(:end - $ - 1)
		andn	H, #bit7
:end

'*******************************************************
' cb bd Reset bit 7 in L
'
z_res_7_l	long	(:end - $ - 1)
		andn	L, #bit7
:end

'*******************************************************
' cb be Reset bit 7 in memory (HL)
'
z_res_7_m	long	(:end - $ - 1)
		call	#rd_byte_hl
		andn	alu, #bit7
		call	#wr_byte
:end

'*******************************************************
' cb bf Reset bit 7 in A
'
z_res_7_a	long	(:end - $ - 1)
		andn	A, #bit7
:end

'*******************************************************
' cb c0 Set bit 0 in B
'
z_set_0_b	long	(:end - $ - 1)
		or	B, #bit0
:end

'*******************************************************
' cb c1 Set bit 0 in C
'
z_set_0_c	long	(:end - $ - 1)
		or	C, #bit0
:end

'*******************************************************
' cb c2 Set bit 0 in D
'
z_set_0_d	long	(:end - $ - 1)
		or	D, #bit0
:end

'*******************************************************
' cb c3 Set bit 0 in E
'
z_set_0_e	long	(:end - $ - 1)
		or	E, #bit0
:end

'*******************************************************
' cb c4 Set bit 0 in H
'
z_set_0_h	long	(:end - $ - 1)
		or	H, #bit0
:end

'*******************************************************
' cb c5 Set bit 0 in L
'
z_set_0_l	long	(:end - $ - 1)
		or	L, #bit0
:end

'*******************************************************
' cb c6 Set bit 0 in memory (HL)
'
z_set_0_m	long	(:end - $ - 1)
		call	#rd_byte_hl
		or	alu, #bit0
		call	#wr_byte
:end

'*******************************************************
' cb c7 Set bit 0 in A
'
z_set_0_a	long	(:end - $ - 1)
		or	A, #bit0
:end

'*******************************************************
' cb c8 Set bit 1 in B
'
z_set_1_b	long	(:end - $ - 1)
		or	B, #bit1
:end

'*******************************************************
' cb c9 Set bit 1 in C
'
z_set_1_c	long	(:end - $ - 1)
		or	C, #bit1
:end

'*******************************************************
' cb ca Set bit 1 in D
'
z_set_1_d	long	(:end - $ - 1)
		or	D, #bit1
:end

'*******************************************************
' cb cb Set bit 1 in E
'
z_set_1_e	long	(:end - $ - 1)
		or	E, #bit1
:end

'*******************************************************
' cb cc Set bit 1 in H
'
z_set_1_h	long	(:end - $ - 1)
		or	H, #bit1
:end

'*******************************************************
' cb cd Set bit 1 in L
'
z_set_1_l	long	(:end - $ - 1)
		or	L, #bit1
:end

'*******************************************************
' cb ce Set bit 1 in memory (HL)
'
z_set_1_m	long	(:end - $ - 1)
		call	#rd_byte_hl
		or	alu, #bit1
		call	#wr_byte
:end

'*******************************************************
' cb cf Set bit 1 in A
'
z_set_1_a	long	(:end - $ - 1)
		or	A, #bit1
:end

'*******************************************************
' cb d0 Set bit 2 in B
'
z_set_2_b	long	(:end - $ - 1)
		or	B, #bit2
:end

'*******************************************************
' cb d1 Set bit 2 in C
'
z_set_2_c	long	(:end - $ - 1)
		or	C, #bit2
:end

'*******************************************************
' cb d2 Set bit 2 in D
'
z_set_2_d	long	(:end - $ - 1)
		or	D, #bit2
:end

'*******************************************************
' cb d3 Set bit 2 in E
'
z_set_2_e	long	(:end - $ - 1)
		or	E, #bit2
:end

'*******************************************************
' cb d4 Set bit 2 in H
'
z_set_2_h	long	(:end - $ - 1)
		or	H, #bit2
:end

'*******************************************************
' cb d5 Set bit 2 in L
'
z_set_2_l	long	(:end - $ - 1)
		or	L, #bit2
:end

'*******************************************************
' cb d6 Set bit 2 in memory (HL)
'
z_set_2_m	long	(:end - $ - 1)
		call	#rd_byte_hl
		or	alu, #bit2
		call	#wr_byte
:end

'*******************************************************
' cb d7 Set bit 2 in A
'
z_set_2_a	long	(:end - $ - 1)
		or	A, #bit2
:end

'*******************************************************
' cb d8 Set bit 3 in B
'
z_set_3_b	long	(:end - $ - 1)
		or	B, #bit3
:end

'*******************************************************
' cb d9 Set bit 3 in C
'
z_set_3_c	long	(:end - $ - 1)
		or	C, #bit3
:end

'*******************************************************
' cb da Set bit 3 in D
'
z_set_3_d	long	(:end - $ - 1)
		or	D, #bit3
:end

'*******************************************************
' cb db Set bit 3 in E
'
z_set_3_e	long	(:end - $ - 1)
		or	E, #bit3
:end

'*******************************************************
' cb dc Set bit 3 in H
'
z_set_3_h	long	(:end - $ - 1)
		or	H, #bit3
:end

'*******************************************************
' cb dd Set bit 3 in L
'
z_set_3_l	long	(:end - $ - 1)
		or	L, #bit3
:end

'*******************************************************
' cb de Set bit 3 in memory (HL)
'
z_set_3_m	long	(:end - $ - 1)
		call	#rd_byte_hl
		or	alu, #bit3
		call	#wr_byte
:end

'*******************************************************
' cb df Set bit 3 in A
'
z_set_3_a	long	(:end - $ - 1)
		or	A, #bit3
:end

'*******************************************************
' cb e0 Set bit 4 in B
'
z_set_4_b	long	(:end - $ - 1)
		or	B, #bit4
:end

'*******************************************************
' cb e1 Set bit 4 in C
'
z_set_4_c	long	(:end - $ - 1)
		or	C, #bit4
:end

'*******************************************************
' cb e2 Set bit 4 in D
'
z_set_4_d	long	(:end - $ - 1)
		or	D, #bit4
:end

'*******************************************************
' cb e3 Set bit 4 in E
'
z_set_4_e	long	(:end - $ - 1)
		or	E, #bit4
:end

'*******************************************************
' cb e4 Set bit 4 in H
'
z_set_4_h	long	(:end - $ - 1)
		or	H, #bit4
:end

'*******************************************************
' cb e5 Set bit 4 in L
'
z_set_4_l	long	(:end - $ - 1)
		or	L, #bit4
:end

'*******************************************************
' cb e6 Set bit 4 in memory (HL)
'
z_set_4_m	long	(:end - $ - 1)
		call	#rd_byte_hl
		or	alu, #bit4
		call	#wr_byte
:end

'*******************************************************
' cb e7 Set bit 4 in A
'
z_set_4_a	long	(:end - $ - 1)
		or	A, #bit4
:end

'*******************************************************
' cb e8 Set bit 5 in B
'
z_set_5_b	long	(:end - $ - 1)
		or	B, #bit5
:end

'*******************************************************
' cb e9 Set bit 5 in C
'
z_set_5_c	long	(:end - $ - 1)
		or	C, #bit5
:end

'*******************************************************
' cb ea Set bit 5 in D
'
z_set_5_d	long	(:end - $ - 1)
		or	D, #bit5
:end

'*******************************************************
' cb eb Set bit 5 in E
'
z_set_5_e	long	(:end - $ - 1)
		or	E, #bit5
:end

'*******************************************************
' cb ec Set bit 5 in H
'
z_set_5_h	long	(:end - $ - 1)
		or	H, #bit5
:end

'*******************************************************
' cb ed Set bit 5 in L
'
z_set_5_l	long	(:end - $ - 1)
		or	L, #bit5
:end

'*******************************************************
' cb ee Set bit 5 in memory (HL)
'
z_set_5_m	long	(:end - $ - 1)
		call	#rd_byte_hl
		or	alu, #bit5
		call	#wr_byte
:end

'*******************************************************
' cb ef Set bit 5 in A
'
z_set_5_a	long	(:end - $ - 1)
		or	A, #bit5
:end

'*******************************************************
' cb f0 Set bit 6 in B
'
z_set_6_b	long	(:end - $ - 1)
		or	B, #bit6
:end

'*******************************************************
' cb f1 Set bit 6 in C
'
z_set_6_c	long	(:end - $ - 1)
		or	C, #bit6
:end

'*******************************************************
' cb f2 Set bit 6 in D
'
z_set_6_d	long	(:end - $ - 1)
		or	D, #bit6
:end

'*******************************************************
' cb f3 Set bit 6 in E
'
z_set_6_e	long	(:end - $ - 1)
		or	E, #bit6
:end

'*******************************************************
' cb f4 Set bit 6 in H
'
z_set_6_h	long	(:end - $ - 1)
		or	H, #bit6
:end

'*******************************************************
' cb f5 Set bit 6 in L
'
z_set_6_l	long	(:end - $ - 1)
		or	L, #bit6
:end

'*******************************************************
' cb f6 Set bit 6 in memory (HL)
'
z_set_6_m	long	(:end - $ - 1)
		call	#rd_byte_hl
		or	alu, #bit6
		call	#wr_byte
:end

'*******************************************************
' cb f7 Set bit 6 in A
'
z_set_6_a	long	(:end - $ - 1)
		or	A, #bit6
:end

'*******************************************************
' cb f8 Set bit 7 in B
'
z_set_7_b	long	(:end - $ - 1)
		or	B, #bit7
:end

'*******************************************************
' cb f9 Set bit 7 in C
'
z_set_7_c	long	(:end - $ - 1)
		or	C, #bit7
:end

'*******************************************************
' cb fa Set bit 7 in D
'
z_set_7_d	long	(:end - $ - 1)
		or	D, #bit7
:end

'*******************************************************
' cb fb Set bit 7 in E
'
z_set_7_e	long	(:end - $ - 1)
		or	E, #bit7
:end

'*******************************************************
' cb fc Set bit 7 in H
'
z_set_7_h	long	(:end - $ - 1)
		or	H, #bit7
:end

'*******************************************************
' cb fd Set bit 7 in L
'
z_set_7_l	long	(:end - $ - 1)
		or	L, #bit7
:end

'*******************************************************
' cb fe Set bit 7 in memory (HL)
'
z_set_7_m	long	(:end - $ - 1)
		call	#rd_byte_hl
		or	alu, #bit7
		call	#wr_byte
:end

'*******************************************************
' cb ff Set bit 7 in A
'
z_set_7_a	long	(:end - $ - 1)
		or	A, #bit7
:end

'******************************************************************************
'******************************************************************************
'
'	ED xx opcodes
'
'******************************************************************************
'******************************************************************************

'*******************************************************
' Add with carry 16 bits tmp to HL
'
adc16	long	(alu_adc16_end - $ - 1)
		shr	F, #1	WC, NR
		andn	F, #(HF|NF)
		mov	alu, H
		shl	alu, #8
		or	alu, L
		mov	aux, alu
		addx	alu, tmp
		shl	alu, #16	WZ,WC
		muxc	F, #CF
		muxz	F, #ZF
		shr	alu, #16
		xor	aux, alu
		xor	aux, tmp
		shr	aux, #8
		and	aux, #HF
		or	F, aux
		mov	L, alu
		mov	H, alu
		and	L, #$ff
		shr	H, #8
alu_adc16_end

'*******************************************************
' Subtract with carry 16 bits tmp from HL
'
alu_sbc16	long	(alu_sbc16_end - $ - 1)
		shr	F, #1	WC, NR
		andn	F, #HF
		or	F, #NF
		mov	alu, H
		shl	alu, #8
		or	alu, L
		mov	aux, alu
		subx	alu, tmp
		shl	alu, #16	WZ,WC
		muxc	F, #CF
		muxz	F, #ZF
		shr	alu, #16
		xor	aux, alu
		xor	aux, tmp
		shr	aux, #8
		and	aux, #HF
		or	F, aux
		mov	L, alu
		mov	H, alu
		and	L, #$ff
		shr	H, #8
alu_sbc16_end

'*******************************************************
' Load (DE) with (HL) and increment registers
' sets P flag if BC is not yet zero
'
do_ldi		long	(do_ldi_end - $ - 1)
		call	#rd_byte_hl
		mov	ea, D
		shl	ea, #8
		or	ea, E
		call	#wr_byte
		add	L, #1
		test	L, #$100	WC
		and	L, #$ff
		addx	H, #0
		and	H, #$ff
		add	E, #1
		test	E, #$100	WC
		and	E, #$ff
		addx	D, #0
		and	D, #$ff
		sub	C, #1	WC
		and	C, #$ff	WZ
		subx	B, #0
	if_z	and	B, #$ff	WZ
		and	F, #(SF|ZF|CF)
	if_nz	or	F, #PF
do_ldi_end

'*******************************************************
' Load (DE) with (HL) and decrement registers
' sets P flag if BC is not yet zero
'
do_ldd		long	(do_ldd_end - $ - 1)
		call	#rd_byte_hl
		mov	ea, D
		shl	ea, #8
		or	ea, E
		call	#wr_byte
		sub	L, #1	WC
		and	L, #$ff
		subx	H, #0
		and	H, #$ff	
		sub	E, #1	WC
		and	E, #$ff
		subx	D, #0
		and	D, #$ff	
		sub	C, #1	WC
		and	C, #$ff	WZ
		subx	B, #0
	if_z	and	B, #$ff	WZ
		and	F, #(SF|ZF|CF)
	if_nz	or	F, #PF
do_ldd_end

'*******************************************************
' Compare A with (HL) and increment registers
' sets P flag if BC is not yet zero
'
do_cpi		long	(do_cpi_end - $ - 1)
		call	#rd_byte_hl
		mov	tmp, alu
		and	F, #CF			' keep C flag
		or	F, #NF			' set N flag
		add	L, #1
		test	L, #$100	WC
		and	L, #$ff
		addx	H, #0
		and	H, #$ff
		sub	C, #1		WC
		and	C, #$ff		WZ
		subx	B, #0
	if_z	and	B, #$ff		WZ
	if_nz	or	F, #PF			' set P flag if BC is non zero
		mov	alu, A
		mov	aux, alu		' keep a copy in aux for H flag
		sub	alu, tmp	WZ
		muxz	F, #ZF			' set Z from zero
		test	alu, #$80	WZ
		muxnz	F, #SF			' set S from bit #7
		xor	aux, alu
		xor	aux, tmp
		and	aux, #HF
		or	F, aux			' set H from change of bit #4
do_cpi_end

'*******************************************************
' Compare A with (HL) and decrement registers
' sets P flag if BC is not yet zero
'
do_cpd		long	(do_cpd_end - $ - 1)
		call	#rd_byte_hl
		mov	tmp, alu
		and	F, #CF			' keep C flag
		or	F, #NF			' set N flag
		sub	L, #1		WC
		and	L, #$ff
		subx	H, #0
		and	H, #$ff
		sub	C, #1		WC
		and	C, #$ff		WZ
		subx	B, #0
	if_z	and	B, #$ff		WZ
	if_nz	or	F, #PF			' set P flag if BC is non zero
		mov	alu, A
		mov	aux, alu		' keep a copy in aux for H flag
		sub	alu, tmp	WZ
		muxz	F, #ZF			' set Z from zero
		test	alu, #$80	WZ
		muxnz	F, #SF			' set S from bit #7
		xor	aux, alu
		xor	aux, tmp
		and	aux, #HF
		or	F, aux			' set H from change of bit #4
do_cpd_end

'*******************************************************
' Input from port (BC) store to (HL) and increment
' HL register, decrement B register
' Sets S and Z flags after B register
'
' Not implemented:
' temp = (C + 1) & 0xff + inval
'	WZ = BC + 1
' Set P flag from parity of ((temp & 7) ^ B)
' Set N flag from inval & 0x80
' Set H and C flag from temp > 0xff
'
do_ini		long	(do_ini_end - $ - 1)
		mov	ea, B
		shl	ea, #8
		or	ea, C
		call	#rd_port_alu
		sub	B, #1
		mov	tmp, B
		shl	tmp, #24
		and	B, #$ff
		mov	ea, H
		shl	ea, #8
		or	ea, L
		call	#wr_byte
		add	L, #1
		test	L, #$100	WC
		and	L, #$ff
		addx	H, #0
		and	H, #$ff
		mov	tmp, tmp	WZ,WC		' set zero flag if 0, carry flag if bit 31
		muxz	F, #ZF
		muxc	F, #SF
do_ini_end

'*******************************************************
' Input from port (BC) store to (HL) and increment
' HL register, decrement B register
' Sets S and Z flags after B register
'
' Not implemented:
' temp = (C + 1) & 0xff + inval
'	WZ = BC - 1
' Set P flag from parity of ((((C+1)&0xff) + inval) & 7) ^ B
' Set N flag from inval & 0x80
' Set H and C flag from temp > 0xff
'
do_ind		long	(do_ind_end - $ - 1)
		mov	ea, B
		shl	ea, #8
		or	ea, C
		call	#rd_port_alu
		sub	B, #1
		mov	tmp, B
		shl	tmp, #24
		and	B, #$ff
		mov	ea, H
		shl	ea, #8
		or	ea, L
		call	#wr_byte
		sub	L, #1	WC
		and	L, #$ff
		subx	H, #0
		and	H, #$ff
		mov	tmp, tmp	WZ,WC		' set zero flag if 0, carry flag if bit 31
		muxz	F, #ZF
		muxc	F, #SF
do_ind_end

'*******************************************************
' Output memory (HL) to port (BC) and increment
' HL register, decrement B register
' Sets S and Z flags after B register
'
' Not implemented:
' temp = (L + 1) & 0xff + outval
'	WZ = BC + 1
' Set P flag from parity of ((temp & 7) ^ B)
' Set N flag from inval & 0x80
' Set H and C flag from temp > 0xff
'
do_outi		long	(do_outi_end - $ - 1)
		call	#rd_byte_hl
		add	L, #1
		test	L, #$100	WC
		and	L, #$ff
		addx	H, #0
		and	H, #$ff
		sub	B, #1
		mov	tmp, B
		and	B, #$ff
		shl	tmp, #24
		mov	ea, B
		shl	ea, #8
		or	ea, C
		call	#wr_port_alu
		mov	tmp, tmp	WZ,WC		' set zero flag if 0, carry flag if bit 31
		muxz	F, #ZF
		muxc	F, #SF
do_outi_end

'*******************************************************
' Output memory (HL) to port (BC) and increment
' HL register, decrement B register
' Sets S and Z flags after B register
'
' Not implemented:
' temp = (L + 1) & 0xff + outval
'	WZ = BC - 1
' Set P flag from parity of ((temp & 7) ^ B)
' Set N flag from inval & 0x80
' Set H and C flag from temp > 0xff
'
do_outd		long	(do_outd_end - $ - 1)
		call	#rd_byte_hl
		sub	L, #1		WC
		and	L, #$ff
		subx	H, #0
		and	H, #$ff
		sub	B, #1
		mov	tmp, B
		shl	tmp, #24
		and	B, #$ff
		mov	ea, B
		shl	ea, #8
		or	ea, C
		call	#wr_port_alu
		mov	tmp, tmp	WZ,WC		' set zero flag if 0, carry flag if bit 31
		muxz	F, #ZF
		muxc	F, #SF
do_outd_end

'*******************************************************
' ed 00 Get CNT into B,C,D,E (fake instruction!)
'
z_bcde_cnt	long	(:end - $ - 1)
		mov	B, CNT
		mov	C, B
		mov	E, B
		mov	D, B
		shr	B, #24
		shr	C, #16
		and	C, #$ff
		shr	D, #8
		and	D, #$ff
		and	E, #$ff
:end

'*******************************************************
' ed 40 Input B from port (BC)
'
z_inp_b_bc	long	(:end - $ - 1)
		mov	ea, B
		shl	ea, #8
		or	ea, C
		call	#rd_port_alu
		mov	B, alu
:end

'*******************************************************
' ed 48 Input C from port (BC)
'
z_inp_c_bc	long	(:end - $ - 1)
		mov	ea, B
		shl	ea, #8
		or	ea, C
		call	#rd_port_alu
		mov	C, alu
:end

'*******************************************************
' ed 50 Input D from port (BC)
'
z_inp_d_bc	long	(:end - $ - 1)
		mov	ea, B
		shl	ea, #8
		or	ea, C
		call	#rd_port_alu
		mov	D, alu
:end

'*******************************************************
' ed 58 Input E from port (BC)
'
z_inp_e_bc	long	(:end - $ - 1)
		mov	ea, B
		shl	ea, #8
		or	ea, C
		call	#rd_port_alu
		mov	E, alu
:end

'*******************************************************
' ed 60 Input H from port (BC)
'
z_inp_h_bc	long	(:end - $ - 1)
		mov	ea, B
		shl	ea, #8
		or	ea, C
		call	#rd_port_alu
		mov	H, alu
:end

'*******************************************************
' ed 68 Input L from port (BC)
'
z_inp_l_bc	long	(:end - $ - 1)
		mov	ea, B
		shl	ea, #8
		or	ea, C
		call	#rd_port_alu
		mov	L, alu
:end

'*******************************************************
' ed 70 Input from port (BC)
'
z_inp_0_bc	long	(:end - $ - 1)
		mov	ea, B
		shl	ea, #8
		or	ea, C
		call	#rd_port_alu
:end

'*******************************************************
' ed 78 Input A from port (BC)
'
z_inp_a_bc	long	(:end - $ - 1)
		mov	ea, B
		shl	ea, #8
		or	ea, C
		call	#rd_port_alu
		mov	A, alu
:end

'*******************************************************
' ed 41 Output B to port (BC)
'
z_out_bc_b	long	(:end - $ - 1)
		mov	ea, B
		shl	ea, #8
		or	ea, C
		mov	alu, B
		call	#wr_port_alu
:end

'*******************************************************
' ed 49 Output C to port (BC)
'
z_out_bc_c	long	(:end - $ - 1)
		mov	ea, B
		shl	ea, #8
		or	ea, C
		mov	alu, C
		call	#wr_port_alu
:end

'*******************************************************
' ed 51 Output B to port (BC)
'
z_out_bc_d	long	(:end - $ - 1)
		mov	ea, B
		shl	ea, #8
		or	ea, C
		mov	alu, D
		call	#wr_port_alu
:end

'*******************************************************
' ed 59 Output C to port (BC)
'
z_out_bc_e	long	(:end - $ - 1)
		mov	ea, B
		shl	ea, #8
		or	ea, C
		mov	alu, E
		call	#wr_port_alu
:end

'*******************************************************
' ed 61 Output H to port (BC)
'
z_out_bc_h	long	(:end - $ - 1)
		mov	ea, B
		shl	ea, #8
		or	ea, C
		mov	alu, H
		call	#wr_port_alu
:end

'*******************************************************
' ed 69 Output L to port (BC)
'
z_out_bc_l	long	(:end - $ - 1)
		mov	ea, B
		shl	ea, #8
		or	ea, C
		mov	alu, L
		call	#wr_port_alu
:end

'*******************************************************
' ed 71 Output 0 to port (BC)
'
z_out_bc_0	long	(:end - $ - 1)
		mov	ea, B
		shl	ea, #8
		or	ea, C
		mov	alu, #0
		call	#wr_port_alu
:end

'*******************************************************
' ed 79 Output A to port (BC)
'
z_out_bc_a	long	(:end - $ - 1)
		mov	ea, B
		shl	ea, #8
		or	ea, C
		mov	alu, A
		call	#wr_port_alu
:end

'*******************************************************
' ed 42 Subtract with carry BC from HL
'
z_sbc_hl_bc	long	(:end - $ - 1) + (alu_sbc16_end - alu_sbc16 - 1) - 1
		mov	tmp, B
		shl	tmp, #8
		or	tmp, C
		long	@@@alu_sbc16
:end

'*******************************************************
' ed 4a Add with carry BC to HL
'
z_adc_hl_bc	long	(:end - $ - 1) + (alu_adc16_end - adc16 - 1) - 1
		mov	tmp, B
		shl	tmp, #8
		or	tmp, C
		long	@@@adc16
:end

'*******************************************************
' ed 52 Subtract with carry DE from HL
'
z_sbc_hl_de	long	(:end - $ - 1) + (alu_sbc16_end - alu_sbc16 - 1) - 1
		mov	tmp, D
		shl	tmp, #8
		or	tmp, E
		long	@@@alu_sbc16
:end

'*******************************************************
' ed 5a Add with carry DE to HL
'
z_adc_hl_de	long	(:end - $ - 1) + (alu_adc16_end - adc16 - 1) - 1
		mov	tmp, D
		shl	tmp, #8
		or	tmp, E
		long	@@@adc16
:end

'*******************************************************
' ed 62 Subtract with carry HL from HL
'
z_sbc_hl_hl	long	(:end - $ - 1) + (alu_sbc16_end - alu_sbc16 - 1) - 1
		mov	tmp, H
		shl	tmp, #8
		or	tmp, L
		long	@@@alu_sbc16
:end

'*******************************************************
' ed 6a Add with carry HL to HL
'
z_adc_hl_hl	long	(:end - $ - 1) + (alu_adc16_end - adc16 - 1) - 1
		mov	tmp, H
		shl	tmp, #8
		or	tmp, L
		long	@@@adc16
:end

'*******************************************************
' ed 72 Subtract with carry SP from HL
'
z_sbc_hl_sp	long	(:end - $ - 1) + (alu_sbc16_end - alu_sbc16 - 1) - 1
		mov	tmp, SP
		long	@@@alu_sbc16
:end

'*******************************************************
' ed 7a Add with carry SP to HL
'
z_adc_hl_sp	long	(:end - $ - 1) + (alu_adc16_end - adc16 - 1) - 1
		mov	tmp, SP
		long	@@@adc16
:end

'*******************************************************
' ed 43 Load memory (imm16) with BC
'
z_ld_abs16_bc	long	(:end - $ - 1) + (patch_imm16 << PATCH_SHIFT)
		mov	ea, 0-0		' source is patched with the address of imm16
		mov	alu, C
		call	#wr_byte
		add	ea, #1
		mov	alu, B
		call	#wr_byte
:end

'*******************************************************
' ed 4b Load BC from memory (imm16)
'
z_ld_bc_abs16	long	(:end - $ - 1) + (patch_imm16 << PATCH_SHIFT)
		mov	ea, 0-0		' source is patched with the address of imm16
		call	#rd_byte
		mov	C, alu
		add	ea, #1
		call	#rd_byte
		mov	B, alu
:end

'*******************************************************
' ed 53 Load memory (imm16) with DE
'
z_ld_abs16_de	long	(:end - $ - 1) + (patch_imm16 << PATCH_SHIFT)
		mov	ea, 0-0		' source is patched with the address of imm16
		mov	alu, E
		call	#wr_byte
		add	ea, #1
		mov	alu, D
		call	#wr_byte
:end

'*******************************************************
' ed 5b Load DE from memory (imm16)
'
z_ld_de_abs16	long	(:end - $ - 1) + (patch_imm16 << PATCH_SHIFT)
		mov	ea, 0-0		' source is patched with the address of imm16
		call	#rd_byte
		mov	E, alu
		add	ea, #1
		call	#rd_byte
		mov	D, alu
:end

'*******************************************************
' ed 63 Load memory (imm16) with HL
'
' **** already defined as opcode 22
'
'z_ld_abs16_hl	long	(:end - $ - 1) + (patch_imm16 << PATCH_SHIFT)
'		mov	ea, 0-0		' source is patched with the address of imm16
'		mov	alu, L
'		call	#wr_byte
'		add	ea, #1
'		mov	alu, H
'		call	#wr_byte
':end

'*******************************************************
' ed 6b Load HL from memory (imm16)
'
' **** already defined as opcode 2a
'
'z_ld_hl_abs16	long	(:end - $ - 1) + (patch_imm16 << PATCH_SHIFT)
'		mov	ea, 0-0		' source is patched with the address of imm16
'		call	#rd_byte
'		mov	L, alu
'		add	ea, #1
'		call	#rd_byte
'		mov	H, alu
':end

'*******************************************************
' ed 73 Load memory (imm16) with SP
'
z_ld_abs16_sp	long	(:end - $ - 1) + (patch_imm16 << PATCH_SHIFT)
		mov	ea, 0-0		' source is patched with the address of imm16
		mov	alu, SP		' store SP
		call	#wr_word
:end

'*******************************************************
' ed 7b Load SP from memory (imm16)
'
z_ld_sp_abs16	long	(:end - $ - 1) + (patch_imm16 << PATCH_SHIFT)
		mov	ea, 0-0		' source is patched with the address of imm16
		call	#rd_word
		mov	SP, alu		' to SP
:end

'*******************************************************
' ed 44 Negate accu
'
z_neg		long	(:end - $ - 1) + (alu_sub_end - alu_sub - 1) - 1
		mov	tmp, A
		mov	A, #0
		long	@@@alu_sub
:end

'*******************************************************
' ed 45 Return from non maskable interrupt
'
z_retn		long	(:end - $ - 1) + (execute << PATCH_SHIFT)
		jmp	#handle_ret
:end

'*******************************************************
' ed 5d Return from interrupt
'
z_reti		long	(:end - $ - 1) + (execute << PATCH_SHIFT)
		jmp	#handle_ret
:end

'*******************************************************
' ed 46 Set interrupt mode 0
'
z_im_0		long	(:end - $ - 1)
:end

'*******************************************************
' ed 56 Set interrupt mode 1
'
z_im_1		long	(:end - $ - 1)
:end

'*******************************************************
' ed 5e Set interrupt mode 2
'
z_im_2		long	(:end - $ - 1)
:end

'*******************************************************
' ed 47 Load I with A
'
z_ld_i_a	long	(:end - $ - 1)
		mov	tmp, A
		shl	tmp, #8
		and	R2_I, #$ff
		or	R2_I, tmp
:end

'*******************************************************
' ed 4f Load R with A
'
z_ld_r_a	long	(:end - $ - 1)
		mov	tmp, A
		mov	R, tmp
		and	tmp, #$80
		andn	R2_I, #$ff
		or	R2_I, tmp
:end

'*******************************************************
' ed 57 Load A with I
'
z_ld_a_i	long	(:end - $ - 1)
		mov	tmp, R2_I
		shr	tmp, #8
		mov	A, tmp
		' FIXME: flags handling
:end

'*******************************************************
' ed 5f Load A with R
'
z_ld_a_r	long	(:end - $ - 1)
		mov	tmp, R
		and	tmp, #$7f
		or	tmp, R2_I
		and	tmp, #$ff
		mov	A, tmp
		' FIXME: flags handling
:end

'*******************************************************
' ed 67 Rotate right digit memory (HL)
'
' A:[7][6][5][4][3][2][1][0] M:[7][6][5][4][3][2][1][0]
' before         x  x  x  x     y  y  y  y  z  z  z  z
' after          z  z  z  z     x  x  x  x  y  y  y  y
'
z_rrd_m		long	(:end - $ - 1)
		call	#rd_byte_hl
		mov	tmp, alu
		shr	tmp, #4
		mov	aux, A
		shl	aux, #4
		and	aux, #$f0
		or	tmp, aux
		and	alu, #$0f
		and	A, #$f0
		or	A, alu
		mov	alu, tmp
		call	#wr_byte	
:end

'*******************************************************
' ed 6f Rotate left digit memory (HL)
'
' A:[7][6][5][4][3][2][1][0] M:[7][6][5][4][3][2][1][0]
' before         x  x  x  x     y  y  y  y  z  z  z  z
' after          y  y  y  y     z  z  z  z  x  x  x  x
'
z_rld_m		long	(:end - $ - 1)
		call	#rd_byte_hl
		mov	tmp, alu
		shl	tmp, #4
		mov	aux, A
		and	aux, #$0f
		or	tmp, aux
		shr	alu, #4
		and	A, #$f0
		or	A, alu
		mov	alu, tmp
		and	alu, #$ff
		call	#wr_byte	
:end

'*******************************************************
' ed xx Illegal ED opcode
'
z_illegal	long	(:end - $ - 1)
		' FIXME: do anything?
:end

'*******************************************************
' ed a0 Load memory (DE) with memory (HL) and increment
'
z_ldi		long	(:end - $ - 1) + (do_ldi_end - do_ldi - 1) - 1
		long	@@@do_ldi
:end

'*******************************************************
' ed a8 Load memory (DE) with memory (HL) and decrement
'
z_ldd		long	(:end - $ - 1) + (do_ldd_end - do_ldd - 1) - 1
		long	@@@do_ldd
:end

'*******************************************************
' ed b0 Load memory (DE) with memory (HL), increment
' and repeat until BC == 0
'
z_ldir		long	(:end - $ - 1) + (do_ldi_end - do_ldi - 1) - 1 + (patch_self_1 << PATCH_SHIFT)
		mov	opcode, #0-0		' source is patched with snippet address
		long	@@@do_ldi
	if_nz	jmp	opcode
:end

'*******************************************************
' ed b8 Load memory (DE) with memory (HL), decrement
' and repeat until BC == 0
'
z_lddr		long	(:end - $ - 1) + (do_ldd_end - do_ldd - 1) - 1 + (patch_self_1 << PATCH_SHIFT)
		mov	opcode, #0-0		' source is patched with snippet address
		long	@@@do_ldd
	if_nz	jmp	opcode
:end

'*******************************************************
' ed a1 Compare A with memory (HL) and increment
'
z_cpi		long	(:end - $ - 1) + (do_cpi_end - do_cpi - 1) - 1
		long	@@@do_cpi
:end

'*******************************************************
' ed a9 Compare A with memory (HL) and decrement
'
z_cpd		long	(:end - $ - 1) + (do_cpd_end - do_cpd - 1) - 1
		long	@@@do_cpd
:end

'*******************************************************
' ed b1 Compare A with memory (HL) and increment
' and repeat until Z flag set or BC == 0
'
z_cpir		long	(:end - $ - 1) + (do_cpi_end - do_cpi - 1) - 1 + (patch_self_1 << PATCH_SHIFT)
		mov	opcode, #0-0		' source is patched with snippet address
		long	@@@do_cpi
		test	F, #ZF		WZ
		test	F, #PF		WC
 if_z_and_c	jmp	opcode
:end

'*******************************************************
' ed a9 Compare A with memory (HL) and decrement
' and repeat until Z flag set or BC == 0
'
z_cpdr		long	(:end - $ - 1) + (do_cpd_end - do_cpd - 1) - 1 + (patch_self_1 << PATCH_SHIFT)
		mov	opcode, #0-0		' source is patched with snippet address
		long	@@@do_cpd
		test	F, #ZF		WZ
		test	F, #PF		WC
 if_z_and_c	jmp	opcode
:end

'*******************************************************
' ed a2 Input port (BC) to memory (HL) and increment
'
z_ini		long	(:end - $ - 1) + (do_ini_end - do_ini - 1) - 1
		long	@@@do_ini
:end

'*******************************************************
' ed aa Input port (BC) to memory (HL) and decrement
'
z_ind		long	(:end - $ - 1) + (do_ind_end - do_ind - 1) - 1
		long	@@@do_ind
:end

'*******************************************************
' ed b2 Input port (BC) to memory (HL) and increment
'
z_inir		long	(:end - $ - 1) + (do_ini_end - do_ini - 1) - 1 + (patch_self_1 << PATCH_SHIFT)
		mov	opcode, #0-0		' source is patched with snippet address
		long	@@@do_ini
		test	F, #ZF	WZ
	if_nz	jmp	opcode
:end

'*******************************************************
' ed ba Input port (BC) to memory (HL) and decrement
'
z_indr		long	(:end - $ - 1) + (do_ind_end - do_ind - 1) - 1 + (patch_self_1 << PATCH_SHIFT)
		mov	opcode, #0-0		' source is patched with snippet address
		long	@@@do_ind
		test	F, #ZF	WZ
	if_nz	jmp	opcode
:end

'*******************************************************
' ed a3 Output memory (HL) to port (BC) and increment
'
z_outi		long	(:end - $ - 1) + (do_outi_end - do_outi - 1) - 1
		long	@@@do_outi
:end

'*******************************************************
' ed ab Output memory (HL) to port (BC) and decrement
'
z_outd		long	(:end - $ - 1) + (do_outd_end - do_outd - 1) - 1
		long	@@@do_outd
:end

'*******************************************************
' ed b3 Input port (BC) to memory (HL) and increment
'
z_otir		long	(:end - $ - 1) + (do_outi_end - do_outi - 1) - 1 + (patch_self_1 << PATCH_SHIFT)
		mov	opcode, #0-0		' source is patched with snippet address
		long	@@@do_outi
		test	F, #ZF	WZ
	if_nz	jmp	opcode
:end

'*******************************************************
' ed bb Input port (BC) to memory (HL) and decrement
'
z_otdr		long	(:end - $ - 1) + (do_outd_end - do_outd - 1) - 1 + (patch_self_1 << PATCH_SHIFT)
		mov	opcode, #0-0		' source is patched with snippet address
		long	@@@do_outd
		test	F, #ZF	WZ
	if_nz	jmp	opcode
:end

'******************************************************************************
'******************************************************************************
'
'	DD xx opcodes
'
'******************************************************************************
'******************************************************************************

'*******************************************************
' dd/fd 09 Add BC to IX or IY
'
z_add_xy_bc	long	(:end - $ - 1) + (add16_xy_end - add16_xy - 1) - 1
		mov	tmp, B
		shl	tmp, #8
		or	tmp, C
		long	@@@add16_xy
:end

'*******************************************************
' dd/fd 19 Add DE to IX or IY
'
z_add_xy_de	long	(:end - $ - 1) + (add16_xy_end - add16_xy - 1) - 1
		mov	tmp, D
		shl	tmp, #8
		or	tmp, E
		long	@@@add16_xy
:end

'*******************************************************
' dd/fd 29 Add IX or IY to IX or IY
'
z_add_xy_xy	long	(:end - $ - 1) + (add16_xy_end - add16_xy - 1) - 1
		rdword	tmp, XY
		long	@@@add16_xy
:end

'*******************************************************
' dd/fd 39 Add SP to IX or IY
'
z_add_xy_sp	long	(:end - $ - 1) + (add16_xy_end - add16_xy - 1) - 1
		mov	tmp, SP
		long	@@@add16_xy
:end

'*******************************************************
' dd/fd 21 Load IX or IY with imm16
'
z_ld_xy_imm16	long	(:end - $ - 1) + (patch_imm16 << PATCH_SHIFT)
		mov	alu, 0-0		' source is patched with the address of imm16
		wrword	alu, XY			' store to IX or IY
:end

'*******************************************************
' dd/fd 22 Load memory (imm16) with IX or IY
'
z_ld_abs16_xy	long	(:end - $ - 1) + (patch_imm16 << PATCH_SHIFT)
		mov	ea, 0-0			' source is patched with the address of imm16
		rdword	alu, XY			' get IX or IY
		call	#wr_word
:end

'*******************************************************
' dd/fd 2a Load IX or IY from memory (imm16)
'
z_ld_xy_abs16	long	(:end - $ - 1) + (patch_imm16 << PATCH_SHIFT)
		mov	ea, 0-0			' source is patched with the address of imm16
		call	#rd_word
		wrword	alu, XY			' store to IX or IY
:end

'*******************************************************
' dd/fd 23 Increment IX or IY
'
z_inc_xy	long	(:end - $ - 1)
		rdword	alu, XY
		add	alu, #1
		wrword	alu, XY
:end

'*******************************************************
' dd/fd 2b Decrement IX or IY
'
z_dec_xy	long	(:end - $ - 1)
		rdword	alu, XY
		sub	alu, #1
		wrword	alu, XY
:end


'*******************************************************
' dd/fd 24 Increment HX or HY
'
z_inc_hxy	long	(:end - $ - 1) + (alu_inc_end - alu_inc - 1) - 1
		add	XY, #1			' point to MSB of IX or IY
		rdbyte	alu, XY
		long	@@@alu_inc
		wrbyte	alu, XY
:end


'*******************************************************
' dd/fd 2c Increment LX or LY
'
z_inc_lxy	long	(:end - $ - 1) + (alu_inc_end - alu_inc - 1) - 1
		rdbyte	alu, XY
		long	@@@alu_inc
		wrbyte	alu, XY
:end


'*******************************************************
' dd/fd 25 Decrement HX or HY
'
z_dec_hxy	long	(:end - $ - 1) + (alu_dec_end - alu_dec - 1) - 1
		add	XY, #1			' point to MSB of IX or IY
		rdbyte	alu, XY
		long	@@@alu_dec
		wrbyte	alu, XY
:end

'*******************************************************
' dd/fd 2d Decrement LX or LY
'
z_dec_lxy	long	(:end - $ - 1) + (alu_dec_end - alu_dec - 1) - 1
		rdbyte	alu, XY
		long	@@@alu_dec
		wrbyte	alu, XY
:end


'*******************************************************
' dd/fd 26 Load HX or HY with imm8
'
z_ld_hxy_imm8	long	(:end - $ - 1) + (patch_imm8 << PATCH_SHIFT)
		mov	tmp, #0			' source is patched with the imm8
		add	XY, #1
		wrbyte	tmp, XY			' store to MSB of IX or IY
:end

'*******************************************************
' dd/fd 2e Load LX or LY with imm8
'
z_ld_lxy_imm8	long	(:end - $ - 1) + (patch_imm8 << PATCH_SHIFT)
		mov	tmp, #0			' source is patched with the imm8
		wrbyte	tmp, XY			' store to LSB of IX or IY
:end

'*******************************************************
' dd/fd 34 Increment memory (IX/IY+disp8)
'
z_inc_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_inc_end - alu_inc - 1) - 2 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0			' source is patched with the disp8
		long	@@@mem_xy
		call	#rd_byte
		long	@@@alu_inc
		call	#wr_byte
:end

'*******************************************************
' dd/fd 35 Decrement memory (IX/IY+disp8)
'
z_dec_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_dec_end - alu_dec - 1) - 2 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0			' source is patched with the disp8
		long	@@@mem_xy
		call	#rd_byte
		long	@@@alu_dec
		call	#wr_byte
:end

'*******************************************************
' dd/fd 36 Load memory (IX/IY+offs) with imm8
'
z_ld_mxy_imm8	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (patch_imm8_1 << PATCH_SHIFT)
		mov	ea, #0			' source is patched with the disp8
		mov	alu, #0			' source is patched with the imm8
		long	@@@mem_xy
		call	#wr_byte
:end

'*******************************************************
' dd/fd 44 Load B with HX or HY
'
z_ld_b_hxy	long	(:end - $ - 1)
		add	XY, #1
		rdbyte	B, XY
:end

'*******************************************************
' dd/fd 45 Load B with LX or LY
'
z_ld_b_lxy	long	(:end - $ - 1)
		rdbyte	B, XY
:end

'*******************************************************
' dd/fd 46 Load B with memory (IX/IY+disp8)
'
z_ld_b_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0			' source is patched with the disp8
		long	@@@mem_xy
		call	#rd_byte
		mov	B, alu
:end

'*******************************************************
' dd/fd 4c Load C with HX or HY
'
z_ld_c_hxy	long	(:end - $ - 1)
		add	XY, #1
		rdbyte	C, XY
:end

'*******************************************************
' dd/fd 4d Load C with LX or LY
'
z_ld_c_lxy	long	(:end - $ - 1)
		rdbyte	C, XY
:end

'*******************************************************
' dd/fd 4e Load C with memory (IX/IY+disp8)
'
z_ld_c_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0			' source is patched with the disp8
		long	@@@mem_xy
		call	#rd_byte
		mov	C, alu
:end

'*******************************************************
' dd/fd 54 Load D with HX or HY
'
z_ld_d_hxy	long	(:end - $ - 1)
		add	XY, #1
		rdbyte	D, XY
:end

'*******************************************************
' dd/fd 55 Load D with LX or LY
'
z_ld_d_lxy	long	(:end - $ - 1)
		rdbyte	D, XY
:end

'*******************************************************
' dd/fd 56 Load D with memory (IX/IY+disp8)
'
z_ld_d_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0		' source is patched with the disp8
		long	@@@mem_xy
		call	#rd_byte
		mov	D, alu
:end

'*******************************************************
' dd/fd 5c Load E with HX or HY
'
z_ld_e_hxy	long	(:end - $ - 1)
		add	XY, #1
		rdbyte	E, XY
:end

'*******************************************************
' dd/fd 5d Load E with LX or LY
'
z_ld_e_lxy	long	(:end - $ - 1)
		rdbyte	E, XY
:end

'*******************************************************
' dd/fd 5e Load E with memory (IX/IY+disp8)
'
z_ld_e_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0		' source is patched with the disp8
		long	@@@mem_xy
		call	#rd_byte
		mov	E, alu
:end


'*******************************************************
' dd/fd 60 Load HX or HY with B
'
z_ld_hxy_b	long	(:end - $ - 1)
		add	XY, #1
		wrbyte	B, XY
:end

'*******************************************************
' dd/fd 61 Load HX or HY with C
'
z_ld_hxy_c	long	(:end - $ - 1)
		add	XY, #1
		wrbyte	B, XY
:end

'*******************************************************
' dd/fd 62 Load HX or HY with D
'
z_ld_hxy_d	long	(:end - $ - 1)
		add	XY, #1
		wrbyte	D, XY
:end

'*******************************************************
' dd/fd 63 Load HX or HY with E
'
z_ld_hxy_e	long	(:end - $ - 1)
		add	XY, #1
		wrbyte	E, XY
:end

'*******************************************************
' dd/fd 64 Load HX or HY with HX or HY (NOP)
'
'z_ld_hxy_hxy	long	(:end - $ - 1)
':end

'*******************************************************
' dd/fd 65 Load HX or HY with LX or LY
'
z_ld_hxy_lxy	long	(:end - $ - 1)
		rdbyte	tmp, XY
		add	XY, #1
		wrbyte	tmp, XY
:end

'*******************************************************
' dd/fd 66 Load H with memory (IX/IY+disp8)
'
z_ld_h_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0		' source is patched with the disp8
		long	@@@mem_xy
		call	#rd_byte
		mov	H, alu
:end

'*******************************************************
' dd/fd 67 Load HX or HY with A
'
z_ld_hxy_a	long	(:end - $ - 1)
		add	XY, #1
		wrbyte	A, XY
:end

'*******************************************************
' dd/fd 68 Load LX or LY with B
'
z_ld_lxy_b	long	(:end - $ - 1)
		wrbyte	B, XY
:end

'*******************************************************
' dd/fd 69 Load LX or LY with C
'
z_ld_lxy_c	long	(:end - $ - 1)
		wrbyte	C, XY
:end

'*******************************************************
' dd/fd 6a Load LX or LY with D
'
z_ld_lxy_d	long	(:end - $ - 1)
		wrbyte	D, XY
:end

'*******************************************************
' dd/fd 6b Load LX or LY with E
'
z_ld_lxy_e	long	(:end - $ - 1)
		wrbyte	E, XY
:end

'*******************************************************
' dd/fd 6c Load LX or LY with HX or HY
'
z_ld_lxy_hxy	long	(:end - $ - 1)
		add	XY, #1
		rdbyte	tmp, XY
		sub	XY, #1
		wrbyte	tmp, XY
:end

'*******************************************************
' dd/fd 6d Load LX or LY with LX or LY (NOP)
'
'z_ld_lxy_lxy	long	(:end - $ - 1)
':end

'*******************************************************
' dd/fd 6e Load L with memory (IX/IY+disp8)
'
z_ld_l_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1)  - 1 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0		' source is patched with the disp8
		long	@@@mem_xy
		call	#rd_byte
		mov	L, alu
:end

'*******************************************************
' dd/fd 6f Load LX or LY with A
'
z_ld_lxy_a	long	(:end - $ - 1)
		wrbyte	A, XY
:end

'*******************************************************
' dd/fd 70 Load memory (IX/IY+disp8) with B
'
z_ld_mxy_b	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0		' source is patched with the disp8
		long	@@@mem_xy
		mov	alu, B
		call	#wr_byte
:end

'*******************************************************
' dd/fd 71 Load memory (IX/IY+disp8) with C
'
z_ld_mxy_c	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0		' source is patched with the disp8
		long	@@@mem_xy
		mov	alu, C
		call	#wr_byte
:end

'*******************************************************
' dd/fd 72 Load memory (IX/IY+disp8) with D
'
z_ld_mxy_d	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0		' source is patched with the disp8
		long	@@@mem_xy
		mov	alu, D
		call	#wr_byte
:end

'*******************************************************
' dd/fd 73 Load memory (IX/IY+disp8) with E
'
z_ld_mxy_e	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0		' source is patched with the disp8
		long	@@@mem_xy
		mov	alu, E
		call	#wr_byte
:end

'*******************************************************
' dd/fd 74 Load memory (IX/IY+disp8) with H
'
z_ld_mxy_h	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0		' source is patched with the disp8
		long	@@@mem_xy
		mov	alu, H
		call	#wr_byte
:end

'*******************************************************
' dd/fd 75 Load memory (IX/IY+disp8) with L
'
z_ld_mxy_l	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0		' source is patched with the disp8
		long	@@@mem_xy
		mov	alu, L
		call	#wr_byte
:end

'*******************************************************
' dd/fd 77 Load memory (IX/IY+disp8) with A
'
z_ld_mxy_a	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0		' source is patched with the disp8
		long	@@@mem_xy
		mov	alu, A
		call	#wr_byte
:end

'*******************************************************
' dd/fd 7c Load A with HX or HY
'
z_ld_a_hxy	long	(:end - $ - 1)
		add	XY, #1
		rdbyte	A, XY
:end

'*******************************************************
' dd/fd 7d Load A with LX or LY
'
z_ld_a_lxy	long	(:end - $ - 1)
		rdbyte	A, XY
:end

'*******************************************************
' dd/fd 7e Load A with memory (IX/IY+disp8)
'
z_ld_a_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0		' source is patched with the disp8
		long	@@@mem_xy
		call	#rd_byte
		mov	A, alu
:end

'*******************************************************
' dd/fd 84 Add HX or HY to A
'
z_add_a_hxy	long	(:end - $ - 1) + (alu_add_end - alu_add - 1) - 1
		add	XY, #1
		rdbyte	tmp, XY
		long	@@@alu_add
:end

'*******************************************************
' dd/fd 85 Add LX or LY to A
'
z_add_a_lxy	long	(:end - $ - 1) + (alu_add_end - alu_add - 1) - 1
		rdbyte	tmp, XY
		long	@@@alu_add
:end

'*******************************************************
' dd/fd 86 Add memory (IX/IY+disp8) to A
'
z_add_a_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_add_end - alu_add - 1) - 2 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0		' source is patched with the disp8
		long	@@@mem_xy
		call	#rd_byte
		mov	tmp, alu
		long	@@@alu_add
:end

'*******************************************************
' dd/fd 8c Add with carry HX or HY to A
'
z_adc_a_hxy	long	(:end - $ - 1) + (alu_adc_end - alu_adc - 1) - 1
		add	XY, #1
		rdbyte	tmp, XY
		long	@@@alu_adc
:end

'*******************************************************
' dd/fd 8d Add with carry LX or LY to A
'
z_adc_a_lxy	long	(:end - $ - 1) + (alu_adc_end - alu_adc - 1) - 1
		rdbyte	tmp, XY
		long	@@@alu_adc
:end

'*******************************************************
' dd/fd 8e Add with carry memory (IX/IY+disp8) to A
'
z_adc_a_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_adc_end - alu_adc - 1) - 2 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0		' source is patched with the disp8
		long	@@@mem_xy
		call	#rd_byte
		mov	tmp, alu
		long	@@@alu_adc
:end

'*******************************************************
' dd/fd 94 Subtract HX or HY from A 
'
z_sub_a_hxy	long	(:end - $ - 1) + (alu_sub_end - alu_sub - 1) - 1
		add	XY, #1
		rdbyte	tmp, XY
		long	@@@alu_sub
:end

'*******************************************************
' dd/fd 95 Subtract LX or LY from A 
'
z_sub_a_lxy	long	(:end - $ - 1) + (alu_sub_end - alu_sub - 1) - 1
		rdbyte	tmp, XY
		long	@@@alu_sub
:end

'*******************************************************
' dd/fd 96 Subtract memory (IX+disp8) from A 
'
z_sub_a_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_sub_end - alu_sub - 1) - 2 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0		' source is patched with the disp8
		long	@@@mem_xy
		call	#rd_byte
		mov	tmp, alu
		long	@@@alu_sub
:end

'*******************************************************
' dd/fd 9c Subtract with carry HX or HY from A 
'
z_sbc_a_hxy	long	(:end - $ - 1) + (alu_sbc_end - alu_sbc - 1) - 1
		add	XY, #1
		rdbyte	tmp, XY
		long	@@@alu_sbc
:end

'*******************************************************
' dd/fd 9d Subtract with carry LX or LY from A 
'
z_sbc_a_lxy	long	(:end - $ - 1) + (alu_sbc_end - alu_sbc - 1) - 1
		rdbyte	tmp, XY
		long	@@@alu_sbc
:end

'*******************************************************
' dd/fd 9e Subtract with carry memory (IX/IY+disp8) from A 
'
z_sbc_a_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_sbc_end - alu_sbc - 1) - 2 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0		' source is patched with the disp8
		long	@@@mem_xy
		call	#rd_byte
		mov	tmp, alu
		long	@@@alu_sbc
:end

'*******************************************************
' dd/fd a4 And A with HX or HY
'
z_and_a_hxy	long	(:end - $ - 1) + (alu_and_end - alu_and - 1) - 1
		add	XY, #1
		rdbyte	tmp, XY
		long	@@@alu_and
:end

'*******************************************************
' dd/fd a5 And A with LX or LY
'
z_and_a_lxy	long	(:end - $ - 1) + (alu_and_end - alu_and - 1) - 1
		rdbyte	tmp, XY
		long	@@@alu_and
:end

'*******************************************************
' dd/fd a6 And A with memory (IX/IY+disp8)
'
z_and_a_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_and_end - alu_and - 1) - 2 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0		' source is patched with the disp8
		long	@@@mem_xy
		call	#rd_byte
		mov	tmp, alu
		long	@@@alu_and
:end

'*******************************************************
' dd/fd ac Exclusive or A with HX or HY
'
z_xor_a_hxy	long	(:end - $ - 1) + (alu_xor_end - alu_xor - 1) - 1
		add	XY, #1
		rdbyte	tmp, XY
		long	@@@alu_xor
:end

'*******************************************************
' dd/fd ad Exclusive or A with LX or LY
'
z_xor_a_lxy	long	(:end - $ - 1) + (alu_xor_end - alu_xor - 1) - 1
		rdbyte	tmp, XY
		long	@@@alu_xor
:end

'*******************************************************
' dd/fd ae Exclusive or A with memory (IX/IY+disp8)
'
z_xor_a_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_xor_end - alu_xor - 1) - 2 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0			' source is patched with the disp8
		long	@@@mem_xy
		call	#rd_byte
		mov	tmp, alu
		long	@@@alu_xor
:end

'*******************************************************
' dd/fd b4 Or A with HX or HY
'
z_or_a_hxy	long	(:end - $ - 1) + (alu_or_end - alu_or - 1) - 1
		add	XY, #1
		rdbyte	tmp, XY
		long	@@@alu_or
:end

'*******************************************************
' dd/fd b5 Or A with LX or LY
'
z_or_a_lxy	long	(:end - $ - 1) + (alu_or_end - alu_or - 1) - 1
		rdbyte	tmp, XY
		long	@@@alu_or
:end

'*******************************************************
' dd/fd b6 Or A with memory (IX/IY+disp8)
'
z_or_a_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_or_end - alu_or - 1) - 2 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0		' source is patched with the disp8
		long	@@@mem_xy
		call	#rd_byte
		mov	tmp, alu
		long	@@@alu_or
:end

'*******************************************************
' dd/fd bc Compare A with HX or HY
'
z_cp_a_hxy	long	(:end - $ - 1) + (alu_cp_end - alu_cp - 1) - 1
		add	XY, #1
		rdbyte	tmp, XY
		long	@@@alu_cp
:end

'*******************************************************
' dd/fd bd Compare A with LX or LY
'
z_cp_a_lxy	long	(:end - $ - 1) + (alu_cp_end - alu_cp - 1) - 1
		rdbyte	tmp, XY
		long	@@@alu_cp
:end

'*******************************************************
' dd/fd be Compare A with memory (IX/IY+disp8)
'
z_cp_a_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_cp_end - alu_cp - 1) - 2 + (patch_imm8 << PATCH_SHIFT)
		mov	ea, #0		' source is patched with the disp8
		long	@@@mem_xy
		call	#rd_byte
		mov	tmp, alu
		long	@@@alu_cp
:end

'*******************************************************
' dd/fd cb disp8 xx 2nd prefix (rotate, shift, bit-ops)
'
' unused, because 2nd prefix CB is detected
'
'z_xy_cb_xx	long	(:end - $ - 1) + (prefix_xy_cb << PATCH_SHIFT)
':end

'*******************************************************
' dd e1 Pop IX or IY off the stack
'
z_pop_xy	long	(:end - $ - 1)
		call	#pop_alu
		wrword	alu, XY
:end

'*******************************************************
' dd/fd e3 Exchange stack contents with IX
'
z_ex_sp_xy	long	(:end - $ - 1)
		mov	ea, SP
		call	#rd_word
		rdword	tmp, XY
		xor	tmp, alu
		xor	alu, tmp
		xor	tmp, alu
		wrword	tmp, XY
		call	#wr_word
:end

'*******************************************************
' dd/fd e5 Push IX or IY to the stack
'
z_push_xy	long	(:end - $ - 1)
		rdword	alu, XY
		call	#push_alu
:end

'*******************************************************
' dd/fd e9 Jump to address in IX or IY
'
z_jp_xy		long	(:end - $ - 1)
		rdword	PC, XY
:end

'*******************************************************
' dd/fd f9 Load SP with IX or IY
'
z_ld_sp_xy	long	(:end - $ - 1)
		rdword	SP, XY
:end

'******************************************************************************
'******************************************************************************
'
'	DD/FD CB disp8 xx opcodes
'
'******************************************************************************
'******************************************************************************

'*******************************************************
' dd/fd cb disp8 00-07 Rotate left circular (IXY+disp8)
'
z_rlc_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_rlc_end - alu_rlc - 1) - 2 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		long	@@@alu_rlc
		call	#wr_byte
:end

'*******************************************************
' dd/fd cb disp8 08-0f Rotate right circular (IXY+disp8)
'
z_rrc_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_rrc_end - alu_rrc - 1) - 2 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		long	@@@alu_rrc
		call	#wr_byte
:end

'*******************************************************
' dd/fd cb disp8 10-17 Rotate left through carry (IXY+disp8)
'
z_rl_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_rl_end - alu_rl - 1) - 2 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		long	@@@alu_rl
		call	#wr_byte
:end

'*******************************************************
' dd/fd cb disp8 18-1f Rotate right through carry (IXY+disp8)
'
z_rr_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_rr_end - alu_rr - 1) - 2 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		long	@@@alu_rr
		call	#wr_byte
:end

'*******************************************************
' dd/fd cb disp8 20-27 Shift left arithmetic (IXY+disp8)
'
z_sla_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_sla_end - alu_sla - 1) - 2 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		long	@@@alu_sla
		call	#wr_byte
:end

'*******************************************************
' dd/fd cb disp8 28-2f Shift right arithmetic (IXY+disp8)
'
z_sra_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_sra_end - alu_sra - 1) - 2 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		long	@@@alu_sra
		call	#wr_byte
:end

'*******************************************************
' dd/fd cb disp8 30-37 Shift left inverted (IXY+disp8)
'
z_sli_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_sli_end - alu_sli - 1) - 2 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		long	@@@alu_sli
		call	#wr_byte
:end

'*******************************************************
' dd/fd cb disp8 38-3f Shift right logic (IXY+disp8)
'
z_srl_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_srl_end - alu_srl - 1) - 2 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		long	@@@alu_srl
		call	#wr_byte
:end

'*******************************************************
' dd/fd cb disp8 40-47 Test bit 0 in (IXY+disp8)
'
z_bit_0_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_bit_end - alu_bit - 1) - 2 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		mov	tmp, #bit0
		long	@@@alu_bit
:end

'*******************************************************
' dd/fd cb disp8 48-4f Test bit 1 in (IXY+disp8)
'
z_bit_1_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_bit_end - alu_bit - 1) - 2 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		mov	tmp, #bit1
		long	@@@alu_bit
:end

'*******************************************************
' dd/fd cb disp8 50-57 Test bit 2 in (IXY+disp8)
'
z_bit_2_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_bit_end - alu_bit - 1) - 2 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		mov	tmp, #bit2
		long	@@@alu_bit
:end

'*******************************************************
' dd/fd cb disp8 58-5f Test bit 3 in (IXY+disp8)
'
z_bit_3_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_bit_end - alu_bit - 1) - 2 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		mov	tmp, #bit3
		long	@@@alu_bit
:end

'*******************************************************
' dd/fd cb disp8 60-67 Test bit 4 in (IXY+disp8)
'
z_bit_4_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_bit_end - alu_bit - 1) - 2 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		mov	tmp, #bit4
		long	@@@alu_bit
:end

'*******************************************************
' dd/fd cb disp8 68-6f Test bit 5 in (IXY+disp8)
'
z_bit_5_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_bit_end - alu_bit - 1) - 2 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		mov	tmp, #bit5
		long	@@@alu_bit
:end

'*******************************************************
' dd/fd cb disp8 70-77 Test bit 6 in (IXY+disp8)
'
z_bit_6_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_bit_end - alu_bit - 1) - 2 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		mov	tmp, #bit6
		long	@@@alu_bit
:end

'*******************************************************
' dd/fd cb disp8 78-7f Test bit 7 in (IXY+disp8)
'
z_bit_7_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) + (alu_bit_end - alu_bit - 1) - 2 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		mov	tmp, #bit7
		long	@@@alu_bit
:end

'*******************************************************
' dd/fd cb disp8 80-87 Reset bit 0 in (IXY+disp8)
'
z_res_0_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		andn	alu, #bit0
		call	#wr_byte
:end

'*******************************************************
' dd/fd cb disp8 88-8f Reset bit 1 in (IXY+disp8)
'
z_res_1_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		andn	alu, #bit1
		call	#wr_byte
:end

'*******************************************************
' dd/fd cb disp8 90-97 Reset bit 2 in (IXY+disp8)
'
z_res_2_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		andn	alu, #bit2
		call	#wr_byte
:end

'*******************************************************
' dd/fd cb disp8 98-9f Reset bit 3 in (IXY+disp8)
'
z_res_3_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		andn	alu, #bit3
		call	#wr_byte
:end

'*******************************************************
' dd/fd cb disp8 a0-a7 Reset bit 4 in (IXY+disp8)
'
z_res_4_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		andn	alu, #bit4
		call	#wr_byte
:end

'*******************************************************
' dd/fd cb disp8 a8-af Reset bit 5 in (IXY+disp8)
'
z_res_5_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		andn	alu, #bit5
		call	#wr_byte
:end

'*******************************************************
' dd/fd cb disp8 b0-b7 Reset bit 6 in (IXY+disp8)
'
z_res_6_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		andn	alu, #bit6
		call	#wr_byte
:end

'*******************************************************
' dd/fd cb disp8 b8-bf Reset bit 7 in (IXY+disp8)
'
z_res_7_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		andn	alu, #bit7
		call	#wr_byte
:end

'*******************************************************
' dd/fd cb disp8 c0-c7 Set bit 0 in (IXY+disp8)
'
z_set_0_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		or	alu, #bit0
		call	#wr_byte
:end

'*******************************************************
' dd/fd cb disp8 c8-cf Set bit 1 in (IXY+disp8)
'
z_set_1_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		or	alu, #bit1
		call	#wr_byte
:end

'*******************************************************
' dd/fd cb disp8 d0-d7 Set bit 2 in (IXY+disp8)
'
z_set_2_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		or	alu, #bit2
		call	#wr_byte
:end

'*******************************************************
' dd/fd cb disp8 d8-df Set bit 3 in (IXY+disp8)
'
z_set_3_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		or	alu, #bit3
		call	#wr_byte
:end

'*******************************************************
' dd/fd cb disp8 e0-e7 Set bit 4 in (IXY+disp8)
'
z_set_4_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		or	alu, #bit4
		call	#wr_byte
:end

'*******************************************************
' dd/fd cb disp8 e8-ef Set bit 5 in (IXY+disp8)
'
z_set_5_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		or	alu, #bit5
		call	#wr_byte
:end

'*******************************************************
' dd/fd cb disp8 f0-f7 Set bit 6 in (IXY+disp8)
'
z_set_6_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		or	alu, #bit6
		call	#wr_byte
:end

'*******************************************************
' dd/fd cb disp8 f8-ff Set bit 7 in (IXY+disp8)
'
z_set_7_mxy	long	(:end - $ - 1) + (mem_xy_end - mem_xy - 1) - 1 + (prefix_xy_cb << PATCH_SHIFT)
		mov	ea, #0-0		' source is patched with disp8
		long	@@@mem_xy
		call	#rd_byte
		or	alu, #bit7
		call	#wr_byte
:end

'******************************************************************************
'******************************************************************************
'
'	code snippets tables
'
'******************************************************************************
'******************************************************************************

snippets_00
	{00}	word	@@@z_nop
	{01}	word	@@@z_ld_bc_imm16
	{02}	word	@@@z_ld_mem_bc_a
	{03}	word	@@@z_inc_bc
	{04}	word	@@@z_inc_b
	{05}	word	@@@z_dec_b
	{06}	word	@@@z_ld_b_imm8
	{07}	word	@@@z_rlca
	{08}	word	@@@z_ex_af_af2
	{09}	word	@@@z_add_hl_bc
	{0a}	word	@@@z_ld_a_mem_bc
	{0b}	word	@@@z_dec_bc
	{0c}	word	@@@z_inc_c
	{0d}	word	@@@z_dec_c
	{0e}	word	@@@z_ld_c_imm8
	{0f}	word	@@@z_rrca
	{10}	word	@@@z_djnz
	{11}	word	@@@z_ld_de_imm16
	{12}	word	@@@z_ld_mem_de_a
	{13}	word	@@@z_inc_de
	{14}	word	@@@z_inc_d
	{15}	word	@@@z_dec_d
	{16}	word	@@@z_ld_d_imm8
	{17}	word	@@@z_rla
	{18}	word	@@@z_jr
	{19}	word	@@@z_add_hl_de
	{1a}	word	@@@z_ld_a_mem_de
	{1b}	word	@@@z_dec_de
	{1c}	word	@@@z_inc_e
	{1d}	word	@@@z_dec_e
	{1e}	word	@@@z_ld_e_imm8
	{1f}	word	@@@z_rra
	{20}	word	@@@z_jr_nz
	{21}	word	@@@z_ld_hl_imm16
	{22}	word	@@@z_ld_abs16_hl
	{23}	word	@@@z_inc_hl
	{24}	word	@@@z_inc_h
	{25}	word	@@@z_dec_h
	{26}	word	@@@z_ld_h_imm8
	{27}	word	@@@z_daa
	{28}	word	@@@z_jr_z
	{29}	word	@@@z_add_hl_hl
	{2a}	word	@@@z_ld_hl_abs16
	{2b}	word	@@@z_dec_hl
	{2c}	word	@@@z_inc_l
	{2d}	word	@@@z_dec_l
	{2e}	word	@@@z_ld_l_imm8
	{2f}	word	@@@z_cpl
	{30}	word	@@@z_jr_nc
	{31}	word	@@@z_ld_sp_imm16
	{32}	word	@@@z_ld_mem_a
	{33}	word	@@@z_inc_sp
	{34}	word	@@@z_inc_m
	{35}	word	@@@z_dec_m
	{36}	word	@@@z_ld_m_imm8
	{37}	word	@@@z_scf
	{38}	word	@@@z_jr_c
	{39}	word	@@@z_add_hl_sp
	{3a}	word	@@@z_ld_a_mem
	{3b}	word	@@@z_dec_sp
	{3c}	word	@@@z_inc_a
	{3d}	word	@@@z_dec_a
	{3e}	word	@@@z_ld_a_imm8
	{3f}	word	@@@z_ccf
	{40}	word	@@@z_nop		' z_ld_b_b
	{41}	word	@@@z_ld_b_c
	{42}	word	@@@z_ld_b_d
	{43}	word	@@@z_ld_b_e
	{44}	word	@@@z_ld_b_h
	{45}	word	@@@z_ld_b_l
	{46}	word	@@@z_ld_b_m
	{47}	word	@@@z_ld_b_a
	{48}	word	@@@z_ld_c_b
	{49}	word	@@@z_nop		' z_ld_c_c
	{4a}	word	@@@z_ld_c_d
	{4b}	word	@@@z_ld_c_e
	{4c}	word	@@@z_ld_c_h
	{4d}	word	@@@z_ld_c_l
	{4e}	word	@@@z_ld_c_m
	{4f}	word	@@@z_ld_c_a
	{50}	word	@@@z_ld_d_b
	{51}	word	@@@z_ld_d_c
	{52}	word	@@@z_nop		' z_ld_d_d
	{53}	word	@@@z_ld_d_e
	{54}	word	@@@z_ld_d_h
	{55}	word	@@@z_ld_d_l
	{56}	word	@@@z_ld_d_m
	{57}	word	@@@z_ld_d_a
	{58}	word	@@@z_ld_e_b
	{59}	word	@@@z_ld_e_c
	{5a}	word	@@@z_ld_e_d
	{5b}	word	@@@z_nop		' z_ld_e_e
	{5c}	word	@@@z_ld_e_h
	{5d}	word	@@@z_ld_e_l
	{5e}	word	@@@z_ld_e_m
	{5f}	word	@@@z_ld_e_a
	{60}	word	@@@z_ld_h_b
	{61}	word	@@@z_ld_h_c
	{62}	word	@@@z_ld_h_d
	{63}	word	@@@z_ld_h_e
	{64}	word	@@@z_nop		' z_ld_h_h
	{65}	word	@@@z_ld_h_l
	{66}	word	@@@z_ld_h_m
	{67}	word	@@@z_ld_h_a
	{68}	word	@@@z_ld_l_b
	{69}	word	@@@z_ld_l_c
	{6a}	word	@@@z_ld_l_d
	{6b}	word	@@@z_ld_l_e
	{6c}	word	@@@z_ld_l_h
	{6d}	word	@@@z_nop		' z_ld_l_l
	{6e}	word	@@@z_ld_l_m
	{6f}	word	@@@z_ld_l_a
	{70}	word	@@@z_ld_m_b
	{71}	word	@@@z_ld_m_c
	{72}	word	@@@z_ld_m_d
	{73}	word	@@@z_ld_m_e
	{74}	word	@@@z_ld_m_h
	{75}	word	@@@z_ld_m_l
	{76}	word	@@@z_halt
	{77}	word	@@@z_ld_m_a
	{78}	word	@@@z_ld_a_b
	{79}	word	@@@z_ld_a_c
	{7a}	word	@@@z_ld_a_d
	{7b}	word	@@@z_ld_a_e
	{7c}	word	@@@z_ld_a_h
	{7d}	word	@@@z_ld_a_l
	{7e}	word	@@@z_ld_a_m
	{7f}	word	@@@z_nop		' z_ld_a_a
	{80}	word	@@@z_add_a_b
	{81}	word	@@@z_add_a_c
	{82}	word	@@@z_add_a_d
	{83}	word	@@@z_add_a_e
	{84}	word	@@@z_add_a_h
	{85}	word	@@@z_add_a_l
	{86}	word	@@@z_add_a_m
	{87}	word	@@@z_add_a_a
	{88}	word	@@@z_adc_a_b
	{89}	word	@@@z_adc_a_c
	{8a}	word	@@@z_adc_a_d
	{8b}	word	@@@z_adc_a_e
	{8c}	word	@@@z_adc_a_h
	{8d}	word	@@@z_adc_a_l
	{8e}	word	@@@z_adc_a_m
	{8f}	word	@@@z_adc_a_a
	{90}	word	@@@z_sub_a_b
	{91}	word	@@@z_sub_a_c
	{92}	word	@@@z_sub_a_d
	{93}	word	@@@z_sub_a_e
	{94}	word	@@@z_sub_a_h
	{95}	word	@@@z_sub_a_l
	{96}	word	@@@z_sub_a_m
	{97}	word	@@@z_sub_a_a
	{98}	word	@@@z_sbc_a_b
	{99}	word	@@@z_sbc_a_c
	{9a}	word	@@@z_sbc_a_d
	{9b}	word	@@@z_sbc_a_e
	{9c}	word	@@@z_sbc_a_h
	{9d}	word	@@@z_sbc_a_l
	{9e}	word	@@@z_sbc_a_m
	{9f}	word	@@@z_sbc_a_a
	{a0}	word	@@@z_and_a_b
	{a1}	word	@@@z_and_a_c
	{a2}	word	@@@z_and_a_d
	{a3}	word	@@@z_and_a_e
	{a4}	word	@@@z_and_a_h
	{a5}	word	@@@z_and_a_l
	{a6}	word	@@@z_and_a_m
	{a7}	word	@@@z_and_a_a
	{a8}	word	@@@z_xor_a_b
	{a9}	word	@@@z_xor_a_c
	{aa}	word	@@@z_xor_a_d
	{ab}	word	@@@z_xor_a_e
	{ac}	word	@@@z_xor_a_h
	{ad}	word	@@@z_xor_a_l
	{ae}	word	@@@z_xor_a_m
	{af}	word	@@@z_xor_a_a
	{b0}	word	@@@z_or_a_b
	{b1}	word	@@@z_or_a_c
	{b2}	word	@@@z_or_a_d
	{b3}	word	@@@z_or_a_e
	{b4}	word	@@@z_or_a_h
	{b5}	word	@@@z_or_a_l
	{b6}	word	@@@z_or_a_m
	{b7}	word	@@@z_or_a_a
	{b8}	word	@@@z_cp_a_b
	{b9}	word	@@@z_cp_a_c
	{ba}	word	@@@z_cp_a_d
	{bb}	word	@@@z_cp_a_e
	{bc}	word	@@@z_cp_a_h
	{bd}	word	@@@z_cp_a_l
	{be}	word	@@@z_cp_a_m
	{bf}	word	@@@z_cp_a_a
	{c0}	word	@@@z_ret_nz
	{c1}	word	@@@z_pop_bc
	{c2}	word	@@@z_jp_nz
	{c3}	word	@@@z_jp
	{c4}	word	@@@z_call_nz
	{c5}	word	@@@z_push_bc
	{c6}	word	@@@z_add_a_imm8
	{c7}	word	@@@z_rst_00
	{c8}	word	@@@z_ret_z
	{c9}	word	@@@z_ret
	{ca}	word	@@@z_jp_z
	{cb}	word	@@@z_cb_xx
	{cc}	word	@@@z_call_z
	{cd}	word	@@@z_call
	{ce}	word	@@@z_adc_a_imm8
	{c7}	word	@@@z_rst_08
	{d0}	word	@@@z_ret_nc
	{d1}	word	@@@z_pop_de
	{d2}	word	@@@z_jp_nc
	{d3}	word	@@@z_out_port_a
	{d4}	word	@@@z_call_nc
	{d5}	word	@@@z_push_de
	{d6}	word	@@@z_sub_imm8
	{d7}	word	@@@z_rst_10
	{d8}	word	@@@z_ret_c
	{d9}	word	@@@z_exx
	{da}	word	@@@z_jp_c
	{db}	word	@@@z_in_a_port
	{dc}	word	@@@z_call_c
	{dd}	word	@@@z_dd_xx
	{de}	word	@@@z_sbc_a_imm8
	{df}	word	@@@z_rst_18
	{e0}	word	@@@z_ret_po
	{e1}	word	@@@z_pop_hl
	{e2}	word	@@@z_jp_po
	{e3}	word	@@@z_ex_sp_hl
	{e4}	word	@@@z_call_po
	{e5}	word	@@@z_push_hl
	{e6}	word	@@@z_and_imm8
	{e7}	word	@@@z_rst_20
	{e8}	word	@@@z_ret_pe
	{e9}	word	@@@z_jp_hl
	{ea}	word	@@@z_jp_pe
	{eb}	word	@@@z_ex_de_hl
	{ec}	word	@@@z_call_pe
	{ed}	word	@@@z_ed_xx
	{ee}	word	@@@z_xor_imm8
	{ef}	word	@@@z_rst_28
	{f0}	word	@@@z_ret_p
	{f1}	word	@@@z_pop_af
	{f2}	word	@@@z_jp_p
	{f3}	word	@@@z_di
	{f4}	word	@@@z_call_p
	{f5}	word	@@@z_push_af
	{f6}	word	@@@z_or_imm8
	{f7}	word	@@@z_rst_30
	{f8}	word	@@@z_ret_m
	{f9}	word	@@@z_ld_sp_hl
	{fa}	word	@@@z_jp_m
	{fb}	word	@@@z_ei
	{fc}	word	@@@z_call_m
	{fd}	word	@@@z_fd_xx
	{fe}	word	@@@z_cp_imm8
	{ff}	word	@@@z_rst_38

snippets_cb
	{00}	word	@@@z_rlc_b
	{01}	word	@@@z_rlc_c
	{02}	word	@@@z_rlc_d
	{03}	word	@@@z_rlc_e
	{04}	word	@@@z_rlc_h
	{05}	word	@@@z_rlc_l
	{06}	word	@@@z_rlc_m
	{07}	word	@@@z_rlc_a
	{08}	word	@@@z_rrc_b
	{09}	word	@@@z_rrc_c
	{0a}	word	@@@z_rrc_d
	{0b}	word	@@@z_rrc_e
	{0c}	word	@@@z_rrc_h
	{0d}	word	@@@z_rrc_l
	{0e}	word	@@@z_rrc_m
	{0f}	word	@@@z_rrc_a
	{10}	word	@@@z_rl_b
	{11}	word	@@@z_rl_c
	{12}	word	@@@z_rl_d
	{13}	word	@@@z_rl_e
	{14}	word	@@@z_rl_h
	{15}	word	@@@z_rl_l
	{16}	word	@@@z_rl_m
	{17}	word	@@@z_rl_a
	{18}	word	@@@z_rr_b
	{19}	word	@@@z_rr_c
	{1a}	word	@@@z_rr_d
	{1b}	word	@@@z_rr_e
	{1c}	word	@@@z_rr_h
	{1d}	word	@@@z_rr_l
	{1e}	word	@@@z_rr_m
	{1f}	word	@@@z_rr_a
	{20}	word	@@@z_sla_b
	{21}	word	@@@z_sla_c
	{22}	word	@@@z_sla_d
	{23}	word	@@@z_sla_e
	{24}	word	@@@z_sla_h
	{25}	word	@@@z_sla_l
	{26}	word	@@@z_sla_m
	{27}	word	@@@z_sla_a
	{28}	word	@@@z_sra_b
	{29}	word	@@@z_sra_c
	{2a}	word	@@@z_sra_d
	{2b}	word	@@@z_sra_e
	{2c}	word	@@@z_sra_h
	{2d}	word	@@@z_sra_l
	{2e}	word	@@@z_sra_m
	{2f}	word	@@@z_sra_a
	{30}	word	@@@z_sli_b
	{31}	word	@@@z_sli_c
	{32}	word	@@@z_sli_d
	{33}	word	@@@z_sli_e
	{34}	word	@@@z_sli_h
	{35}	word	@@@z_sli_l
	{36}	word	@@@z_sli_m
	{37}	word	@@@z_sli_a
	{38}	word	@@@z_srl_b
	{39}	word	@@@z_srl_c
	{3a}	word	@@@z_srl_d
	{3b}	word	@@@z_srl_e
	{3c}	word	@@@z_srl_h
	{3d}	word	@@@z_srl_l
	{3e}	word	@@@z_srl_m
	{3f}	word	@@@z_srl_a
	{40}	word	@@@z_bit_0_b
	{41}	word	@@@z_bit_0_c
	{42}	word	@@@z_bit_0_d
	{43}	word	@@@z_bit_0_e
	{44}	word	@@@z_bit_0_h
	{45}	word	@@@z_bit_0_l
	{46}	word	@@@z_bit_0_m
	{47}	word	@@@z_bit_0_a
	{48}	word	@@@z_bit_1_b
	{49}	word	@@@z_bit_1_c
	{4a}	word	@@@z_bit_1_d
	{4b}	word	@@@z_bit_1_e
	{4c}	word	@@@z_bit_1_h
	{4d}	word	@@@z_bit_1_l
	{4e}	word	@@@z_bit_1_m
	{4f}	word	@@@z_bit_1_a
	{50}	word	@@@z_bit_2_b
	{51}	word	@@@z_bit_2_c
	{52}	word	@@@z_bit_2_d
	{53}	word	@@@z_bit_2_e
	{54}	word	@@@z_bit_2_h
	{55}	word	@@@z_bit_2_l
	{56}	word	@@@z_bit_2_m
	{57}	word	@@@z_bit_2_a
	{58}	word	@@@z_bit_3_b
	{59}	word	@@@z_bit_3_c
	{5a}	word	@@@z_bit_3_d
	{5b}	word	@@@z_bit_3_e
	{5c}	word	@@@z_bit_3_h
	{5d}	word	@@@z_bit_3_l
	{5e}	word	@@@z_bit_3_m
	{5f}	word	@@@z_bit_3_a
	{60}	word	@@@z_bit_4_b
	{61}	word	@@@z_bit_4_c
	{62}	word	@@@z_bit_4_d
	{63}	word	@@@z_bit_4_e
	{64}	word	@@@z_bit_4_h
	{65}	word	@@@z_bit_4_l
	{66}	word	@@@z_bit_4_m
	{67}	word	@@@z_bit_4_a
	{68}	word	@@@z_bit_5_b
	{69}	word	@@@z_bit_5_c
	{6a}	word	@@@z_bit_5_d
	{6b}	word	@@@z_bit_5_e
	{6c}	word	@@@z_bit_5_h
	{6d}	word	@@@z_bit_5_l
	{6e}	word	@@@z_bit_5_m
	{6f}	word	@@@z_bit_5_a
	{70}	word	@@@z_bit_6_b
	{71}	word	@@@z_bit_6_c
	{72}	word	@@@z_bit_6_d
	{73}	word	@@@z_bit_6_e
	{74}	word	@@@z_bit_6_h
	{75}	word	@@@z_bit_6_l
	{76}	word	@@@z_bit_6_m
	{77}	word	@@@z_bit_6_a
	{78}	word	@@@z_bit_7_b
	{79}	word	@@@z_bit_7_c
	{7a}	word	@@@z_bit_7_d
	{7b}	word	@@@z_bit_7_e
	{7c}	word	@@@z_bit_7_h
	{7d}	word	@@@z_bit_7_l
	{7e}	word	@@@z_bit_7_m
	{7f}	word	@@@z_bit_7_a
	{80}	word	@@@z_res_0_b
	{81}	word	@@@z_res_0_c
	{82}	word	@@@z_res_0_d
	{83}	word	@@@z_res_0_e
	{84}	word	@@@z_res_0_h
	{85}	word	@@@z_res_0_l
	{86}	word	@@@z_res_0_m
	{87}	word	@@@z_res_0_a
	{88}	word	@@@z_res_1_b
	{89}	word	@@@z_res_1_c
	{8a}	word	@@@z_res_1_d
	{8b}	word	@@@z_res_1_e
	{8c}	word	@@@z_res_1_h
	{8d}	word	@@@z_res_1_l
	{8e}	word	@@@z_res_1_m
	{8f}	word	@@@z_res_1_a
	{90}	word	@@@z_res_2_b
	{91}	word	@@@z_res_2_c
	{92}	word	@@@z_res_2_d
	{93}	word	@@@z_res_2_e
	{94}	word	@@@z_res_2_h
	{95}	word	@@@z_res_2_l
	{96}	word	@@@z_res_2_m
	{97}	word	@@@z_res_2_a
	{98}	word	@@@z_res_3_b
	{99}	word	@@@z_res_3_c
	{9a}	word	@@@z_res_3_d
	{9b}	word	@@@z_res_3_e
	{9c}	word	@@@z_res_3_h
	{9d}	word	@@@z_res_3_l
	{9e}	word	@@@z_res_3_m
	{9f}	word	@@@z_res_3_a
	{a0}	word	@@@z_res_4_b
	{a1}	word	@@@z_res_4_c
	{a2}	word	@@@z_res_4_d
	{a3}	word	@@@z_res_4_e
	{a4}	word	@@@z_res_4_h
	{a5}	word	@@@z_res_4_l
	{a6}	word	@@@z_res_4_m
	{a7}	word	@@@z_res_4_a
	{a8}	word	@@@z_res_5_b
	{a9}	word	@@@z_res_5_c
	{aa}	word	@@@z_res_5_d
	{ab}	word	@@@z_res_5_e
	{ac}	word	@@@z_res_5_h
	{ad}	word	@@@z_res_5_l
	{ae}	word	@@@z_res_5_m
	{af}	word	@@@z_res_5_a
	{b0}	word	@@@z_res_6_b
	{b1}	word	@@@z_res_6_c
	{b2}	word	@@@z_res_6_d
	{b3}	word	@@@z_res_6_e
	{b4}	word	@@@z_res_6_h
	{b5}	word	@@@z_res_6_l
	{b6}	word	@@@z_res_6_m
	{b7}	word	@@@z_res_6_a
	{b8}	word	@@@z_res_7_b
	{b9}	word	@@@z_res_7_c
	{ba}	word	@@@z_res_7_d
	{bb}	word	@@@z_res_7_e
	{bc}	word	@@@z_res_7_h
	{bd}	word	@@@z_res_7_l
	{be}	word	@@@z_res_7_m
	{bf}	word	@@@z_res_7_a
	{c0}	word	@@@z_set_0_b
	{c1}	word	@@@z_set_0_c
	{c2}	word	@@@z_set_0_d
	{c3}	word	@@@z_set_0_e
	{c4}	word	@@@z_set_0_h
	{c5}	word	@@@z_set_0_l
	{c6}	word	@@@z_set_0_m
	{c7}	word	@@@z_set_0_a
	{c8}	word	@@@z_set_1_b
	{c9}	word	@@@z_set_1_c
	{ca}	word	@@@z_set_1_d
	{cb}	word	@@@z_set_1_e
	{cc}	word	@@@z_set_1_h
	{cd}	word	@@@z_set_1_l
	{ce}	word	@@@z_set_1_m
	{cf}	word	@@@z_set_1_a
	{d0}	word	@@@z_set_2_b
	{d1}	word	@@@z_set_2_c
	{d2}	word	@@@z_set_2_d
	{d3}	word	@@@z_set_2_e
	{d4}	word	@@@z_set_2_h
	{d5}	word	@@@z_set_2_l
	{d6}	word	@@@z_set_2_m
	{d7}	word	@@@z_set_2_a
	{d8}	word	@@@z_set_3_b
	{d9}	word	@@@z_set_3_c
	{da}	word	@@@z_set_3_d
	{db}	word	@@@z_set_3_e
	{dc}	word	@@@z_set_3_h
	{dd}	word	@@@z_set_3_l
	{de}	word	@@@z_set_3_m
	{df}	word	@@@z_set_3_a
	{e0}	word	@@@z_set_4_b
	{e1}	word	@@@z_set_4_c
	{e2}	word	@@@z_set_4_d
	{e3}	word	@@@z_set_4_e
	{e4}	word	@@@z_set_4_h
	{e5}	word	@@@z_set_4_l
	{e6}	word	@@@z_set_4_m
	{e7}	word	@@@z_set_4_a
	{e8}	word	@@@z_set_5_b
	{e9}	word	@@@z_set_5_c
	{ea}	word	@@@z_set_5_d
	{eb}	word	@@@z_set_5_e
	{ec}	word	@@@z_set_5_h
	{ed}	word	@@@z_set_5_l
	{ee}	word	@@@z_set_5_m
	{ef}	word	@@@z_set_5_a
	{f0}	word	@@@z_set_6_b
	{f1}	word	@@@z_set_6_c
	{f2}	word	@@@z_set_6_d
	{f3}	word	@@@z_set_6_e
	{f4}	word	@@@z_set_6_h
	{f5}	word	@@@z_set_6_l
	{f6}	word	@@@z_set_6_m
	{f7}	word	@@@z_set_6_a
	{f8}	word	@@@z_set_7_b
	{f9}	word	@@@z_set_7_c
	{fa}	word	@@@z_set_7_d
	{fb}	word	@@@z_set_7_e
	{fc}	word	@@@z_set_7_h
	{fd}	word	@@@z_set_7_l
	{fe}	word	@@@z_set_7_m
	{ff}	word	@@@z_set_7_a

snippets_xy
	{00}	word	@@@z_nop
	{01}	word	@@@z_ld_bc_imm16
	{02}	word	@@@z_ld_mem_bc_a
	{03}	word	@@@z_inc_bc
	{04}	word	@@@z_inc_b
	{05}	word	@@@z_dec_b
	{06}	word	@@@z_ld_b_imm8
	{07}	word	@@@z_rlca
	{08}	word	@@@z_ex_af_af2
	{09}	word	@@@z_add_xy_bc
	{0a}	word	@@@z_ld_a_mem_bc
	{0b}	word	@@@z_dec_bc
	{0c}	word	@@@z_inc_c
	{0d}	word	@@@z_dec_c
	{0e}	word	@@@z_ld_c_imm8
	{0f}	word	@@@z_rrca
	{10}	word	@@@z_djnz
	{11}	word	@@@z_ld_de_imm16
	{12}	word	@@@z_ld_mem_de_a
	{13}	word	@@@z_inc_de
	{14}	word	@@@z_inc_d
	{15}	word	@@@z_dec_d
	{16}	word	@@@z_ld_d_imm8
	{17}	word	@@@z_rla
	{18}	word	@@@z_jr
	{19}	word	@@@z_add_xy_de
	{1a}	word	@@@z_ld_a_mem_de
	{1b}	word	@@@z_dec_de
	{1c}	word	@@@z_inc_e
	{1d}	word	@@@z_dec_e
	{1e}	word	@@@z_ld_e_imm8
	{1f}	word	@@@z_rra
	{20}	word	@@@z_jr_nz
	{21}	word	@@@z_ld_xy_imm16
	{22}	word	@@@z_ld_abs16_xy
	{23}	word	@@@z_inc_xy
	{24}	word	@@@z_inc_hxy
	{25}	word	@@@z_dec_hxy
	{26}	word	@@@z_ld_hxy_imm8
	{27}	word	@@@z_daa
	{28}	word	@@@z_jr_z
	{29}	word	@@@z_add_xy_xy
	{2a}	word	@@@z_ld_xy_abs16
	{2b}	word	@@@z_dec_xy
	{2c}	word	@@@z_inc_lxy
	{2d}	word	@@@z_dec_lxy
	{2e}	word	@@@z_ld_lxy_imm8
	{2f}	word	@@@z_cpl
	{30}	word	@@@z_jr_nc
	{31}	word	@@@z_ld_sp_imm16
	{32}	word	@@@z_ld_mem_a
	{33}	word	@@@z_inc_sp
	{34}	word	@@@z_inc_mxy
	{35}	word	@@@z_dec_mxy
	{36}	word	@@@z_ld_mxy_imm8
	{37}	word	@@@z_scf
	{38}	word	@@@z_jr_c
	{39}	word	@@@z_add_xy_sp
	{3a}	word	@@@z_ld_a_mem
	{3b}	word	@@@z_dec_sp
	{3c}	word	@@@z_inc_a
	{3d}	word	@@@z_dec_a
	{3e}	word	@@@z_ld_a_imm8
	{3f}	word	@@@z_ccf
	{40}	word	@@@z_nop		' z_ld_b_b
	{41}	word	@@@z_ld_b_c
	{42}	word	@@@z_ld_b_d
	{43}	word	@@@z_ld_b_e
	{44}	word	@@@z_ld_b_hxy
	{45}	word	@@@z_ld_b_lxy
	{46}	word	@@@z_ld_b_mxy
	{47}	word	@@@z_ld_b_a
	{48}	word	@@@z_ld_c_b
	{49}	word	@@@z_nop		' z_ld_c_c
	{4a}	word	@@@z_ld_c_d
	{4b}	word	@@@z_ld_c_e
	{4c}	word	@@@z_ld_c_hxy
	{4d}	word	@@@z_ld_c_lxy
	{4e}	word	@@@z_ld_c_mxy
	{4f}	word	@@@z_ld_c_a
	{50}	word	@@@z_ld_d_b
	{51}	word	@@@z_ld_d_c
	{52}	word	@@@z_nop		' z_ld_d_d
	{53}	word	@@@z_ld_d_e
	{54}	word	@@@z_ld_d_hxy
	{55}	word	@@@z_ld_d_lxy
	{56}	word	@@@z_ld_d_mxy
	{57}	word	@@@z_ld_d_a
	{58}	word	@@@z_ld_e_b
	{59}	word	@@@z_ld_e_c
	{5a}	word	@@@z_ld_e_d
	{5b}	word	@@@z_nop		' z_ld_e_e
	{5c}	word	@@@z_ld_e_hxy
	{5d}	word	@@@z_ld_e_lxy
	{5e}	word	@@@z_ld_e_mxy
	{5f}	word	@@@z_ld_e_a
	{60}	word	@@@z_ld_hxy_b
	{61}	word	@@@z_ld_hxy_c
	{62}	word	@@@z_ld_hxy_d
	{63}	word	@@@z_ld_hxy_e
	{64}	word	@@@z_nop		' z_ld_hxy_hxy
	{65}	word	@@@z_ld_hxy_lxy
	{66}	word	@@@z_ld_h_mxy
	{67}	word	@@@z_ld_hxy_a
	{68}	word	@@@z_ld_lxy_b
	{69}	word	@@@z_ld_lxy_c
	{6a}	word	@@@z_ld_lxy_d
	{6b}	word	@@@z_ld_lxy_e
	{6c}	word	@@@z_ld_lxy_hxy
	{6d}	word	@@@z_nop		' z_ld_lxy_lxy
	{6e}	word	@@@z_ld_l_mxy
	{6f}	word	@@@z_ld_lxy_a
	{70}	word	@@@z_ld_mxy_b
	{71}	word	@@@z_ld_mxy_c
	{72}	word	@@@z_ld_mxy_d
	{73}	word	@@@z_ld_mxy_e
	{74}	word	@@@z_ld_mxy_h
	{75}	word	@@@z_ld_mxy_l
	{76}	word	@@@z_halt
	{77}	word	@@@z_ld_mxy_a
	{78}	word	@@@z_ld_a_b
	{79}	word	@@@z_ld_a_c
	{7a}	word	@@@z_ld_a_d
	{7b}	word	@@@z_ld_a_e
	{7c}	word	@@@z_ld_a_hxy
	{7d}	word	@@@z_ld_a_lxy
	{7e}	word	@@@z_ld_a_mxy
	{7f}	word	@@@z_nop		' z_ld_a_a
	{80}	word	@@@z_add_a_b
	{81}	word	@@@z_add_a_c
	{82}	word	@@@z_add_a_d
	{83}	word	@@@z_add_a_e
	{84}	word	@@@z_add_a_hxy
	{85}	word	@@@z_add_a_lxy
	{86}	word	@@@z_add_a_mxy
	{87}	word	@@@z_add_a_a
	{88}	word	@@@z_adc_a_b
	{89}	word	@@@z_adc_a_c
	{8a}	word	@@@z_adc_a_d
	{8b}	word	@@@z_adc_a_e
	{8c}	word	@@@z_adc_a_hxy
	{8d}	word	@@@z_adc_a_lxy
	{8e}	word	@@@z_adc_a_mxy
	{8f}	word	@@@z_adc_a_a
	{90}	word	@@@z_sub_a_b
	{91}	word	@@@z_sub_a_c
	{92}	word	@@@z_sub_a_d
	{93}	word	@@@z_sub_a_e
	{94}	word	@@@z_sub_a_hxy
	{95}	word	@@@z_sub_a_lxy
	{96}	word	@@@z_sub_a_mxy
	{97}	word	@@@z_sub_a_a
	{98}	word	@@@z_sbc_a_b
	{99}	word	@@@z_sbc_a_c
	{9a}	word	@@@z_sbc_a_d
	{9b}	word	@@@z_sbc_a_e
	{9c}	word	@@@z_sbc_a_hxy
	{9d}	word	@@@z_sbc_a_lxy
	{9e}	word	@@@z_sbc_a_mxy
	{9f}	word	@@@z_sbc_a_a
	{a0}	word	@@@z_and_a_b
	{a1}	word	@@@z_and_a_c
	{a2}	word	@@@z_and_a_d
	{a3}	word	@@@z_and_a_e
	{a4}	word	@@@z_and_a_hxy
	{a5}	word	@@@z_and_a_lxy
	{a6}	word	@@@z_and_a_mxy
	{a7}	word	@@@z_and_a_a
	{a8}	word	@@@z_xor_a_b
	{a9}	word	@@@z_xor_a_c
	{aa}	word	@@@z_xor_a_d
	{ab}	word	@@@z_xor_a_e
	{ac}	word	@@@z_xor_a_hxy
	{ad}	word	@@@z_xor_a_lxy
	{ae}	word	@@@z_xor_a_mxy
	{af}	word	@@@z_xor_a_a
	{b0}	word	@@@z_or_a_b
	{b1}	word	@@@z_or_a_c
	{b2}	word	@@@z_or_a_d
	{b3}	word	@@@z_or_a_e
	{b4}	word	@@@z_or_a_hxy
	{b5}	word	@@@z_or_a_lxy
	{b6}	word	@@@z_or_a_mxy
	{b7}	word	@@@z_or_a_a
	{b8}	word	@@@z_cp_a_b
	{b9}	word	@@@z_cp_a_c
	{ba}	word	@@@z_cp_a_d
	{bb}	word	@@@z_cp_a_e
	{bc}	word	@@@z_cp_a_hxy
	{bd}	word	@@@z_cp_a_lxy
	{be}	word	@@@z_cp_a_mxy
	{bf}	word	@@@z_cp_a_a
	{c0}	word	@@@z_ret_nz
	{c1}	word	@@@z_pop_bc
	{c2}	word	@@@z_jp_nz
	{c3}	word	@@@z_jp
	{c4}	word	@@@z_call_nz
	{c5}	word	@@@z_push_bc
	{c6}	word	@@@z_add_a_imm8
	{c7}	word	@@@z_rst_00
	{c8}	word	@@@z_ret_z
	{c9}	word	@@@z_ret
	{ca}	word	@@@z_jp_z
	{cb}	word	@@@z_nop		' z_dd_cb_xx
	{cc}	word	@@@z_call_z
	{cd}	word	@@@z_call
	{ce}	word	@@@z_adc_a_imm8
	{c7}	word	@@@z_rst_08
	{d0}	word	@@@z_ret_nc
	{d1}	word	@@@z_pop_de
	{d2}	word	@@@z_jp_nc
	{d3}	word	@@@z_out_port_a
	{d4}	word	@@@z_call_nc
	{d5}	word	@@@z_push_de
	{d6}	word	@@@z_sub_imm8
	{d7}	word	@@@z_rst_10
	{d8}	word	@@@z_ret_c
	{d9}	word	@@@z_exx
	{da}	word	@@@z_jp_c
	{db}	word	@@@z_in_a_port
	{dc}	word	@@@z_call_c
	{dd}	word	@@@z_dd_xx
	{de}	word	@@@z_sbc_a_imm8
	{df}	word	@@@z_rst_18
	{e0}	word	@@@z_ret_po
	{e1}	word	@@@z_pop_xy
	{e2}	word	@@@z_jp_po
	{e3}	word	@@@z_ex_sp_xy
	{e4}	word	@@@z_call_po
	{e5}	word	@@@z_push_xy
	{e6}	word	@@@z_and_imm8
	{e7}	word	@@@z_rst_20
	{e8}	word	@@@z_ret_pe
	{e9}	word	@@@z_jp_xy
	{ea}	word	@@@z_jp_pe
	{eb}	word	@@@z_ex_de_hl
	{ec}	word	@@@z_call_pe
	{ed}	word	@@@z_ed_xx
	{ee}	word	@@@z_xor_imm8
	{ef}	word	@@@z_rst_28
	{f0}	word	@@@z_ret_p
	{f1}	word	@@@z_pop_af
	{f2}	word	@@@z_jp_p
	{f3}	word	@@@z_di
	{f4}	word	@@@z_call_p
	{f5}	word	@@@z_push_af
	{f6}	word	@@@z_or_imm8
	{f7}	word	@@@z_rst_30
	{f8}	word	@@@z_ret_m
	{f9}	word	@@@z_ld_sp_xy
	{fa}	word	@@@z_jp_m
	{fb}	word	@@@z_ei
	{fc}	word	@@@z_call_m
	{fd}	word	@@@z_fd_xx
	{fe}	word	@@@z_cp_imm8
	{ff}	word	@@@z_rst_38

snippets_ed
	{00}	word	@@@z_bcde_cnt
	{01}	word	@@@z_illegal
	{02}	word	@@@z_illegal
	{03}	word	@@@z_illegal
	{04}	word	@@@z_illegal
	{05}	word	@@@z_illegal
	{06}	word	@@@z_illegal
	{07}	word	@@@z_illegal
	{08}	word	@@@z_illegal
	{09}	word	@@@z_illegal
	{0a}	word	@@@z_illegal
	{0b}	word	@@@z_illegal
	{0c}	word	@@@z_illegal
	{0d}	word	@@@z_illegal
	{0e}	word	@@@z_illegal
	{0f}	word	@@@z_illegal
	{10}	word	@@@z_illegal
	{11}	word	@@@z_illegal
	{12}	word	@@@z_illegal
	{13}	word	@@@z_illegal
	{14}	word	@@@z_illegal
	{15}	word	@@@z_illegal
	{16}	word	@@@z_illegal
	{17}	word	@@@z_illegal
	{18}	word	@@@z_illegal
	{19}	word	@@@z_illegal
	{1a}	word	@@@z_illegal
	{1b}	word	@@@z_illegal
	{1c}	word	@@@z_illegal
	{1d}	word	@@@z_illegal
	{1e}	word	@@@z_illegal
	{1f}	word	@@@z_illegal
	{20}	word	@@@z_illegal
	{21}	word	@@@z_illegal
	{22}	word	@@@z_illegal
	{23}	word	@@@z_illegal
	{24}	word	@@@z_illegal
	{25}	word	@@@z_illegal
	{26}	word	@@@z_illegal
	{27}	word	@@@z_illegal
	{28}	word	@@@z_illegal
	{29}	word	@@@z_illegal
	{2a}	word	@@@z_illegal
	{2b}	word	@@@z_illegal
	{2c}	word	@@@z_illegal
	{2d}	word	@@@z_illegal
	{2e}	word	@@@z_illegal
	{2f}	word	@@@z_illegal
	{30}	word	@@@z_illegal
	{31}	word	@@@z_illegal
	{32}	word	@@@z_illegal
	{33}	word	@@@z_illegal
	{34}	word	@@@z_illegal
	{35}	word	@@@z_illegal
	{36}	word	@@@z_illegal
	{37}	word	@@@z_illegal
	{38}	word	@@@z_illegal
	{39}	word	@@@z_illegal
	{3a}	word	@@@z_illegal
	{3b}	word	@@@z_illegal
	{3c}	word	@@@z_illegal
	{3d}	word	@@@z_illegal
	{3e}	word	@@@z_illegal
	{3f}	word	@@@z_illegal
	{40}	word	@@@z_inp_b_bc
	{41}	word	@@@z_out_bc_b
	{42}	word	@@@z_sbc_hl_bc
	{43}	word	@@@z_ld_abs16_bc
	{44}	word	@@@z_neg
	{45}	word	@@@z_retn
	{46}	word	@@@z_im_0
	{47}	word	@@@z_ld_i_a
	{48}	word	@@@z_inp_c_bc
	{49}	word	@@@z_out_bc_c
	{4a}	word	@@@z_adc_hl_bc
	{4b}	word	@@@z_ld_bc_abs16
	{4c}	word	@@@z_neg
	{4d}	word	@@@z_reti
	{4e}	word	@@@z_im_0
	{4f}	word	@@@z_ld_r_a
	{50}	word	@@@z_inp_d_bc
	{51}	word	@@@z_out_bc_d
	{52}	word	@@@z_sbc_hl_de
	{53}	word	@@@z_ld_abs16_de
	{54}	word	@@@z_neg
	{55}	word	@@@z_retn
	{56}	word	@@@z_im_1
	{57}	word	@@@z_ld_a_i
	{58}	word	@@@z_inp_e_bc
	{59}	word	@@@z_out_bc_e
	{5a}	word	@@@z_adc_hl_de
	{5b}	word	@@@z_ld_de_abs16
	{5c}	word	@@@z_neg
	{5d}	word	@@@z_reti
	{5e}	word	@@@z_im_2
	{5f}	word	@@@z_ld_a_r
	{60}	word	@@@z_inp_h_bc
	{61}	word	@@@z_out_bc_h
	{62}	word	@@@z_sbc_hl_hl
	{63}	word	@@@z_ld_abs16_hl
	{64}	word	@@@z_neg
	{65}	word	@@@z_retn
	{66}	word	@@@z_im_0
	{67}	word	@@@z_rrd_m
	{68}	word	@@@z_inp_l_bc
	{69}	word	@@@z_out_bc_l
	{6a}	word	@@@z_adc_hl_hl
	{6b}	word	@@@z_ld_hl_abs16
	{6c}	word	@@@z_neg
	{6d}	word	@@@z_reti
	{6e}	word	@@@z_im_0
	{6f}	word	@@@z_rld_m
	{70}	word	@@@z_inp_0_bc
	{71}	word	@@@z_out_bc_0
	{72}	word	@@@z_sbc_hl_sp
	{73}	word	@@@z_ld_abs16_sp
	{74}	word	@@@z_neg
	{75}	word	@@@z_retn
	{76}	word	@@@z_im_1
	{77}	word	@@@z_illegal
	{78}	word	@@@z_inp_a_bc
	{79}	word	@@@z_out_bc_a
	{7a}	word	@@@z_adc_hl_sp
	{7b}	word	@@@z_ld_sp_abs16
	{7c}	word	@@@z_neg
	{7d}	word	@@@z_reti
	{7e}	word	@@@z_im_2
	{7f}	word	@@@z_illegal
	{80}	word	@@@z_illegal
	{81}	word	@@@z_illegal
	{82}	word	@@@z_illegal
	{83}	word	@@@z_illegal
	{84}	word	@@@z_illegal
	{85}	word	@@@z_illegal
	{86}	word	@@@z_illegal
	{87}	word	@@@z_illegal
	{88}	word	@@@z_illegal
	{89}	word	@@@z_illegal
	{8a}	word	@@@z_illegal
	{8b}	word	@@@z_illegal
	{8c}	word	@@@z_illegal
	{8d}	word	@@@z_illegal
	{8e}	word	@@@z_illegal
	{8f}	word	@@@z_illegal
	{90}	word	@@@z_illegal
	{91}	word	@@@z_illegal
	{92}	word	@@@z_illegal
	{93}	word	@@@z_illegal
	{94}	word	@@@z_illegal
	{95}	word	@@@z_illegal
	{96}	word	@@@z_illegal
	{97}	word	@@@z_illegal
	{98}	word	@@@z_illegal
	{99}	word	@@@z_illegal
	{9a}	word	@@@z_illegal
	{9b}	word	@@@z_illegal
	{9c}	word	@@@z_illegal
	{9d}	word	@@@z_illegal
	{9e}	word	@@@z_illegal
	{9f}	word	@@@z_illegal
	{a0}	word	@@@z_ldi
	{a1}	word	@@@z_cpi
	{a2}	word	@@@z_ini
	{a3}	word	@@@z_outi
	{a4}	word	@@@z_illegal
	{a5}	word	@@@z_illegal
	{a6}	word	@@@z_illegal
	{a7}	word	@@@z_illegal
	{a8}	word	@@@z_ldd
	{a9}	word	@@@z_cpd
	{aa}	word	@@@z_ind
	{ab}	word	@@@z_outd
	{ac}	word	@@@z_illegal
	{ad}	word	@@@z_illegal
	{ae}	word	@@@z_illegal
	{af}	word	@@@z_illegal
	{b0}	word	@@@z_ldir
	{b1}	word	@@@z_cpir
	{b2}	word	@@@z_inir
	{b3}	word	@@@z_otir
	{b4}	word	@@@z_illegal
	{b5}	word	@@@z_illegal
	{b6}	word	@@@z_illegal
	{b7}	word	@@@z_illegal
	{b8}	word	@@@z_lddr
	{b9}	word	@@@z_cpdr
	{ba}	word	@@@z_indr
	{bb}	word	@@@z_otdr
	{bc}	word	@@@z_illegal
	{bd}	word	@@@z_illegal
	{be}	word	@@@z_illegal
	{bf}	word	@@@z_illegal
	{c0}	word	@@@z_illegal
	{c1}	word	@@@z_illegal
	{c2}	word	@@@z_illegal
	{c3}	word	@@@z_illegal
	{c4}	word	@@@z_illegal
	{c5}	word	@@@z_illegal
	{c6}	word	@@@z_illegal
	{c7}	word	@@@z_illegal
	{c8}	word	@@@z_illegal
	{c9}	word	@@@z_illegal
	{ca}	word	@@@z_illegal
	{cb}	word	@@@z_illegal
	{cc}	word	@@@z_illegal
	{cd}	word	@@@z_illegal
	{ce}	word	@@@z_illegal
	{cf}	word	@@@z_illegal
	{d0}	word	@@@z_illegal
	{d1}	word	@@@z_illegal
	{d2}	word	@@@z_illegal
	{d3}	word	@@@z_illegal
	{d4}	word	@@@z_illegal
	{d5}	word	@@@z_illegal
	{d6}	word	@@@z_illegal
	{d7}	word	@@@z_illegal
	{d8}	word	@@@z_illegal
	{d9}	word	@@@z_illegal
	{da}	word	@@@z_illegal
	{db}	word	@@@z_illegal
	{dc}	word	@@@z_illegal
	{dd}	word	@@@z_illegal
	{de}	word	@@@z_illegal
	{df}	word	@@@z_illegal
	{e0}	word	@@@z_illegal
	{e1}	word	@@@z_illegal
	{e2}	word	@@@z_illegal
	{e3}	word	@@@z_illegal
	{e4}	word	@@@z_illegal
	{e5}	word	@@@z_illegal
	{e6}	word	@@@z_illegal
	{e7}	word	@@@z_illegal
	{e8}	word	@@@z_illegal
	{e9}	word	@@@z_illegal
	{ea}	word	@@@z_illegal
	{eb}	word	@@@z_illegal
	{ec}	word	@@@z_illegal
	{ed}	word	@@@z_illegal
	{ee}	word	@@@z_illegal
	{ef}	word	@@@z_illegal
	{f0}	word	@@@z_illegal
	{f1}	word	@@@z_illegal
	{f2}	word	@@@z_illegal
	{f3}	word	@@@z_illegal
	{f4}	word	@@@z_illegal
	{f5}	word	@@@z_illegal
	{f6}	word	@@@z_illegal
	{f7}	word	@@@z_illegal
	{f8}	word	@@@z_illegal
	{f9}	word	@@@z_illegal
	{fa}	word	@@@z_illegal
	{fb}	word	@@@z_illegal
	{fc}	word	@@@z_illegal
	{fd}	word	@@@z_illegal
	{fe}	word	@@@z_illegal
	{ff}	word	@@@z_illegal

snippets_xy_cb
	{00-07}	word	@@@z_rlc_mxy
	{08-0f}	word	@@@z_rrc_mxy
	{10-17}	word	@@@z_rl_mxy
	{18-1f}	word	@@@z_rr_mxy
	{20-27}	word	@@@z_sla_mxy
	{28-2f}	word	@@@z_sra_mxy
	{30-37}	word	@@@z_sli_mxy
	{38-3f}	word	@@@z_srl_mxy
	{40-47}	word	@@@z_bit_0_mxy
	{48-4f}	word	@@@z_bit_1_mxy
	{50-57}	word	@@@z_bit_2_mxy
	{58-5f}	word	@@@z_bit_3_mxy
	{60-67}	word	@@@z_bit_4_mxy
	{68-6f}	word	@@@z_bit_5_mxy
	{70-77}	word	@@@z_bit_6_mxy
	{78-7f}	word	@@@z_bit_7_mxy
	{80-87}	word	@@@z_res_0_mxy
	{88-8f}	word	@@@z_res_1_mxy
	{90-97}	word	@@@z_res_2_mxy
	{98-9f}	word	@@@z_res_3_mxy
	{a0-a7}	word	@@@z_res_4_mxy
	{a8-af}	word	@@@z_res_5_mxy
	{b0-b7}	word	@@@z_res_6_mxy
	{b8-bf}	word	@@@z_res_7_mxy
	{c0-c7}	word	@@@z_set_0_mxy
	{c8-cf}	word	@@@z_set_1_mxy
	{d0-d7}	word	@@@z_set_2_mxy
	{d8-df}	word	@@@z_set_3_mxy
	{e0-e7}	word	@@@z_set_4_mxy
	{e8-ef}	word	@@@z_set_5_mxy
	{f0-f7}	word	@@@z_set_6_mxy
	{f8-ff}	word	@@@z_set_7_mxy
